@@ -9,19 +9,23 @@
 
 #include "system.h"
 
-#include <rpmio_internal.h>
-#include <rpmlib.h>
+#if MAJOR_IN_MKDEV
+#include <sys/mkdev.h>
+#elif MAJOR_IN_SYSMACROS
+#include <sys/sysmacros.h>
+#else
+#include <sys/types.h> /* already included from system.h */
+#endif
+#include <errno.h>
 
-#include "cpio.h"
-#include "fsm.h"
+#include <rpm/rpmio.h>
+#include <rpm/rpmlog.h>
+
+#include "lib/cpio.h"
+#include "lib/fsm.h"
 
-#include "rpmerr.h"
 #include "debug.h"
 
-/*@access FSM_t @*/
-
-/*@unchecked@*/
-int _cpio_debug = 0;
 
 /**
  * Convert string to unsigned integer (with buffer size check).
@@ -31,38 +35,32 @@ int _cpio_debug = 0;
  * @param num		max no. of bytes to read
  * @return		converted integer
  */
-static int strntoul(const char *str, /*@out@*/char **endptr, int base, int num)
-	/*@modifies *endptr @*/
-	/*@requires maxSet(endptr) >= 0 @*/
+static unsigned long strntoul(const char *str,char **endptr, int base, size_t num)
 {
-    char * buf, * end;
+    char buf[num+1], * end;
     unsigned long ret;
 
-    buf = alloca(num + 1);
     strncpy(buf, str, num);
     buf[num] = '\0';
 
     ret = strtoul(buf, &end, base);
-/*@-boundsread@*/ /* LCL: strtoul annotations */
     if (*end != '\0')
 	*endptr = ((char *)str) + (end - buf);	/* XXX discards const */
     else
 	*endptr = ((char *)str) + strlen(buf);
-/*@=boundsread@*/
 
     return ret;
 }
 
 #define GET_NUM_FIELD(phys, log) \
-	/*@-boundswrite@*/ \
+	\
 	log = strntoul(phys, &end, 16, sizeof(phys)); \
-	/*@=boundswrite@*/ \
+	\
 	if ( (end - phys) != sizeof(phys) ) return CPIOERR_BAD_HEADER;
 #define SET_NUM_FIELD(phys, val, space) \
 	sprintf(space, "%8.8lx", (unsigned long) (val)); \
-	/*@-boundsread@*/ \
+	\
 	memcpy(phys, space, 8) \
-	/*@=boundsread@*/
 
 int cpioTrailerWrite(FSM_t fsm)
 {
@@ -70,13 +68,11 @@ int cpioTrailerWrite(FSM_t fsm)
 	(struct cpioCrcPhysicalHeader *)fsm->rdbuf;
     int rc;
 
-/*@-boundswrite@*/
     memset(hdr, '0', PHYS_HDR_SIZE);
     memcpy(hdr->magic, CPIO_NEWC_MAGIC, sizeof(hdr->magic));
     memcpy(hdr->nlink, "00000001", 8);
     memcpy(hdr->namesize, "0000000b", 8);
     memcpy(fsm->rdbuf + PHYS_HDR_SIZE, CPIO_TRAILER, sizeof(CPIO_TRAILER));
-/*@=boundswrite@*/
 
     /* XXX DWRITE uses rdnb for I/O length. */
     fsm->rdnb = PHYS_HDR_SIZE + sizeof(CPIO_TRAILER);
@@ -100,9 +96,6 @@ int cpioHeaderWrite(FSM_t fsm, struct stat * st)
     dev_t dev;
     int rc = 0;
 
-if (_cpio_debug)
-fprintf(stderr, "    %s(%p, %p)\n", __FUNCTION__, fsm, st);
-
     memcpy(hdr->magic, CPIO_NEWC_MAGIC, sizeof(hdr->magic));
     SET_NUM_FIELD(hdr->inode, st->st_ino, field);
     SET_NUM_FIELD(hdr->mode, st->st_mode, field);
@@ -112,16 +105,14 @@ fprintf(stderr, "    %s(%p, %p)\n", __FUNCTION__, fsm, st);
     SET_NUM_FIELD(hdr->mtime, st->st_mtime, field);
     SET_NUM_FIELD(hdr->filesize, st->st_size, field);
 
-    dev = major((unsigned)st->st_dev); SET_NUM_FIELD(hdr->devMajor, dev, field);
-    dev = minor((unsigned)st->st_dev); SET_NUM_FIELD(hdr->devMinor, dev, field);
-    dev = major((unsigned)st->st_rdev); SET_NUM_FIELD(hdr->rdevMajor, dev, field);
-    dev = minor((unsigned)st->st_rdev); SET_NUM_FIELD(hdr->rdevMinor, dev, field);
+    dev = major(st->st_dev); SET_NUM_FIELD(hdr->devMajor, dev, field);
+    dev = minor(st->st_dev); SET_NUM_FIELD(hdr->devMinor, dev, field);
+    dev = major(st->st_rdev); SET_NUM_FIELD(hdr->rdevMajor, dev, field);
+    dev = minor(st->st_rdev); SET_NUM_FIELD(hdr->rdevMinor, dev, field);
 
     len = strlen(fsm->path) + 1; SET_NUM_FIELD(hdr->namesize, len, field);
     memcpy(hdr->checksum, "00000000", 8);
-/*@-boundswrite@*/
     memcpy(fsm->rdbuf + PHYS_HDR_SIZE, fsm->path, len);
-/*@=boundswrite@*/
 
     /* XXX DWRITE uses rdnb for I/O length. */
     fsm->rdnb = PHYS_HDR_SIZE + len;
@@ -130,41 +121,23 @@ fprintf(stderr, "    %s(%p, %p)\n", __FUNCTION__, fsm, st);
 	rc = CPIOERR_WRITE_FAILED;
     if (!rc)
 	rc = fsmNext(fsm, FSM_PAD);
-
-    if (!rc && S_ISLNK(st->st_mode)) {
-assert(fsm->lpath != NULL);
-	fsm->rdnb = strlen(fsm->lpath);
-	memcpy(fsm->rdbuf, fsm->lpath, fsm->rdnb);
-	rc = fsmNext(fsm, FSM_DWRITE);
-	if (!rc && fsm->rdnb != fsm->wrnb)
-	    rc = CPIOERR_WRITE_FAILED;
-	if (!rc)
-	    rc = fsmNext(fsm, FSM_PAD);
-    }
-
     return rc;
 }
 
 int cpioHeaderRead(FSM_t fsm, struct stat * st)
-	/*@modifies fsm, *st @*/
 {
     struct cpioCrcPhysicalHeader hdr;
     int nameSize;
     char * end;
-    int major, minor;
+    unsigned int major, minor;
     int rc = 0;
 
-if (_cpio_debug)
-fprintf(stderr, "    %s(%p, %p)\n", __FUNCTION__, fsm, st);
-
     fsm->wrlen = PHYS_HDR_SIZE;
     rc = fsmNext(fsm, FSM_DREAD);
     if (!rc && fsm->rdnb != fsm->wrlen)
 	rc = CPIOERR_READ_FAILED;
     if (rc) return rc;
-/*@-boundswrite@*/
     memcpy(&hdr, fsm->wrbuf, fsm->rdnb);
-/*@=boundswrite@*/
 
     if (strncmp(CPIO_CRC_MAGIC, hdr.magic, sizeof(CPIO_CRC_MAGIC)-1) &&
 	strncmp(CPIO_NEWC_MAGIC, hdr.magic, sizeof(CPIO_NEWC_MAGIC)-1))
@@ -180,15 +153,11 @@ fprintf(stderr, "    %s(%p, %p)\n", __FUNCTION__, fsm, st);
 
     GET_NUM_FIELD(hdr.devMajor, major);
     GET_NUM_FIELD(hdr.devMinor, minor);
-    /*@-shiftimplementation@*/
     st->st_dev = makedev(major, minor);
-    /*@=shiftimplementation@*/
 
     GET_NUM_FIELD(hdr.rdevMajor, major);
     GET_NUM_FIELD(hdr.rdevMinor, minor);
-    /*@-shiftimplementation@*/
     st->st_rdev = makedev(major, minor);
-    /*@=shiftimplementation@*/
 
     GET_NUM_FIELD(hdr.namesize, nameSize);
     if (nameSize >= fsm->wrsize)
@@ -204,50 +173,29 @@ fprintf(stderr, "    %s(%p, %p)\n", __FUNCTION__, fsm, st);
 	    fsm->path = NULL;
 	    return rc;
 	}
-/*@-boundswrite@*/
 	memcpy(t, fsm->wrbuf, fsm->rdnb);
 	t[nameSize] = '\0';
-/*@=boundswrite@*/
 	fsm->path = t;
     }
 
-    if (S_ISLNK(st->st_mode)) {
-	rc = fsmNext(fsm, FSM_POS);
-	if (rc) return rc;
-	fsm->wrlen = st->st_size;
-	rc = fsmNext(fsm, FSM_DREAD);
-	if (!rc && fsm->rdnb != fsm->wrlen)
-	    rc = CPIOERR_READ_FAILED;
-	if (rc) return rc;
-/*@-boundswrite@*/
-	fsm->wrbuf[st->st_size] = '\0';
-/*@=boundswrite@*/
-	fsm->lpath = xstrdup(fsm->wrbuf);
-    }
-
-if (_cpio_debug)
-fprintf(stderr, "\t     %06o%3d (%4d,%4d)%10d %s\n\t-> %s\n",
-                (unsigned)st->st_mode, (int)st->st_nlink,
-                (int)st->st_uid, (int)st->st_gid, (int)st->st_size,
-                (fsm->path ? fsm->path : ""), (fsm->lpath ? fsm->lpath : ""));
-
-    return rc;
+    return 0;
 }
 
 const char * cpioStrerror(int rc)
 {
     static char msg[256];
-    char *s;
-    int l, myerrno = errno;
+    const char *s;
+    int myerrno = errno;
+    size_t l;
 
     strcpy(msg, "cpio: ");
-    /*@-branchstate@*/
     switch (rc) {
-    default:
-	s = msg + strlen(msg);
-	sprintf(s, _("(error 0x%x)"), (unsigned)rc);
+    default: {
+	char *t = msg + strlen(msg);
+	sprintf(t, _("(error 0x%x)"), (unsigned)rc);
 	s = NULL;
 	break;
+    }
     case CPIOERR_BAD_MAGIC:	s = _("Bad magic");		break;
     case CPIOERR_BAD_HEADER:	s = _("Bad/unreadable  header");break;
 
@@ -270,30 +218,28 @@ const char * cpioStrerror(int rc)
     case CPIOERR_READ_FAILED:	s = "read";	break;
     case CPIOERR_COPY_FAILED:	s = "copy";	break;
     case CPIOERR_LSETFCON_FAILED: s = "lsetfilecon";	break;
+    case CPIOERR_SETCAP_FAILED: s = "cap_set_file";	break;
 
     case CPIOERR_HDR_SIZE:	s = _("Header size too big");	break;
     case CPIOERR_UNKNOWN_FILETYPE: s = _("Unknown file type");	break;
     case CPIOERR_MISSING_HARDLINK: s = _("Missing hard link(s)"); break;
-    case CPIOERR_DIGEST_MISMATCH: s = _("File digest mismatch");	break;
+    case CPIOERR_DIGEST_MISMATCH: s = _("Digest mismatch");	break;
     case CPIOERR_INTERNAL:	s = _("Internal error");	break;
     case CPIOERR_UNMAPPED_FILE:	s = _("Archive file not in header"); break;
     case CPIOERR_ENOENT:	s = strerror(ENOENT); break;
     case CPIOERR_ENOTEMPTY:	s = strerror(ENOTEMPTY); break;
     }
-    /*@=branchstate@*/
 
     l = sizeof(msg) - strlen(msg) - 1;
     if (s != NULL) {
 	if (l > 0) strncat(msg, s, l);
 	l -= strlen(s);
     }
-    /*@-branchstate@*/
     if ((rc & CPIOERR_CHECK_ERRNO) && myerrno) {
 	s = _(" failed - ");
 	if (l > 0) strncat(msg, s, l);
 	l -= strlen(s);
 	if (l > 0) strncat(msg, strerror(myerrno), l);
     }
-    /*@=branchstate@*/
     return msg;
 }
