@@ -9,15 +9,14 @@
 /* network byte order and is converted on the fly to host order. */
 
 #include "system.h"
-
+#include <netdb.h>
 #include <rpm/rpmtypes.h>
 #include <rpm/rpmstring.h>
 #include "lib/header_internal.h"
+#include "lib/misc.h"			/* tag function proto */
 
 #include "debug.h"
 
-int _hdr_debug = 0;
-
 /** \ingroup header
  */
 const unsigned char rpm_header_magic[8] = {
@@ -68,41 +67,59 @@ static const int typeSizes[16] =  {
     0
 };
 
+enum headerFlags_e {
+    HEADERFLAG_SORTED    = (1 << 0), /*!< Are header entries sorted? */
+    HEADERFLAG_ALLOCATED = (1 << 1), /*!< Is 1st header region allocated? */
+    HEADERFLAG_LEGACY    = (1 << 2), /*!< Header came from legacy source? */
+    HEADERFLAG_DEBUG     = (1 << 3), /*!< Debug this header? */
+};
+
+typedef rpmFlags headerFlags;
+
+/** \ingroup header
+ * The Header data structure.
+ */
+struct headerToken_s {
+    void * blob;		/*!< Header region blob. */
+    indexEntry index;		/*!< Array of tags. */
+    int indexUsed;		/*!< Current size of tag array. */
+    int indexAlloced;		/*!< Allocated size of tag array. */
+    unsigned int instance;	/*!< Rpmdb instance (offset) */
+    headerFlags flags;
+    int nrefs;			/*!< Reference count. */
+};
+
 /** \ingroup header
  * Maximum no. of bytes permitted in a header.
  */
 static const size_t headerMaxbytes = (32*1024*1024);
 
-/** \ingroup header
- * HEADER_EXT_TAG format function prototype.
- * This is allowed to fail, which indicates the tag doesn't exist.
- *
- * @param h		header
- * @retval td		tag data container
- * @param flags		modifier flags
- * @return		0 on success
- */
-typedef int (*headerTagTagFunction) (Header h, rpmtd td, headerGetFlags hgflags);
+#define	INDEX_MALLOC_SIZE	8
 
-extern void *rpmHeaderTagFunc(rpmTag tag);
+#define	ENTRY_IS_REGION(_e) \
+	(((_e)->info.tag >= RPMTAG_HEADERIMAGE) && ((_e)->info.tag < RPMTAG_HEADERREGIONS))
+#define	ENTRY_IN_REGION(_e)	((_e)->info.offset < 0)
+
+/* Convert a 64bit value to network byte order. */
+static uint64_t htonll( uint64_t n ) {
+    uint32_t *i = (uint32_t*)&n;
+    uint32_t b = i[0];
+    i[0] = htonl(i[1]);
+    i[1] = htonl(b);
+    return n;
+}
 
 Header headerLink(Header h)
 {
-    if (h == NULL) return NULL;
-
-    h->nrefs++;
-if (_hdr_debug)
-fprintf(stderr, "--> h  %p ++ %d at %s:%u\n", h, h->nrefs, __FILE__, __LINE__);
-
+    if (h != NULL)
+	h->nrefs++;
     return h;
 }
 
-Header headerUnlink(Header h)
+static Header headerUnlink(Header h)
 {
-    if (h == NULL) return NULL;
-if (_hdr_debug)
-fprintf(stderr, "--> h  %p -- %d at %s:%u\n", h, h->nrefs, __FILE__, __LINE__);
-    h->nrefs--;
+    if (h != NULL)
+	h->nrefs--;
     return NULL;
 }
 
@@ -135,13 +152,17 @@ Header headerFree(Header h)
     return h;
 }
 
-Header headerNew(void)
+static Header headerCreate(void *blob, int32_t indexLen)
 {
     Header h = xcalloc(1, sizeof(*h));
-
-    h->blob = NULL;
-    h->indexAlloced = INDEX_MALLOC_SIZE;
-    h->indexUsed = 0;
+    h->blob = blob;
+    if (blob) {
+	h->indexAlloced = indexLen + 1;
+	h->indexUsed = indexLen;
+    } else {
+	h->indexAlloced = INDEX_MALLOC_SIZE;
+	h->indexUsed = 0;
+    }
     h->instance = 0;
     h->flags |= HEADERFLAG_SORTED;
 
@@ -153,6 +174,11 @@ Header headerNew(void)
     return headerLink(h);
 }
 
+Header headerNew(void)
+{
+    return headerCreate(NULL, 0);
+}
+
 int headerVerifyInfo(int il, int dl, const void * pev, void * iv, int negate)
 {
     entryInfo pe = (entryInfo) pev;
@@ -219,10 +245,13 @@ static int offsetCmp(const void * avp, const void * bvp)
  */
 void headerUnsort(Header h)
 {
-    qsort(h->index, h->indexUsed, sizeof(*h->index), offsetCmp);
+    if (h->flags & HEADERFLAG_SORTED) {
+	qsort(h->index, h->indexUsed, sizeof(*h->index), offsetCmp);
+	h->flags &= ~HEADERFLAG_SORTED;
+    }
 }
 
-unsigned headerSizeof(Header h, enum hMagic magicp)
+unsigned headerSizeof(Header h, int magicp)
 {
     indexEntry entry;
     unsigned int size = 0;
@@ -244,7 +273,7 @@ unsigned headerSizeof(Header h, enum hMagic magicp)
     size += 2 * sizeof(int32_t);	/* count of index entries */
 
     for (i = 0, entry = h->index; i < h->indexUsed; i++, entry++) {
-	rpmTagType type;
+	rpm_tagtype_t type;
 
 	/* Regions go in as is ... */
         if (ENTRY_IS_REGION(entry)) {
@@ -283,7 +312,7 @@ unsigned headerSizeof(Header h, enum hMagic magicp)
  * @param pend		pointer to end of data (or NULL)
  * @return		no. bytes in data, -1 on failure
  */
-static int dataLength(rpmTagType type, rpm_constdata_t p, rpm_count_t count,
+static int dataLength(rpm_tagtype_t type, rpm_constdata_t p, rpm_count_t count,
 			 int onDisk, rpm_constdata_t pend)
 {
     const unsigned char * s = p;
@@ -360,7 +389,7 @@ static int regionSwab(indexEntry entry, int il, int dl,
 {
     for (; il > 0; il--, pe++) {
 	struct indexEntry_s ie;
-	rpmTagType type;
+	rpm_tagtype_t type;
 
 	ie.info.tag = ntohl(pe->tag);
 	ie.info.type = ntohl(pe->type);
@@ -452,7 +481,7 @@ static void * doHeaderUnload(Header h,
     int32_t il = 0;
     int32_t dl = 0;
     indexEntry entry; 
-    rpmTagType type;
+    rpm_tagtype_t type;
     int i;
     int drlen, ndribbles;
 
@@ -668,7 +697,6 @@ static void * doHeaderUnload(Header h,
     if (lengthPtr)
 	*lengthPtr = len;
 
-    h->flags &= ~HEADERFLAG_SORTED;
     headerSort(h);
 
     return (void *) ei;
@@ -693,7 +721,7 @@ void * headerUnload(Header h)
  * @return 		header entry
  */
 static
-indexEntry findEntry(Header h, rpmTag tag, rpmTagType type)
+indexEntry findEntry(Header h, rpmTagVal tag, rpm_tagtype_t type)
 {
     indexEntry entry;
     struct indexEntry_s key;
@@ -720,7 +748,7 @@ indexEntry findEntry(Header h, rpmTag tag, rpmTagType type)
     return NULL;
 }
 
-int headerDel(Header h, rpmTag tag)
+int headerDel(Header h, rpmTagVal tag)
 {
     indexEntry last = h->index + h->indexUsed;
     indexEntry entry, first;
@@ -781,21 +809,13 @@ Header headerLoad(void * uh)
     dataStart = (unsigned char *) (pe + il);
     dataEnd = dataStart + dl;
 
-    h = xcalloc(1, sizeof(*h));
-    h->blob = uh;
-    h->indexAlloced = il + 1;
-    h->indexUsed = il;
-    h->instance = 0;
-    h->index = xcalloc(h->indexAlloced, sizeof(*h->index));
-    h->flags |= HEADERFLAG_SORTED;
-    h->nrefs = 0;
-    h = headerLink(h);
+    h = headerCreate(uh, il);
 
     entry = h->index;
-    if (!(htonl(pe->tag) < HEADER_I18NTABLE)) {
+    if (!(htonl(pe->tag) < RPMTAG_HEADERI18NTABLE)) {
 	h->flags |= HEADERFLAG_LEGACY;
 	entry->info.type = REGION_TAG_TYPE;
-	entry->info.tag = HEADER_IMAGE;
+	entry->info.tag = RPMTAG_HEADERIMAGE;
 	entry->info.count = REGION_TAG_COUNT;
 	entry->info.offset = ((unsigned char *)pe - dataStart); /* negative offset */
 
@@ -837,7 +857,7 @@ Header headerLoad(void * uh)
 	    } else {
 		ril = il;
 		rdl = (ril * sizeof(struct entryInfo_s));
-		entry->info.tag = HEADER_IMAGE;
+		entry->info.tag = RPMTAG_HEADERIMAGE;
 	    }
 	}
 	entry->info.offset = -rdl;	/* negative offset */
@@ -897,7 +917,7 @@ errxit:
     return h;
 }
 
-Header headerReload(Header h, rpmTag tag)
+Header headerReload(Header h, rpmTagVal tag)
 {
     Header nh;
     size_t length;
@@ -912,7 +932,7 @@ Header headerReload(Header h, rpmTag tag)
 	return NULL;
     }
     if (ENTRY_IS_REGION(nh->index)) {
-	if (tag == HEADER_SIGNATURES || tag == HEADER_IMMUTABLE)
+	if (tag == RPMTAG_HEADERSIGNATURES || tag == RPMTAG_HEADERIMMUTABLE)
 	    nh->index[0].info.tag = tag;
     }
     return nh;
@@ -943,7 +963,7 @@ Header headerCopyLoad(const void * uh)
  * @param magicp	read (and verify) 8 bytes of (magic, 0)?
  * @return		header (or NULL on error)
  */
-Header headerRead(FD_t fd, enum hMagic magicp)
+Header headerRead(FD_t fd, int magicp)
 {
     int32_t block[4];
     int32_t reserved;
@@ -1000,7 +1020,7 @@ exit:
     return h;
 }
 
-int headerWrite(FD_t fd, Header h, enum hMagic magicp)
+int headerWrite(FD_t fd, Header h, int magicp)
 {
     ssize_t nb;
     size_t length;
@@ -1026,7 +1046,7 @@ exit:
     return (nb == length ? 0 : 1);
 }
 
-int headerIsEntry(Header h, rpmTag tag)
+int headerIsEntry(Header h, rpmTagVal tag)
 {
    		/* FIX: h modified by sort. */
     return (findEntry(h, tag, RPM_NULL_TYPE) ? 1 : 0);
@@ -1075,7 +1095,7 @@ static int copyTdEntry(const indexEntry entry, rpmtd td, headerGetFlags flags)
 
 	    rdl = entry->rdlen;
 	    count = 2 * sizeof(*ei) + (ril * sizeof(*pe)) + rdl;
-	    if (entry->info.tag == HEADER_IMAGE) {
+	    if (entry->info.tag == RPMTAG_HEADERIMAGE) {
 		ril -= 1;
 		pe += 1;
 	    } else {
@@ -1107,7 +1127,8 @@ static int copyTdEntry(const indexEntry entry, rpmtd td, headerGetFlags flags)
 	}
 	break;
     case RPM_STRING_TYPE:
-	if (count == 1) {
+	/* simple string, but fallthrough if its actually an array */
+	if (count == 1 && !argvArray) {
 	    td->data = allocMem ? xstrdup(entry->data) : entry->data;
 	    break;
 	}
@@ -1240,7 +1261,7 @@ static int copyI18NEntry(Header h, indexEntry entry, rpmtd td,
 	(lang = getenv("LANG")) == NULL)
 	    goto exit;
     
-    if ((table = findEntry(h, HEADER_I18NTABLE, RPM_STRING_ARRAY_TYPE)) == NULL)
+    if ((table = findEntry(h, RPMTAG_HEADERI18NTABLE, RPM_STRING_ARRAY_TYPE)) == NULL)
 	goto exit;
 
     for (l = lang; *l != '\0'; l = le) {
@@ -1320,7 +1341,7 @@ static int intGetTdEntry(Header h, rpmtd td, headerGetFlags flags)
     return ((rc == 1) ? 1 : 0);
 }
 
-int headerGet(Header h, rpmTag tag, rpmtd td, headerGetFlags flags)
+int headerGet(Header h, rpmTagVal tag, rpmtd td, headerGetFlags flags)
 {
     int rc;
     headerTagTagFunction tagfunc = intGetTdEntry;
@@ -1342,7 +1363,7 @@ int headerGet(Header h, rpmTag tag, rpmtd td, headerGetFlags flags)
 
 /**
  */
-static void copyData(rpmTagType type, rpm_data_t dstPtr, 
+static void copyData(rpm_tagtype_t type, rpm_data_t dstPtr, 
 		rpm_constdata_t srcPtr, rpm_count_t cnt, int dataLength)
 {
     switch (type) {
@@ -1376,7 +1397,7 @@ static void copyData(rpmTagType type, rpm_data_t dstPtr,
  * @return 		(malloc'ed) copy of entry data, NULL on error
  */
 static void *
-grabData(rpmTagType type, rpm_constdata_t p, rpm_count_t c, int * lengthPtr)
+grabData(rpm_tagtype_t type, rpm_constdata_t p, rpm_count_t c, int * lengthPtr)
 {
     rpm_data_t data = NULL;
     int length;
@@ -1486,101 +1507,7 @@ int headerPut(Header h, rpmtd td, headerPutFlags flags)
     return rc;
 }
 
-/*
- * Sanity check data types against tag table before putting. Assume
- * append on all array-types.
- */
-static int headerPutType(Header h, rpmTag tag, rpmTagType reqtype,
-			rpm_constdata_t data, rpm_count_t size)
-{
-    struct rpmtd_s td;
-    rpmTagType type = rpmTagGetType(tag);
-    headerPutFlags flags = HEADERPUT_APPEND; 
-    int valid = 1;
-
-    /* Basic sanity checks: type must match and there must be data to put */
-    if ((type & RPM_MASK_TYPE) != reqtype 
-	|| size < 1 || data == NULL || h == NULL) {
-	valid = 0;
-    }
-
-    /*
-     * Non-array types can't be appended to. Binary types use size
-     * for data length, for other non-array types size must be 1.
-     */
-    if ((type & RPM_MASK_RETURN_TYPE) != RPM_ARRAY_RETURN_TYPE) {
-	flags = HEADERPUT_DEFAULT;
-	if ((type & RPM_MASK_TYPE) != RPM_BIN_TYPE && size != 1) {
-	    valid = 0;
-	}
-    }
-
-    if (valid) {
-	rpmtdReset(&td);
-	td.tag = tag;
-	td.type = type & RPM_MASK_TYPE;
-	td.data = (void *) data;
-	td.count = size;
-
-	valid = headerPut(h, &td, flags);
-    }
-
-    return valid;
-}
-	
-int headerPutString(Header h, rpmTag tag, const char *val)
-{
-    rpmTagType type = rpmTagGetType(tag) & RPM_MASK_TYPE;
-    const void *sptr = NULL;
-
-    /* string arrays expect char **, arrange that */
-    if (type == RPM_STRING_ARRAY_TYPE || type == RPM_I18NSTRING_TYPE) {
-	sptr = &val;
-    } else if (type == RPM_STRING_TYPE) {
-	sptr = val;
-    } else {
-	return 0;
-    }
-
-    return headerPutType(h, tag, type, sptr, 1);
-}
-
-int headerPutStringArray(Header h, rpmTag tag, const char **array, rpm_count_t size)
-{
-    return headerPutType(h, tag, RPM_STRING_ARRAY_TYPE, array, size);
-}
-
-int headerPutChar(Header h, rpmTag tag, char *val, rpm_count_t size)
-{
-    return headerPutType(h, tag, RPM_CHAR_TYPE, val, size);
-}
-
-int headerPutUint8(Header h, rpmTag tag, uint8_t *val, rpm_count_t size)
-{
-    return headerPutType(h, tag, RPM_INT8_TYPE, val, size);
-}
-
-int headerPutUint16(Header h, rpmTag tag, uint16_t *val, rpm_count_t size)
-{
-    return headerPutType(h, tag, RPM_INT16_TYPE, val, size);
-}
-
-int headerPutUint32(Header h, rpmTag tag, uint32_t *val, rpm_count_t size)
-{
-    return headerPutType(h, tag, RPM_INT32_TYPE, val, size);
-}
-
-int headerPutUint64(Header h, rpmTag tag, uint64_t *val, rpm_count_t size)
-{
-    return headerPutType(h, tag, RPM_INT64_TYPE, val, size);
-}
-
-int headerPutBin(Header h, rpmTag tag, uint8_t *val, rpm_count_t size)
-{
-    return headerPutType(h, tag, RPM_BIN_TYPE, val, size);
-}
-
-int headerAddI18NString(Header h, rpmTag tag, const char * string,
+int headerAddI18NString(Header h, rpmTagVal tag, const char * string,
 		const char * lang)
 {
     indexEntry table, entry;
@@ -1590,7 +1517,7 @@ int headerAddI18NString(Header h, rpmTag tag, const char * string,
     rpm_count_t i, langNum;
     char * buf;
 
-    table = findEntry(h, HEADER_I18NTABLE, RPM_STRING_ARRAY_TYPE);
+    table = findEntry(h, RPMTAG_HEADERI18NTABLE, RPM_STRING_ARRAY_TYPE);
     entry = findEntry(h, tag, RPM_I18NSTRING_TYPE);
 
     if (!table && entry)
@@ -1608,13 +1535,13 @@ int headerAddI18NString(Header h, rpmTag tag, const char * string,
 	}
 	
 	rpmtdReset(&td);
-	td.tag = HEADER_I18NTABLE;
+	td.tag = RPMTAG_HEADERI18NTABLE;
 	td.type = RPM_STRING_ARRAY_TYPE;
 	td.data = (void *) charArray;
 	td.count = count;
 	if (!headerPut(h, &td, HEADERPUT_DEFAULT))
 	    return 0;
-	table = findEntry(h, HEADER_I18NTABLE, RPM_STRING_ARRAY_TYPE);
+	table = findEntry(h, RPMTAG_HEADERI18NTABLE, RPM_STRING_ARRAY_TYPE);
     }
 
     if (!table)
@@ -1803,7 +1730,7 @@ static indexEntry nextIndex(HeaderIterator hi)
     return entry;
 }
 
-rpmTag headerNextTag(HeaderIterator hi)
+rpmTagVal headerNextTag(HeaderIterator hi)
 {
     indexEntry entry = nextIndex(hi);
     return entry ? entry->info.tag : RPMTAG_NOT_FOUND;
@@ -1822,48 +1749,6 @@ int headerNext(HeaderIterator hi, rpmtd td)
     return ((rc == 1) ? 1 : 0);
 }
 
-/** \ingroup header
- * Duplicate a header.
- * @param h		header
- * @return		new header instance
- */
-Header headerCopy(Header h)
-{
-    Header nh = headerNew();
-    HeaderIterator hi;
-    struct rpmtd_s td;
-   
-    hi = headerInitIterator(h);
-    while (headerNext(hi, &td)) {
-	if (rpmtdCount(&td) > 0) {
-	    (void) headerPut(nh, &td, HEADERPUT_DEFAULT);
-	}
-	rpmtdFreeData(&td);
-    }
-    hi = headerFreeIterator(hi);
-
-    return headerReload(nh, HEADER_IMAGE);
-}
-
-void headerCopyTags(Header headerFrom, Header headerTo, 
-		    const rpmTag * tagstocopy)
-{
-    const rpmTag * p;
-    struct rpmtd_s td;
-
-    if (headerFrom == headerTo)
-	return;
-
-    for (p = tagstocopy; *p != 0; p++) {
-	if (headerIsEntry(headerTo, *p))
-	    continue;
-	if (!headerGet(headerFrom, *p, &td, (HEADERGET_MINMEM|HEADERGET_RAW)))
-	    continue;
-	(void) headerPut(headerTo, &td, HEADERPUT_DEFAULT);
-	rpmtdFreeData(&td);
-    }
-}
-
 unsigned int headerGetInstance(Header h)
 {
     return h ? h->instance : 0;
@@ -1874,44 +1759,3 @@ void headerSetInstance(Header h, unsigned int instance)
     h->instance = instance;
 }    
 
-char * headerGetAsString(Header h, rpmTag tag)
-{
-    char *res = NULL;
-    struct rpmtd_s td;
-
-    if (headerGet(h, tag, &td, HEADERGET_EXT)) {
-	if (rpmtdCount(&td) == 1) {
-	    res = rpmtdFormat(&td, RPMTD_FORMAT_STRING, NULL);
-	}
-	rpmtdFreeData(&td);
-    }
-    return res;
-}
-
-const char * headerGetString(Header h, rpmTag tag)
-{
-    const char *res = NULL;
-    struct rpmtd_s td;
-
-    if (headerGet(h, tag, &td, HEADERGET_MINMEM)) {
-	if (rpmtdCount(&td) == 1) {
-	    res = rpmtdGetString(&td);
-	}
-	rpmtdFreeData(&td);
-    }
-    return res;
-}
-
-uint64_t headerGetNumber(Header h, rpmTag tag)
-{
-    uint64_t res = 0;
-    struct rpmtd_s td;
-
-    if (headerGet(h, tag, &td, HEADERGET_EXT)) {
-	if (rpmtdCount(&td) == 1) {
-	    res = rpmtdGetNumber(&td);
-	}
-	rpmtdFreeData(&td);
-    }
-    return res;
-}
