@@ -15,7 +15,7 @@
 #include "lib/rpmlead.h"
 #include "lib/signature.h"
 #include "rpmio/digest.h"
-#include "rpmio/rpmio_internal.h"	/* fd*Digest(), fd stats */
+#include "rpmio/rpmio_internal.h"	/* fd digest bits */
 #include "lib/header_internal.h"	/* XXX headerCheck */
 
 #include "debug.h"
@@ -40,15 +40,9 @@ void headerMergeLegacySigs(Header h, const Header sigh)
 	case RPMSIGTAG_SIZE:
 	    td.tag = RPMTAG_SIGSIZE;
 	    break;
-	case RPMSIGTAG_LEMD5_1:
-	    td.tag = RPMTAG_SIGLEMD5_1;
-	    break;
 	case RPMSIGTAG_PGP:
 	    td.tag = RPMTAG_SIGPGP;
 	    break;
-	case RPMSIGTAG_LEMD5_2:
-	    td.tag = RPMTAG_SIGLEMD5_2;
-	    break;
 	case RPMSIGTAG_MD5:
 	    td.tag = RPMTAG_SIGMD5;
 	    break;
@@ -115,15 +109,9 @@ Header headerRegenSigHeader(const Header h, int noArchiveSize)
 	case RPMTAG_SIGSIZE:
 	    td.tag = RPMSIGTAG_SIZE;
 	    break;
-	case RPMTAG_SIGLEMD5_1:
-	    td.tag = RPMSIGTAG_LEMD5_1;
-	    break;
 	case RPMTAG_SIGPGP:
 	    td.tag = RPMSIGTAG_PGP;
 	    break;
-	case RPMTAG_SIGLEMD5_2:
-	    td.tag = RPMSIGTAG_LEMD5_2;
-	    break;
 	case RPMTAG_SIGMD5:
 	    td.tag = RPMSIGTAG_MD5;
 	    break;
@@ -191,20 +179,24 @@ static int stashKeyid(pgpDig dig)
     return 0;
 }
 
-/**
- * Check header consistency, performing headerGetEntry() the hard way.
- *
- * Sanity checks on the header are performed while looking for a
- * header-only digest or signature to verify the blob. If found,
- * the digest or signature is verified.
- *
- * @param ts		transaction set
- * @param uh		unloaded header blob
- * @param uc		no. of bytes in blob (or 0 to disable)
- * @retval *msg		signature verification msg
- * @return		RPMRC_OK/RPMRC_NOTFOUND/RPMRC_FAIL
- */
-rpmRC headerCheck(rpmts ts, const void * uh, size_t uc, char ** msg)
+/* Parse the parameters from the OpenPGP packets that will be needed. */
+static rpmRC parsePGP(rpmtd sigtd, const char *type, pgpDig dig)
+{
+    rpmRC rc = RPMRC_FAIL;
+    int debug = (_print_pkts & rpmIsDebug());
+    if ((pgpPrtPkts(sigtd->data, sigtd->count, dig, debug) == 0) &&
+	 (dig->signature.version == 3 || dig->signature.version == 4)) {
+	rc = RPMRC_OK;
+    } else {
+	rpmlog(RPMLOG_ERR,
+	    _("skipping %s with unverifiable V%u signature\n"), type,
+	    dig->signature.version);
+    }
+    return rc;
+}
+
+static rpmRC headerVerify(rpmKeyring keyring, rpmVSFlags vsflags,
+			  const void * uh, size_t uc, char ** msg)
 {
     pgpDig dig = NULL;
     char *buf = NULL;
@@ -218,7 +210,6 @@ rpmRC headerCheck(rpmts ts, const void * uh, size_t uc, char ** msg)
     struct indexEntry_s entry;
     struct entryInfo_s info;
     unsigned const char * b;
-    rpmVSFlags vsflags = rpmtsVSFlags(ts);
     size_t siglen = 0;
     size_t blen;
     size_t nb;
@@ -228,6 +219,7 @@ rpmRC headerCheck(rpmts ts, const void * uh, size_t uc, char ** msg)
     int xx;
     int i;
     struct rpmtd_s sigtd;
+    DIGEST_CTX ctx = NULL;
 
     /* Is the blob the right size? */
     if (uc > 0 && pvlen != uc) {
@@ -384,7 +376,6 @@ verifyinfo_exit:
     dig = pgpNewDig();
     if (dig == NULL)
 	goto verifyinfo_exit;
-    dig->nbytes = 0;
 
     sigtd.tag = info.tag;
     sigtd.type = info.type;
@@ -394,96 +385,43 @@ verifyinfo_exit:
 
     switch (info.tag) {
     case RPMTAG_RSAHEADER:
-	/* Parse the parameters from the OpenPGP packets that will be needed. */
-	xx = pgpPrtPkts(sigtd.data, sigtd.count, dig, (_print_pkts & rpmIsDebug()));
-	if (dig->signature.version != 3 && dig->signature.version != 4) {
-	    rpmlog(RPMLOG_ERR,
-		_("skipping header with unverifiable V%u signature\n"),
-		dig->signature.version);
-	    pgpFreeDig(dig);
-	    rc = RPMRC_FAIL;
-	    goto exit;
-	}
-
-	ildl[0] = htonl(ril);
-	ildl[1] = (regionEnd - dataStart);
-	ildl[1] = htonl(ildl[1]);
-
-	(void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_DIGEST), 0);
-	dig->hdrmd5ctx = rpmDigestInit(dig->signature.hash_algo, RPMDIGEST_NONE);
-
-	b = (unsigned char *) rpm_header_magic;
-	nb = sizeof(rpm_header_magic);
-        (void) rpmDigestUpdate(dig->hdrmd5ctx, b, nb);
-        dig->nbytes += nb;
-
-	b = (unsigned char *) ildl;
-	nb = sizeof(ildl);
-        (void) rpmDigestUpdate(dig->hdrmd5ctx, b, nb);
-        dig->nbytes += nb;
-
-	b = (unsigned char *) pe;
-	nb = (htonl(ildl[0]) * sizeof(*pe));
-        (void) rpmDigestUpdate(dig->hdrmd5ctx, b, nb);
-        dig->nbytes += nb;
-
-	b = (unsigned char *) dataStart;
-	nb = htonl(ildl[1]);
-        (void) rpmDigestUpdate(dig->hdrmd5ctx, b, nb);
-        dig->nbytes += nb;
-	(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_DIGEST), dig->nbytes);
-
-	break;
     case RPMTAG_DSAHEADER:
-	/* Parse the parameters from the OpenPGP packets that will be needed. */
-	xx = pgpPrtPkts(sigtd.data, sigtd.count, dig, (_print_pkts & rpmIsDebug()));
-	if (dig->signature.version != 3 && dig->signature.version != 4) {
-	    rpmlog(RPMLOG_ERR,
-		_("skipping header with unverifiable V%u signature\n"),
-		dig->signature.version);
+	if ((rc = parsePGP(&sigtd, "header", dig)) != RPMRC_OK) {
 	    pgpFreeDig(dig);
-	    rc = RPMRC_FAIL;
 	    goto exit;
 	}
-    case RPMTAG_SHA1HEADER:
+	/* fallthrough */
+    case RPMTAG_SHA1HEADER: {
+	int hashalgo = (info.tag == RPMTAG_SHA1HEADER) ?
+			PGPHASHALGO_SHA1 : dig->signature.hash_algo;
 	ildl[0] = htonl(ril);
 	ildl[1] = (regionEnd - dataStart);
 	ildl[1] = htonl(ildl[1]);
 
-	(void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_DIGEST), 0);
-	dig->hdrsha1ctx = rpmDigestInit(PGPHASHALGO_SHA1, RPMDIGEST_NONE);
+	ctx = rpmDigestInit(hashalgo, RPMDIGEST_NONE);
 
 	b = (unsigned char *) rpm_header_magic;
 	nb = sizeof(rpm_header_magic);
-        (void) rpmDigestUpdate(dig->hdrsha1ctx, b, nb);
-        dig->nbytes += nb;
+        (void) rpmDigestUpdate(ctx, b, nb);
 
 	b = (unsigned char *) ildl;
 	nb = sizeof(ildl);
-        (void) rpmDigestUpdate(dig->hdrsha1ctx, b, nb);
-        dig->nbytes += nb;
+        (void) rpmDigestUpdate(ctx, b, nb);
 
 	b = (unsigned char *) pe;
 	nb = (htonl(ildl[0]) * sizeof(*pe));
-        (void) rpmDigestUpdate(dig->hdrsha1ctx, b, nb);
-        dig->nbytes += nb;
+        (void) rpmDigestUpdate(ctx, b, nb);
 
 	b = (unsigned char *) dataStart;
 	nb = htonl(ildl[1]);
-        (void) rpmDigestUpdate(dig->hdrsha1ctx, b, nb);
-        dig->nbytes += nb;
-	(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_DIGEST), dig->nbytes);
-
-	break;
+        (void) rpmDigestUpdate(ctx, b, nb);
+	} break;
     default:
 	sigtd.data = _free(sigtd.data); /* Hmm...? */
 	break;
     }
 
-    {	rpmKeyring keyring = rpmtsGetKeyring(ts, 1);
-    	rc = rpmVerifySignature(keyring, &sigtd, dig, &buf);
-	rpmKeyringFree(keyring);
-    }
+    rc = rpmVerifySignature(keyring, &sigtd, dig, ctx, &buf);
 
     if (msg) 
 	*msg = buf;
@@ -492,10 +430,26 @@ verifyinfo_exit:
 
     rpmtdFreeData(&sigtd);
     pgpFreeDig(dig);
+    rpmDigestFinal(ctx, NULL, NULL, 0);
     return rc;
 }
 
-rpmRC rpmReadHeader(rpmts ts, FD_t fd, Header *hdrp, char ** msg)
+rpmRC headerCheck(rpmts ts, const void * uh, size_t uc, char ** msg)
+{
+    rpmRC rc;
+    rpmVSFlags vsflags = rpmtsVSFlags(ts);
+    rpmKeyring keyring = rpmtsGetKeyring(ts, 1);
+
+    rpmswEnter(rpmtsOp(ts, RPMTS_OP_DIGEST), 0);
+    rc = headerVerify(keyring, vsflags, uh, uc, msg);
+    rpmswExit(rpmtsOp(ts, RPMTS_OP_DIGEST), uc);
+    rpmKeyringFree(keyring);
+
+    return rc;
+}
+
+static rpmRC rpmpkgReadHeader(rpmKeyring keyring, rpmVSFlags vsflags, 
+		       FD_t fd, Header *hdrp, char ** msg)
 {
     char *buf = NULL;
     int32_t block[4];
@@ -546,7 +500,7 @@ rpmRC rpmReadHeader(rpmts ts, FD_t fd, Header *hdrp, char ** msg)
     }
 
     /* Sanity check header tags */
-    rc = headerCheck(ts, ei, uc, msg);
+    rc = headerVerify(keyring, vsflags, ei, uc, msg);
     if (rc != RPMRC_OK)
 	goto exit;
 
@@ -557,7 +511,6 @@ rpmRC rpmReadHeader(rpmts ts, FD_t fd, Header *hdrp, char ** msg)
 	rc = RPMRC_FAIL;
         goto exit;
     }
-    h->flags |= HEADERFLAG_ALLOCATED;
     ei = NULL;	/* XXX will be freed with header */
     
 exit:
@@ -575,24 +528,34 @@ exit:
     return rc;
 }
 
-rpmRC rpmReadPackageFile(rpmts ts, FD_t fd, const char * fn, Header * hdrp)
+rpmRC rpmReadHeader(rpmts ts, FD_t fd, Header *hdrp, char ** msg)
+{
+    rpmRC rc;
+    rpmKeyring keyring = rpmtsGetKeyring(ts, 1);
+    rpmVSFlags vsflags = rpmtsVSFlags(ts);
+
+    rc = rpmpkgReadHeader(keyring, vsflags, fd, hdrp, msg);
+
+    rpmKeyringFree(keyring);
+    return rc;
+}
+
+static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags, 
+			FD_t fd, const char * fn, Header * hdrp)
 {
     pgpDig dig = NULL;
     char buf[8*BUFSIZ];
     ssize_t count;
     rpmlead l = NULL;
     Header sigh = NULL;
-    rpmSigTag sigtag;
+    rpmTagVal sigtag;
     struct rpmtd_s sigtd;
-    rpmtsOpX opx;
-    size_t nb;
     Header h = NULL;
     char * msg;
-    rpmVSFlags vsflags;
     rpmRC rc = RPMRC_FAIL;	/* assume failure */
-    int xx;
     int leadtype = -1;
     headerGetFlags hgeflags = HEADERGET_DEFAULT;
+    DIGEST_CTX ctx = NULL;
 
     if (hdrp) *hdrp = NULL;
 
@@ -631,7 +594,8 @@ rpmRC rpmReadPackageFile(rpmts ts, FD_t fd, const char * fn, Header * hdrp)
     }
     msg = _free(msg);
 
-#define	_chk(_mask)	(sigtag == 0 && !(vsflags & (_mask)))
+#define	_chk(_mask, _tag) \
+	(sigtag == 0 && !(vsflags & (_mask)) && headerIsEntry(sigh, (_tag)))
 
     /*
      * Figger the most effective available signature.
@@ -640,52 +604,28 @@ rpmRC rpmReadPackageFile(rpmts ts, FD_t fd, const char * fn, Header * hdrp)
      * Note that NEEDPAYLOAD prevents header+payload signatures and digests.
      */
     sigtag = 0;
-    opx = 0;
-    vsflags = rpmtsVSFlags(ts);
-    if (_chk(RPMVSF_NODSAHEADER) && headerIsEntry(sigh, RPMSIGTAG_DSA)) {
+    if (_chk(RPMVSF_NODSAHEADER, RPMSIGTAG_DSA)) {
 	sigtag = RPMSIGTAG_DSA;
-    } else
-    if (_chk(RPMVSF_NORSAHEADER) && headerIsEntry(sigh, RPMSIGTAG_RSA)) {
+    } else if (_chk(RPMVSF_NORSAHEADER, RPMSIGTAG_RSA)) {
 	sigtag = RPMSIGTAG_RSA;
-    } else
-    if (_chk(RPMVSF_NODSA|RPMVSF_NEEDPAYLOAD) &&
-	headerIsEntry(sigh, RPMSIGTAG_GPG))
-    {
+    } else if (_chk(RPMVSF_NODSA|RPMVSF_NEEDPAYLOAD, RPMSIGTAG_GPG)) {
 	sigtag = RPMSIGTAG_GPG;
 	fdInitDigest(fd, PGPHASHALGO_SHA1, 0);
-	opx = RPMTS_OP_SIGNATURE;
-    } else
-    if (_chk(RPMVSF_NORSA|RPMVSF_NEEDPAYLOAD) &&
-	headerIsEntry(sigh, RPMSIGTAG_PGP))
-    {
+    } else if (_chk(RPMVSF_NORSA|RPMVSF_NEEDPAYLOAD, RPMSIGTAG_PGP)) {
 	sigtag = RPMSIGTAG_PGP;
 	fdInitDigest(fd, PGPHASHALGO_MD5, 0);
-	opx = RPMTS_OP_SIGNATURE;
-    } else
-    if (_chk(RPMVSF_NOSHA1HEADER) && headerIsEntry(sigh, RPMSIGTAG_SHA1)) {
+    } else if (_chk(RPMVSF_NOSHA1HEADER, RPMSIGTAG_SHA1)) {
 	sigtag = RPMSIGTAG_SHA1;
-    } else
-    if (_chk(RPMVSF_NOMD5|RPMVSF_NEEDPAYLOAD) &&
-	headerIsEntry(sigh, RPMSIGTAG_MD5))
-    {
+    } else if (_chk(RPMVSF_NOMD5|RPMVSF_NEEDPAYLOAD, RPMSIGTAG_MD5)) {
 	sigtag = RPMSIGTAG_MD5;
 	fdInitDigest(fd, PGPHASHALGO_MD5, 0);
-	opx = RPMTS_OP_DIGEST;
     }
 
     /* Read the metadata, computing digest(s) on the fly. */
     h = NULL;
     msg = NULL;
 
-    /* XXX stats will include header i/o and setup overhead. */
-    /* XXX repackaged packages have appended tags, legacy dig/sig check fails */
-    if (opx > 0)
-	(void) rpmswEnter(rpmtsOp(ts, opx), 0);
-    nb = -fd->stats->ops[FDSTAT_READ].bytes;
-    rc = rpmReadHeader(ts, fd, &h, &msg);
-    nb += fd->stats->ops[FDSTAT_READ].bytes;
-    if (opx > 0)
-	(void) rpmswExit(rpmtsOp(ts, opx), nb);
+    rc = rpmpkgReadHeader(keyring, vsflags, fd, &h, &msg);
 
     if (rc != RPMRC_OK || h == NULL) {
 	rpmlog(RPMLOG_ERR, _("%s: headerRead failed: %s"), fn,
@@ -706,7 +646,6 @@ rpmRC rpmReadPackageFile(rpmts ts, FD_t fd, const char * fn, Header * hdrp)
 	rc = RPMRC_FAIL;
 	goto exit;
     }
-    dig->nbytes = 0;
 
     /* Retrieve the tag parameters from the signature header. */
     if (!headerGet(sigh, sigtag, &sigtd, hgeflags)) {
@@ -716,76 +655,33 @@ rpmRC rpmReadPackageFile(rpmts ts, FD_t fd, const char * fn, Header * hdrp)
 
     switch (sigtag) {
     case RPMSIGTAG_RSA:
-	/* Parse the parameters from the OpenPGP packets that will be needed. */
-	xx = pgpPrtPkts(sigtd.data, sigtd.count, dig, (_print_pkts & rpmIsDebug()));
-	if (dig->signature.version != 3 && dig->signature.version != 4) {
-	    rpmlog(RPMLOG_ERR,
-		_("skipping package %s with unverifiable V%u signature\n"),
-		fn, dig->signature.version);
-	    rc = RPMRC_FAIL;
-	    goto exit;
-	}
-    {	struct rpmtd_s utd;
-
-	if (!headerGet(h, RPMTAG_HEADERIMMUTABLE, &utd, hgeflags))
-	    break;
-	(void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_DIGEST), 0);
-	dig->hdrmd5ctx = rpmDigestInit(dig->signature.hash_algo, RPMDIGEST_NONE);
-	(void) rpmDigestUpdate(dig->hdrmd5ctx, rpm_header_magic, sizeof(rpm_header_magic));
-	dig->nbytes += sizeof(rpm_header_magic);
-	(void) rpmDigestUpdate(dig->hdrmd5ctx, utd.data, utd.count);
-	dig->nbytes += utd.count;
-	(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_DIGEST), dig->nbytes);
-	rpmtsOp(ts, RPMTS_OP_DIGEST)->count--;	/* XXX one too many */
-	rpmtdFreeData(&utd);
-    }	break;
     case RPMSIGTAG_DSA:
-	/* Parse the parameters from the OpenPGP packets that will be needed. */
-	xx = pgpPrtPkts(sigtd.data, sigtd.count, dig, (_print_pkts & rpmIsDebug()));
-	if (dig->signature.version != 3 && dig->signature.version != 4) {
-	    rpmlog(RPMLOG_ERR,
-		_("skipping package %s with unverifiable V%u signature\n"), 
-		fn, dig->signature.version);
-	    rc = RPMRC_FAIL;
+	if ((rc = parsePGP(&sigtd, "package", dig)) != RPMRC_OK) {
 	    goto exit;
 	}
+	/* fallthrough */
     case RPMSIGTAG_SHA1:
     {	struct rpmtd_s utd;
+	int hashalgo = (sigtag == RPMSIGTAG_SHA1) ?
+			PGPHASHALGO_SHA1 : dig->signature.hash_algo;
 
 	if (!headerGet(h, RPMTAG_HEADERIMMUTABLE, &utd, hgeflags))
 	    break;
-	(void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_DIGEST), 0);
-	dig->hdrsha1ctx = rpmDigestInit(PGPHASHALGO_SHA1, RPMDIGEST_NONE);
-	(void) rpmDigestUpdate(dig->hdrsha1ctx, rpm_header_magic, sizeof(rpm_header_magic));
-	dig->nbytes += sizeof(rpm_header_magic);
-	(void) rpmDigestUpdate(dig->hdrsha1ctx, utd.data, utd.count);
-	dig->nbytes += utd.count;
-	(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_DIGEST), dig->nbytes);
-	if (sigtag == RPMSIGTAG_SHA1)
-	    rpmtsOp(ts, RPMTS_OP_DIGEST)->count--;	/* XXX one too many */
+	ctx = rpmDigestInit(hashalgo, RPMDIGEST_NONE);
+	(void) rpmDigestUpdate(ctx, rpm_header_magic, sizeof(rpm_header_magic));
+	(void) rpmDigestUpdate(ctx, utd.data, utd.count);
 	rpmtdFreeData(&utd);
     }	break;
     case RPMSIGTAG_GPG:
     case RPMSIGTAG_PGP5:	/* XXX legacy */
     case RPMSIGTAG_PGP:
-	/* Parse the parameters from the OpenPGP packets that will be needed. */
-	xx = pgpPrtPkts(sigtd.data, sigtd.count, dig, (_print_pkts & rpmIsDebug()));
-
-	if (dig->signature.version != 3 && dig->signature.version != 4) {
-	    rpmlog(RPMLOG_ERR,
-		_("skipping package %s with unverifiable V%u signature\n"),
-		fn, dig->signature.version);
-	    rc = RPMRC_FAIL;
+	if ((rc = parsePGP(&sigtd, "package", dig)) != RPMRC_OK) {
 	    goto exit;
 	}
+	/* fallthrough */
     case RPMSIGTAG_MD5:
 	/* Legacy signatures need the compressed payload in the digest too. */
-	(void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_DIGEST), 0);
-	while ((count = Fread(buf, sizeof(buf[0]), sizeof(buf), fd)) > 0)
-	    dig->nbytes += count;
-	(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_DIGEST), dig->nbytes);
-	rpmtsOp(ts, RPMTS_OP_DIGEST)->count--;	/* XXX one too many */
-	dig->nbytes += nb;	/* XXX include size of header blob. */
+	while ((count = Fread(buf, sizeof(buf[0]), sizeof(buf), fd)) > 0) {}
 	if (count < 0) {
 	    rpmlog(RPMLOG_ERR, _("%s: Fread failed: %s\n"),
 					fn, Fstrerror(fd));
@@ -793,25 +689,15 @@ rpmRC rpmReadPackageFile(rpmts ts, FD_t fd, const char * fn, Header * hdrp)
 	    goto exit;
 	}
 
-	fdStealDigest(fd, dig);
+	ctx = rpmDigestBundleDupCtx(fdGetBundle(fd), (sigtag == RPMSIGTAG_MD5) ?
+				    PGPHASHALGO_MD5 : dig->signature.hash_algo);
 	break;
-    /* shut up gcc */
-    case RPMSIGTAG_LEMD5_1:
-    case RPMSIGTAG_LEMD5_2:
-    case RPMSIGTAG_BADSHA1_1:
-    case RPMSIGTAG_BADSHA1_2:
-    case RPMSIGTAG_SIZE:
-    case RPMSIGTAG_LONGSIZE:
-    case RPMSIGTAG_PAYLOADSIZE:
-    case RPMSIGTAG_LONGARCHIVESIZE:
+    default:
 	break;
     }
 
     /** @todo Implement disable/enable/warn/error/anal policy. */
-    {	rpmKeyring keyring = rpmtsGetKeyring(ts, 1);
-    	rc = rpmVerifySignature(keyring, &sigtd, dig, &msg);
-	rpmKeyringFree(keyring);
-    }
+    rc = rpmVerifySignature(keyring, &sigtd, dig, ctx, &msg);
 	
     switch (rc) {
     case RPMRC_OK:		/* Signature is OK. */
@@ -864,38 +750,46 @@ exit:
 	*hdrp = headerLink(h);
     }
     rpmtdFreeData(&sigtd);
+    rpmDigestFinal(ctx, NULL, NULL, 0);
     h = headerFree(h);
     pgpFreeDig(dig);
     sigh = rpmFreeSignature(sigh);
     return rc;
 }
 
+rpmRC rpmReadPackageFile(rpmts ts, FD_t fd, const char * fn, Header * hdrp)
+{
+    rpmRC rc;
+    rpmVSFlags vsflags = rpmtsVSFlags(ts);
+    rpmKeyring keyring = 0;
+
+    if ((vsflags & _RPMVSF_NOSIGNATURES) != _RPMVSF_NOSIGNATURES)
+	keyring = rpmtsGetKeyring(ts, 1);
+
+    rc = rpmpkgRead(keyring, vsflags, fd, fn, hdrp);
+
+    if (keyring)
+	rpmKeyringFree(keyring);
+    return rc;
+}
+
 /**
  * Check for supported payload format in header.
  * @param h		header to check
  * @return		RPMRC_OK if supported, RPMRC_FAIL otherwise
  */
 rpmRC headerCheckPayloadFormat(Header h) {
-    rpmRC rc = RPMRC_FAIL;
-    const char *payloadfmt = NULL;
-    struct rpmtd_s payload;
-
-    if (headerGet(h, RPMTAG_PAYLOADFORMAT, &payload, HEADERGET_DEFAULT)) {
-	payloadfmt = rpmtdGetString(&payload);
-	rpmtdFreeData(&payload);
-    }
+    rpmRC rc = RPMRC_OK;
+    const char *payloadfmt = headerGetString(h, RPMTAG_PAYLOADFORMAT);
     /* 
      * XXX Ugh, rpm 3.x packages don't have payload format tag. Instead
      * of blinly allowing, should check somehow (HDRID existence or... ?)
      */
-    if (!payloadfmt)
-	return RPMRC_OK;
+    if (!payloadfmt) return rc;
 
-    if (payloadfmt && strncmp(payloadfmt, "cpio", strlen("cpio")) == 0) {
-	rc = RPMRC_OK;
-    } else {
-        char *nevra = headerGetNEVRA(h, NULL);
-        if (payloadfmt && strncmp(payloadfmt, "drpm", strlen("drpm")) == 0) {
+    if (!rstreq(payloadfmt, "cpio")) {
+        char *nevra = headerGetAsString(h, RPMTAG_NEVRA);
+        if (payloadfmt && rstreq(payloadfmt, "drpm")) {
             rpmlog(RPMLOG_ERR,
                      _("%s is a Delta RPM and cannot be directly installed\n"),
                      nevra);
@@ -905,6 +799,7 @@ rpmRC headerCheckPayloadFormat(Header h) {
                      payloadfmt ? payloadfmt : "none", nevra);
         } 
         nevra = _free(nevra);
+	rc = RPMRC_FAIL;
     }
     return rc;
 }
