@@ -11,17 +11,10 @@
 
 #include "debug.h"
 
-/**
- * Enable noisy range comparison debugging message?
- */
-static int _noisy_range_comparison_debug_message = 0;
-
 int _rpmds_debug = 0;
 
 int _rpmds_nopromote = 1;
 
-int _rpmds_unspecified_epoch_noise = 0;
-
 /**
  * A package dependency set.
  */
@@ -32,10 +25,9 @@ struct rpmds_s {
     const char ** EVR;		/*!< Epoch-Version-Release. */
     rpmsenseFlags * Flags;	/*!< Bit(s) identifying context/comparison. */
     rpm_color_t * Color;	/*!< Bit(s) calculated from file color(s). */
-    int32_t * Refs;		/*!< No. of file refs. */
-    time_t BT;			/*!< Package build time tie breaker. */
-    rpmTag tagN;		/*!< Header tag. */
+    rpmTagVal tagN;		/*!< Header tag. */
     int32_t Count;		/*!< No. of elements */
+    unsigned int instance;	/*!< From rpmdb instance? */
     int i;			/*!< Element index. */
     unsigned l;			/*!< Low element (bsearch). */
     unsigned u;			/*!< High element (bsearch). */
@@ -45,13 +37,13 @@ struct rpmds_s {
 
 static const char ** rpmdsDupArgv(const char ** argv, int argc);
 
-static int dsType(rpmTag tag, 
-		  const char ** Type, rpmTag * tagEVR, rpmTag * tagF)
+static int dsType(rpmTagVal tag, 
+		  const char ** Type, rpmTagVal * tagEVR, rpmTagVal * tagF)
 {
     int rc = 0;
     const char *t = NULL;
-    rpmTag evr = RPMTAG_NOT_FOUND;
-    rpmTag f = RPMTAG_NOT_FOUND;
+    rpmTagVal evr = RPMTAG_NOT_FOUND;
+    rpmTagVal f = RPMTAG_NOT_FOUND;
 
     if (tag == RPMTAG_PROVIDENAME) {
 	t = "Provides";
@@ -69,6 +61,10 @@ static int dsType(rpmTag tag,
 	t = "Obsoletes";
 	evr = RPMTAG_OBSOLETEVERSION;
 	f = RPMTAG_OBSOLETEFLAGS;
+    } else if (tag == RPMTAG_ORDERNAME) {
+	t = "Order";
+	evr = RPMTAG_ORDERVERSION;
+	f = RPMTAG_ORDERFLAGS;
     } else if (tag == RPMTAG_TRIGGERNAME) {
 	t = "Trigger";
 	evr = RPMTAG_TRIGGERVERSION;
@@ -82,38 +78,29 @@ static int dsType(rpmTag tag,
     return rc;
 }    
 
-rpmds rpmdsUnlink(rpmds ds, const char * msg)
+static rpmds rpmdsUnlink(rpmds ds)
 {
-    if (ds == NULL) return NULL;
-if (_rpmds_debug && msg != NULL)
-fprintf(stderr, "--> ds %p -- %d %s\n", ds, ds->nrefs, msg);
-    ds->nrefs--;
+    if (ds)
+	ds->nrefs--;
     return NULL;
 }
 
-rpmds rpmdsLink(rpmds ds, const char * msg)
+rpmds rpmdsLink(rpmds ds)
 {
-    if (ds == NULL) return NULL;
-    ds->nrefs++;
-
-if (_rpmds_debug && msg != NULL)
-fprintf(stderr, "--> ds %p ++ %d %s\n", ds, ds->nrefs, msg);
-
+    if (ds)
+	ds->nrefs++;
     return ds;
 }
 
 rpmds rpmdsFree(rpmds ds)
 {
-    rpmTag tagEVR, tagF;
+    rpmTagVal tagEVR, tagF;
 
     if (ds == NULL)
 	return NULL;
 
     if (ds->nrefs > 1)
-	return rpmdsUnlink(ds, ds->Type);
-
-if (_rpmds_debug < 0)
-fprintf(stderr, "*** ds %p\t%s[%d]\n", ds, ds->Type, ds->Count);
+	return rpmdsUnlink(ds);
 
     if (dsType(ds->tagN, NULL, &tagEVR, &tagF))
 	return NULL;
@@ -126,17 +113,16 @@ fprintf(stderr, "*** ds %p\t%s[%d]\n", ds, ds->Type, ds->Count);
 
     ds->DNEVR = _free(ds->DNEVR);
     ds->Color = _free(ds->Color);
-    ds->Refs = _free(ds->Refs);
 
-    (void) rpmdsUnlink(ds, ds->Type);
+    (void) rpmdsUnlink(ds);
     memset(ds, 0, sizeof(*ds));		/* XXX trash and burn */
     ds = _free(ds);
     return NULL;
 }
 
-rpmds rpmdsNew(Header h, rpmTag tagN, int flags)
+rpmds rpmdsNew(Header h, rpmTagVal tagN, int flags)
 {
-    rpmTag tagEVR, tagF;
+    rpmTagVal tagEVR, tagF;
     rpmds ds = NULL;
     const char * Type;
     struct rpmtd_s names;
@@ -146,8 +132,7 @@ rpmds rpmdsNew(Header h, rpmTag tagN, int flags)
 	goto exit;
 
     if (headerGet(h, tagN, &names, hgflags) && rpmtdCount(&names) > 0) {
-	struct rpmtd_s evr, flags, buildtime; 
-	rpm_time_t * BTp;
+	struct rpmtd_s evr, flags; 
 
 	ds = xcalloc(1, sizeof(*ds));
 	ds->Type = Type;
@@ -157,28 +142,25 @@ rpmds rpmdsNew(Header h, rpmTag tagN, int flags)
 	ds->N = names.data;
 	ds->Count = rpmtdCount(&names);
 	ds->nopromote = _rpmds_nopromote;
+	ds->instance = headerGetInstance(h);
 
 	headerGet(h, tagEVR, &evr, hgflags);
 	ds->EVR = evr.data;
 	headerGet(h, tagF, &flags, hgflags);
 	ds->Flags = flags.data;
+	/* ensure rpmlib() requires always have RPMSENSE_RPMLIB flag set */
+	if (tagN == RPMTAG_REQUIRENAME && ds->Flags) {
+	    for (int i = 0; i < ds->Count; i++) {
+		if (!(ds->Flags[i] & RPMSENSE_RPMLIB) &&
+			rstreqn(ds->N[i], "rpmlib(", sizeof("rpmlib(")-1))
+		    ds->Flags[i] |= RPMSENSE_RPMLIB;
+	    }
+	}
 
-	headerGet(h, RPMTAG_BUILDTIME, &buildtime, HEADERGET_MINMEM);
-	BTp = rpmtdGetUint32(&buildtime);
-	ds->BT = BTp ? *BTp : 0;
-	rpmtdFreeData(&buildtime);
-	ds->Color = xcalloc(ds->Count, sizeof(*ds->Color));
-	ds->Refs = xcalloc(ds->Count, sizeof(*ds->Refs));
-
-if (_rpmds_debug < 0)
-fprintf(stderr, "*** ds %p\t%s[%d]\n", ds, ds->Type, ds->Count);
-
+	ds = rpmdsLink(ds);
     }
 
 exit:
-    /* FIX: ds->Flags may be NULL */
-    ds = rpmdsLink(ds, (ds ? ds->Type : RPMDBG()));
-
     return ds;
 }
 
@@ -226,42 +208,8 @@ char * rpmdsNewDNEVR(const char * dspfx, const rpmds ds)
     return tbuf;
 }
 
-rpmds rpmdsThis(Header h, rpmTag tagN, rpmsenseFlags Flags)
-{
-    rpmds ds = NULL;
-    const char * Type;
-    const char * n;
-    char *evr;
-
-    if (dsType(tagN, &Type, NULL, NULL))
-	goto exit;
-
-    evr = headerGetEVR(h, &n);
-
-    ds = xcalloc(1, sizeof(*ds));
-    ds->Type = Type;
-    ds->tagN = tagN;
-    ds->Count = 1;
-    ds->nopromote = _rpmds_nopromote;
-    ds->N = rpmdsDupArgv(&n, 1);
-    ds->EVR = rpmdsDupArgv((const char **)&evr, 1);
-    free(evr);
-
-    ds->Flags = xmalloc(sizeof(*ds->Flags));	ds->Flags[0] = Flags;
-    ds->i = 0;
-    {	char pre[2];
-	pre[0] = ds->Type[0];
-	pre[1] = '\0';
-	/* LCL: ds->Type may be NULL ??? */
-	ds->DNEVR = rpmdsNewDNEVR(pre, ds);
-    }
-
-   
-exit:
-    return rpmdsLink(ds, (ds ? ds->Type : RPMDBG()));
-}
-
-rpmds rpmdsSingle(rpmTag tagN, const char * N, const char * EVR, rpmsenseFlags Flags)
+static rpmds singleDS(rpmTagVal tagN, const char * N, const char * EVR,
+		      rpmsenseFlags Flags, unsigned int instance)
 {
     rpmds ds = NULL;
     const char * Type;
@@ -272,11 +220,9 @@ rpmds rpmdsSingle(rpmTag tagN, const char * N, const char * EVR, rpmsenseFlags F
     ds = xcalloc(1, sizeof(*ds));
     ds->Type = Type;
     ds->tagN = tagN;
-    {	time_t now = time(NULL);
-	ds->BT = now;
-    }
     ds->Count = 1;
     ds->nopromote = _rpmds_nopromote;
+    ds->instance = instance;
 
     ds->N = rpmdsDupArgv(&N, 1);
     ds->EVR = rpmdsDupArgv(&EVR, 1);
@@ -284,14 +230,33 @@ rpmds rpmdsSingle(rpmTag tagN, const char * N, const char * EVR, rpmsenseFlags F
     ds->Flags = xmalloc(sizeof(*ds->Flags));
     ds->Flags[0] = Flags;
     ds->i = 0;
-    {	char t[2];
-	t[0] = ds->Type[0];
-	t[1] = '\0';
-	ds->DNEVR = rpmdsNewDNEVR(t, ds);
-    }
 
 exit:
-    return rpmdsLink(ds, (ds ? ds->Type : NULL));
+    return rpmdsLink(ds);
+}
+
+rpmds rpmdsThis(Header h, rpmTagVal tagN, rpmsenseFlags Flags)
+{
+    char *evr = headerGetAsString(h, RPMTAG_EVR);
+    rpmds ds = singleDS(tagN, headerGetString(h, RPMTAG_NAME),
+			evr, Flags, headerGetInstance(h));
+    free(evr);
+    return ds;
+}
+
+rpmds rpmdsSingle(rpmTagVal tagN, const char * N, const char * EVR, rpmsenseFlags Flags)
+{
+    return singleDS(tagN, N, EVR, Flags, 0);
+}
+
+rpmds rpmdsCurrent(rpmds ds)
+{
+    rpmds cds = NULL;
+    if (ds != NULL && ds->i >= 0 && ds->i < ds->Count) {
+	cds = singleDS(ds->tagN, ds->N[ds->i], ds->EVR[ds->i],
+		       ds->Flags[ds->i], ds->instance);
+    }
+    return cds;
 }
 
 int rpmdsCount(const rpmds ds)
@@ -311,6 +276,7 @@ int rpmdsSetIx(rpmds ds, int ix)
     if (ds != NULL) {
 	i = ds->i;
 	ds->i = ix;
+	ds->DNEVR = _free(ds->DNEVR);
     }
     return i;
 }
@@ -320,8 +286,11 @@ const char * rpmdsDNEVR(const rpmds ds)
     const char * DNEVR = NULL;
 
     if (ds != NULL && ds->i >= 0 && ds->i < ds->Count) {
-	if (ds->DNEVR != NULL)
-	    DNEVR = ds->DNEVR;
+	if (ds->DNEVR == NULL) {
+	    char t[2] = { ds->Type[0], '\0' };
+	    ds->DNEVR = rpmdsNewDNEVR(t, ds);
+	}
+	DNEVR = ds->DNEVR;
     }
     return DNEVR;
 }
@@ -359,31 +328,18 @@ rpmsenseFlags rpmdsFlags(const rpmds ds)
     return Flags;
 }
 
-rpmTag rpmdsTagN(const rpmds ds)
+rpmTagVal rpmdsTagN(const rpmds ds)
 {
-    rpmTag tagN = 0;
+    rpmTagVal tagN = RPMTAG_NOT_FOUND;
 
     if (ds != NULL)
 	tagN = ds->tagN;
     return tagN;
 }
 
-time_t rpmdsBT(const rpmds ds)
+unsigned int rpmdsInstance(rpmds ds)
 {
-    time_t BT = 0;
-    if (ds != NULL && ds->BT > 0)
-	BT = ds->BT;
-    return BT;
-}
-
-time_t rpmdsSetBT(const rpmds ds, time_t BT)
-{
-    time_t oBT = 0;
-    if (ds != NULL) {
-	oBT = ds->BT;
-	ds->BT = BT;
-    }
-    return oBT;
+    return (ds != NULL) ? ds->instance : 0;
 }
 
 int rpmdsNoPromote(const rpmds ds)
@@ -422,49 +378,28 @@ rpm_color_t rpmdsSetColor(const rpmds ds, rpm_color_t color)
     rpm_color_t ocolor = 0;
 
     if (ds != NULL && ds->i >= 0 && ds->i < ds->Count) {
-	if (ds->Color != NULL) {
-	    ocolor = ds->Color[ds->i];
-	    ds->Color[ds->i] = color;
+	if (ds->Color == NULL) {
+	    ds->Color = xcalloc(ds->Count, sizeof(*ds->Color));
 	}
+	ocolor = ds->Color[ds->i];
+	ds->Color[ds->i] = color;
     }
     return ocolor;
 }
 
-int32_t rpmdsRefs(const rpmds ds)
-{
-    int32_t Refs = 0;
-
-    if (ds != NULL && ds->i >= 0 && ds->i < ds->Count) {
-	if (ds->Refs != NULL)
-	    Refs = ds->Refs[ds->i];
-    }
-    return Refs;
-}
-
-int32_t rpmdsSetRefs(const rpmds ds, int32_t refs)
-{
-    int32_t orefs = 0;
-
-    if (ds != NULL && ds->i >= 0 && ds->i < ds->Count) {
-	if (ds->Refs != NULL) {
-	    orefs = ds->Refs[ds->i];
-	    ds->Refs[ds->i] = refs;
-	}
-    }
-    return orefs;
-}
-
 void rpmdsNotify(rpmds ds, const char * where, int rc)
 {
+    const char *DNEVR;
+
     if (!rpmIsDebug())
 	return;
     if (!(ds != NULL && ds->i >= 0 && ds->i < ds->Count))
 	return;
-    if (!(ds->Type != NULL && ds->DNEVR != NULL))
+    if (!(ds->Type != NULL && (DNEVR = rpmdsDNEVR(ds)) != NULL))
 	return;
 
     rpmlog(RPMLOG_DEBUG, "%9s: %-45s %-s %s\n", ds->Type,
-		(!strcmp(ds->DNEVR, "cached") ? ds->DNEVR : ds->DNEVR+2),
+		(rstreq(DNEVR, "cached") ? DNEVR : DNEVR+2),
 		(rc ? _("NO ") : _("YES")),
 		(where != NULL ? where : ""));
 }
@@ -475,13 +410,8 @@ int rpmdsNext(rpmds ds)
 
     if (ds != NULL && ++ds->i >= 0) {
 	if (ds->i < ds->Count) {
-	    char t[2];
 	    i = ds->i;
 	    ds->DNEVR = _free(ds->DNEVR);
-	    t[0] = ((ds->Type != NULL) ? ds->Type[0] : '\0');
-	    t[1] = '\0';
-	    ds->DNEVR = rpmdsNewDNEVR(t, ds);
-
 	} else
 	    ds->i = -1;
 
@@ -495,8 +425,10 @@ fprintf(stderr, "*** ds %p\t%s[%d]: %s\n", ds, (ds->Type ? ds->Type : "?Type?"),
 
 rpmds rpmdsInit(rpmds ds)
 {
-    if (ds != NULL)
+    if (ds != NULL) {
 	ds->i = -1;
+	ds->DNEVR = _free(ds->DNEVR);
+    }
     return ds;
 }
 
@@ -538,21 +470,18 @@ static rpmds rpmdsDup(const rpmds ods)
     ds->u = ods->u;
     ds->nopromote = ods->nopromote;
 
-    nb = (ds->Count+1) * sizeof(*ds->N);
     ds->N = rpmdsDupArgv(ods->N, ods->Count);
 
     /* XXX rpm prior to 3.0.2 did not always supply EVR and Flags. */
 assert(ods->EVR != NULL);
 assert(ods->Flags != NULL);
 
-    nb = (ds->Count+1) * sizeof(*ds->EVR);
     ds->EVR = rpmdsDupArgv(ods->EVR, ods->Count);
 
     nb = (ds->Count * sizeof(*ds->Flags));
     ds->Flags = memcpy(xmalloc(nb), ods->Flags, nb);
 
-/* FIX: ds->Flags is kept, not only */
-    return rpmdsLink(ds, (ds ? ds->Type : NULL));
+    return rpmdsLink(ds);
 
 }
 
@@ -614,7 +543,6 @@ int rpmdsMerge(rpmds * dsp, rpmds ods)
      */
     save = ods->i;
     ods = rpmdsInit(ods);
-    if (ods != NULL)
     while (rpmdsNext(ods) >= 0) {
 	/*
 	 * If this entry is already present, don't bother.
@@ -685,15 +613,15 @@ int rpmdsSearch(rpmds ds, rpmds ods)
 	    l = i + 1;
 	else {
 	    /* Set l to 1st member of set that contains N. */
-	    if (strcmp(ods->N[ods->i], ds->N[l]))
+	    if (!rstreq(ods->N[ods->i], ds->N[l]))
 		l = i;
-	    while (l > 0 && !strcmp(ods->N[ods->i], ds->N[l-1]))
+	    while (l > 0 && rstreq(ods->N[ods->i], ds->N[l-1]))
 		l--;
 	    /* Set u to 1st member of set that does not contain N. */
-	    if (u >= ds->Count || strcmp(ods->N[ods->i], ds->N[u]))
+	    if (u >= ds->Count || !rstreq(ods->N[ods->i], ds->N[u]))
 		u = i;
 	    while (++u < ds->Count) {
-		if (strcmp(ods->N[ods->i], ds->N[u]))
+		if (!rstreq(ods->N[ods->i], ds->N[u]))
 		    /*@innerbreak@*/ break;
 	    }
 	    break;
@@ -763,15 +691,13 @@ void parseEVR(char * evr,
 
 int rpmdsCompare(const rpmds A, const rpmds B)
 {
-    const char *aDepend = (A->DNEVR != NULL ? A->DNEVR+2 : "");
-    const char *bDepend = (B->DNEVR != NULL ? B->DNEVR+2 : "");
     char *aEVR, *bEVR;
     const char *aE, *aV, *aR, *bE, *bV, *bR;
     int result;
     int sense;
 
     /* Different names don't overlap. */
-    if (strcmp(A->N[A->i], B->N[B->i])) {
+    if (!rstreq(A->N[A->i], B->N[B->i])) {
 	result = 0;
 	goto exit;
     }
@@ -806,9 +732,6 @@ int rpmdsCompare(const rpmds A, const rpmds B)
 	sense = rpmvercmp(aE, bE);
     else if (aE && *aE && atol(aE) > 0) {
 	if (!B->nopromote) {
-	    int lvl = (_rpmds_unspecified_epoch_noise  ? RPMLOG_WARNING : RPMLOG_DEBUG);
-	    rpmlog(lvl, _("The \"B\" dependency needs an epoch (assuming same epoch as \"A\")\n\tA = \"%s\"\tB = \"%s\"\n"),
-		aDepend, bDepend);
 	    sense = 0;
 	} else
 	    sense = 1;
@@ -817,8 +740,20 @@ int rpmdsCompare(const rpmds A, const rpmds B)
 
     if (sense == 0) {
 	sense = rpmvercmp(aV, bV);
-	if (sense == 0 && aR && *aR && bR && *bR)
-	    sense = rpmvercmp(aR, bR);
+	if (sense == 0) {
+	    if (aR && *aR && bR && *bR) {
+		sense = rpmvercmp(aR, bR);
+	    } else {
+		/* always matches if the side with no release has SENSE_EQUAL */
+		if ((aR && *aR && (B->Flags[B->i] & RPMSENSE_EQUAL)) ||
+		    (bR && *bR && (A->Flags[A->i] & RPMSENSE_EQUAL))) {
+		    aEVR = _free(aEVR);
+		    bEVR = _free(bEVR);
+		    result = 1;
+		    goto exit;
+		}
+	    }
+	}
     }
     aEVR = _free(aEVR);
     bEVR = _free(bEVR);
@@ -837,40 +772,9 @@ int rpmdsCompare(const rpmds A, const rpmds B)
     }
 
 exit:
-    if (_noisy_range_comparison_debug_message)
-    rpmlog(RPMLOG_DEBUG, "  %s    A %s\tB %s\n",
-	(result ? "YES" : "NO "), aDepend, bDepend);
     return result;
 }
 
-void rpmdsProblem(rpmps ps, const char * pkgNEVR, const rpmds ds,
-	const fnpyKey * suggestedKeys, int adding)
-{
-    const char * Name =  rpmdsN(ds);
-    const char * DNEVR = rpmdsDNEVR(ds);
-    const char * EVR = rpmdsEVR(ds);
-    rpmProblemType type;
-    fnpyKey key;
-
-    if (ps == NULL) return;
-
-    if (Name == NULL) Name = "?N?";
-    if (EVR == NULL) EVR = "?EVR?";
-    if (DNEVR == NULL) DNEVR = "? ?N? ?OP? ?EVR?";
-
-    rpmlog(RPMLOG_DEBUG, "package %s has unsatisfied %s: %s\n",
-	    pkgNEVR, ds->Type, DNEVR+2);
-
-    switch ((unsigned)DNEVR[0]) {
-    case 'C':	type = RPMPROB_CONFLICT;	break;
-    default:
-    case 'R':	type = RPMPROB_REQUIRES;	break;
-    }
-
-    key = (suggestedKeys ? suggestedKeys[0] : NULL);
-    rpmpsAppend(ps, type, pkgNEVR, key, NULL, NULL, DNEVR, adding);
-}
-
 int rpmdsAnyMatchesDep (const Header h, const rpmds req, int nopromote)
 {
     rpmds provides = NULL;
@@ -901,11 +805,10 @@ int rpmdsAnyMatchesDep (const Header h, const rpmds req, int nopromote)
     }
 
     result = 0;
-    if (provides != NULL)
     while (rpmdsNext(provides) >= 0) {
 
 	/* Filter out provides that came along for the ride. */
-	if (strcmp(provides->N[provides->i], req->N[req->i]))
+	if (!rstreq(provides->N[provides->i], req->N[req->i]))
 	    continue;
 
 	result = rpmdsCompare(provides, req);
@@ -923,9 +826,6 @@ exit:
 
 int rpmdsNVRMatchesDep(const Header h, const rpmds req, int nopromote)
 {
-    const char * pkgN;
-    char * pkgEVR;
-    rpmsenseFlags pkgFlags = RPMSENSE_EQUAL;
     rpmds pkg;
     int rc = 1;	/* XXX assume match, names already match here */
 
@@ -937,14 +837,11 @@ int rpmdsNVRMatchesDep(const Header h, const rpmds req, int nopromote)
 	return rc;
 
     /* Get package information from header */
-    pkgEVR = headerGetEVR(h, &pkgN);
-    if ((pkg = rpmdsSingle(RPMTAG_PROVIDENAME, pkgN, pkgEVR, pkgFlags)) != NULL) {
-	if (nopromote)
-	    (void) rpmdsSetNoPromote(pkg, nopromote);
-	rc = rpmdsCompare(pkg, req);
-	pkg = rpmdsFree(pkg);
-    }
-    free(pkgEVR);
+    pkg = rpmdsThis(h, RPMTAG_PROVIDENAME, RPMSENSE_EQUAL);
+    if (nopromote)
+	rpmdsSetNoPromote(pkg, nopromote);
+    rc = rpmdsCompare(pkg, req);
+    rpmdsFree(pkg);
 
     return rc;
 }
@@ -1009,11 +906,14 @@ static const struct rpmlibProvides_s rpmlibProvides[] = {
 	(		 RPMSENSE_EQUAL),
     N_("support for POSIX.1e file capabilities") },
 #endif
+    { "rpmlib(ScriptletExpansion)",    "4.9.0-1",
+	(		RPMSENSE_EQUAL),
+    N_("package scriptlets can be expanded at install time.") },
     { NULL,				NULL, 0,	NULL }
 };
 
 
-int rpmdsRpmlib(rpmds * dsp, void * tblp)
+int rpmdsRpmlib(rpmds * dsp, const void * tblp)
 {
     const struct rpmlibProvides_s * rltblp = tblp;
     const struct rpmlibProvides_s * rlp;
