@@ -15,10 +15,6 @@
 extern "C" {
 #endif
 
-/**
- */
-extern int _rpmds_debug;
-
 /**
  */
 extern int _rpmds_nopromote;
@@ -26,15 +22,15 @@ extern int _rpmds_nopromote;
 /** \ingroup rpmds
  * Dependency Attributes.
  */
-typedef	enum rpmsenseFlags_e {
+enum rpmsenseFlags_e {
     RPMSENSE_ANY	= 0,
     RPMSENSE_LESS	= (1 << 1),
     RPMSENSE_GREATER	= (1 << 2),
     RPMSENSE_EQUAL	= (1 << 3),
-    RPMSENSE_PROVIDES	= (1 << 4), /* only used internally by builds */
-    RPMSENSE_CONFLICTS	= (1 << 5), /* only used internally by builds */
-    RPMSENSE_PREREQ	= (1 << 6), /* legacy prereq dependency */
-    RPMSENSE_OBSOLETES	= (1 << 7), /* only used internally by builds */
+    /* bit 4 unused */
+    RPMSENSE_POSTTRANS	= (1 << 5),	/*!< %posttrans dependency */
+    RPMSENSE_PREREQ	= (1 << 6), 	/* legacy prereq dependency */
+    RPMSENSE_PRETRANS	= (1 << 7),	/*!< Pre-transaction dependency. */
     RPMSENSE_INTERP	= (1 << 8),	/*!< Interpreter used by scriptlet. */
     RPMSENSE_SCRIPT_PRE	= (1 << 9),	/*!< %pre dependency. */
     RPMSENSE_SCRIPT_POST = (1 << 10),	/*!< %post dependency. */
@@ -48,16 +44,15 @@ typedef	enum rpmsenseFlags_e {
     RPMSENSE_TRIGGERUN	= (1 << 17),	/*!< %triggerun dependency. */
     RPMSENSE_TRIGGERPOSTUN = (1 << 18),	/*!< %triggerpostun dependency. */
     RPMSENSE_MISSINGOK	= (1 << 19),	/*!< suggests/enhances hint. */
-    RPMSENSE_SCRIPT_PREP = (1 << 20),	/*!< %prep build dependency. */
-    RPMSENSE_SCRIPT_BUILD = (1 << 21),	/*!< %build build dependency. */
-    RPMSENSE_SCRIPT_INSTALL = (1 << 22),/*!< %install build dependency. */
-    RPMSENSE_SCRIPT_CLEAN = (1 << 23),	/*!< %clean build dependency. */
+    /* bits 20-23 unused */
     RPMSENSE_RPMLIB = (1 << 24),	/*!< rpmlib(feature) dependency. */
     RPMSENSE_TRIGGERPREIN = (1 << 25),	/*!< %triggerprein dependency. */
     RPMSENSE_KEYRING	= (1 << 26),
-    RPMSENSE_PATCHES	= (1 << 27),
+    /* bit 27 unused */
     RPMSENSE_CONFIG	= (1 << 28)
-} rpmsenseFlags;
+};
+
+typedef rpmFlags rpmsenseFlags;
 
 #define	RPMSENSE_SENSEMASK	15	 /* Mask to get senses, ie serial, */
                                          /* less, greater, equal.          */
@@ -73,17 +68,15 @@ typedef	enum rpmsenseFlags_e {
     RPMSENSE_SCRIPT_POSTUN | \
     RPMSENSE_SCRIPT_VERIFY | \
     RPMSENSE_FIND_REQUIRES | \
-    RPMSENSE_SCRIPT_PREP | \
-    RPMSENSE_SCRIPT_BUILD | \
-    RPMSENSE_SCRIPT_INSTALL | \
-    RPMSENSE_SCRIPT_CLEAN | \
     RPMSENSE_RPMLIB | \
     RPMSENSE_KEYRING | \
+    RPMSENSE_PRETRANS | \
+    RPMSENSE_POSTTRANS | \
     RPMSENSE_PREREQ)
 
 #define	_notpre(_x)		((_x) & ~RPMSENSE_PREREQ)
 #define	_INSTALL_ONLY_MASK \
-    _notpre(RPMSENSE_SCRIPT_PRE|RPMSENSE_SCRIPT_POST|RPMSENSE_RPMLIB|RPMSENSE_KEYRING)
+    _notpre(RPMSENSE_SCRIPT_PRE|RPMSENSE_SCRIPT_POST|RPMSENSE_RPMLIB|RPMSENSE_KEYRING|RPMSENSE_PRETRANS|RPMSENSE_POSTTRANS)
 #define	_ERASE_ONLY_MASK  \
     _notpre(RPMSENSE_SCRIPT_PREUN|RPMSENSE_SCRIPT_POSTUN)
 
@@ -91,20 +84,12 @@ typedef	enum rpmsenseFlags_e {
 #define	isInstallPreReq(_x)	((_x) & _INSTALL_ONLY_MASK)
 #define	isErasePreReq(_x)	((_x) & _ERASE_ONLY_MASK)
 
-/** \ingroup rpmds
- * Unreference a dependency set instance.
- * @param ds		dependency set
- * @param msg
- * @return		NULL always
- */
-rpmds rpmdsUnlink (rpmds ds, const char * msg);
 /** \ingroup rpmds
  * Reference a dependency set instance.
  * @param ds		dependency set
- * @param msg
  * @return		new dependency set reference
  */
-rpmds rpmdsLink (rpmds ds, const char * msg);
+rpmds rpmdsLink(rpmds ds);
 
 /** \ingroup rpmds
  * Destroy a dependency set.
@@ -119,7 +104,7 @@ rpmds rpmdsFree(rpmds ds);
  * @param flags		unused
  * @return		new dependency set
  */
-rpmds rpmdsNew(Header h, rpmTag tagN, int flags);
+rpmds rpmdsNew(Header h, rpmTagVal tagN, int flags);
 
 /** \ingroup rpmds
  * Return new formatted dependency string.
@@ -136,7 +121,7 @@ char * rpmdsNewDNEVR(const char * dspfx, const rpmds ds);
  * @param Flags		comparison flags
  * @return		new dependency set
  */
-rpmds rpmdsThis(Header h, rpmTag tagN, rpmsenseFlags Flags);
+rpmds rpmdsThis(Header h, rpmTagVal tagN, rpmsenseFlags Flags);
 
 /** \ingroup rpmds
  * Create, load and initialize a dependency set of size 1.
@@ -146,7 +131,14 @@ rpmds rpmdsThis(Header h, rpmTag tagN, rpmsenseFlags Flags);
  * @param Flags		comparison flags
  * @return		new dependency set
  */
-rpmds rpmdsSingle(rpmTag tagN, const char * N, const char * EVR, rpmsenseFlags Flags);
+rpmds rpmdsSingle(rpmTagVal tagN, const char * N, const char * EVR, rpmsenseFlags Flags);
+
+/** \ingroup rpmds
+ * Return a new dependency set of size 1 from the current iteration index
+ * @param ds		dependency set
+ * @return		new dependency set
+ */
+rpmds rpmdsCurrent(rpmds ds);
 
 /** \ingroup rpmds
  * Return dependency set count.
@@ -203,22 +195,15 @@ rpmsenseFlags rpmdsFlags(const rpmds ds);
  * @param ds		dependency set
  * @return		current dependency type, 0 on invalid
  */
-rpmTag rpmdsTagN(const rpmds ds);
+rpmTagVal rpmdsTagN(const rpmds ds);
 
 /** \ingroup rpmds
- * Return dependency build time.
+ * Return dependency header instance, ie whether the dependency comes from 
+ * an installed header or not.
  * @param ds		dependency set
- * @return		dependency build time, 0 on invalid
+ * @return		header instance of dependency (0 for not installed)
  */
-time_t rpmdsBT(const rpmds ds);
-
-/** \ingroup rpmds
- * Set dependency build time.
- * @param ds		dependency set
- * @param BT		build time
- * @return		dependency build time, 0 on invalid
- */
-time_t rpmdsSetBT(const rpmds ds, time_t BT);
+unsigned int rpmdsInstance(rpmds ds);
 
 /** \ingroup rpmds
  * Return current "Don't promote Epoch:" flag.
@@ -258,21 +243,6 @@ rpm_color_t rpmdsColor(const rpmds ds);
  */
 rpm_color_t rpmdsSetColor(const rpmds ds, rpm_color_t color);
 
-/** \ingroup rpmds
- * Return current dependency file refs.
- * @param ds		dependency set
- * @return		current dependency file refs, -1 on global
- */
-int32_t rpmdsRefs(const rpmds ds);
-
-/** \ingroup rpmds
- * Return current dependency color.
- * @param ds		dependency set
- * @param refs		new dependency refs
- * @return		previous dependency refs
- */
-int32_t rpmdsSetRefs(const rpmds ds, int32_t refs);
-
 /** \ingroup rpmds
  * Notify of results of dependency match.
  * @param ds		dependency set
@@ -322,7 +292,6 @@ int rpmdsMerge(rpmds * dsp, rpmds ods);
  **/
 int rpmdsSearch(rpmds ds, rpmds ods);
 
-
 /** \ingroup rpmds
  * Compare two versioned dependency ranges, looking for overlap.
  * @param A		1st dependency
@@ -331,18 +300,6 @@ int rpmdsSearch(rpmds ds, rpmds ods);
  */
 int rpmdsCompare(const rpmds A, const rpmds B);
 
-/** \ingroup rpmds
- * Report a Requires: or Conflicts: dependency problem.
- * @param ps		transaction set problems
- * @param pkgNEVR	package name/epoch/version/release
- * @param ds		dependency set
- * @param suggestedKeys	filename or python object address
- * @param adding	dependency problem is from added package set?
- */
-void rpmdsProblem(rpmps ps, const char * pkgNEVR, const rpmds ds,
-		const fnpyKey * suggestedKeys,
-		int adding);
-
 /** \ingroup rpmds
  * Compare package provides dependencies from header with a single dependency.
  * @param h		header
@@ -367,7 +324,7 @@ int rpmdsNVRMatchesDep(const Header h, const rpmds req, int nopromote);
  * @param tblp		rpmlib provides table (NULL uses internal table)
  * @return		0 on success
  */
-int rpmdsRpmlib(rpmds * dsp, void * tblp);
+int rpmdsRpmlib(rpmds * dsp, const void * tblp);
 
 #ifdef __cplusplus
 }
