@@ -12,8 +12,8 @@
 #include <rpm/rpmts.h>
 #include <rpm/rpmlog.h>
 #include <rpm/rpmfileutil.h>
-#include <rpm/rpmgi.h>
 
+#include "lib/rpmgi.h"
 #include "lib/manifest.h"
 #include "debug.h"
 
@@ -81,11 +81,11 @@ static rpmVSFlags setvsFlags(struct rpmInstallArguments_s * ia)
     else
 	vsflags = rpmExpandNumeric("%{?_vsflags_install}");
 
-    if (ia->qva_flags & VERIFY_DIGEST)
+    if (rpmcliQueryFlags & VERIFY_DIGEST)
 	vsflags |= _RPMVSF_NODIGESTS;
-    if (ia->qva_flags & VERIFY_SIGNATURE)
+    if (rpmcliQueryFlags & VERIFY_SIGNATURE)
 	vsflags |= _RPMVSF_NOSIGNATURES;
-    if (ia->qva_flags & VERIFY_HDRCHK)
+    if (rpmcliQueryFlags & VERIFY_HDRCHK)
 	vsflags |= RPMVSF_NOHDRCHK;
 
     return vsflags;
@@ -104,7 +104,6 @@ void * rpmShowProgress(const void * arg,
     void * rc = NULL;
     const char * filename = (const char *)key;
     static FD_t fd = NULL;
-    int xx;
 
     switch (what) {
     case RPMCALLBACK_INST_OPEN_FILE:
@@ -116,19 +115,19 @@ void * rpmShowProgress(const void * arg,
 	    rpmlog(RPMLOG_ERR, _("open of %s failed: %s\n"), filename,
 			Fstrerror(fd));
 	    if (fd != NULL) {
-		xx = Fclose(fd);
+		Fclose(fd);
 		fd = NULL;
 	    }
 	} else
-	    fd = fdLink(fd, RPMDBG_M("persist (showProgress)"));
+	    fd = fdLink(fd);
 	return (void *)fd;
 	break;
 
     case RPMCALLBACK_INST_CLOSE_FILE:
 	/* FIX: still necessary? */
-	fd = fdFree(fd, RPMDBG_M("persist (showProgress)"));
+	fd = fdFree(fd);
 	if (fd != NULL) {
-	    xx = Fclose(fd);
+	    Fclose(fd);
 	    fd = NULL;
 	}
 	break;
@@ -208,16 +207,14 @@ void * rpmShowProgress(const void * arg,
 static void setNotifyFlag(struct rpmInstallArguments_s * ia,
 			  rpmts ts)
 {
-    int notifyFlags, xx;
+    int notifyFlags;
 
     notifyFlags = ia->installInterfaceFlags | (rpmIsVerbose() ? INSTALL_LABEL : 0 );
-    xx = rpmtsSetNotifyCallback(ts,
-				rpmShowProgress, (void *) ((long)notifyFlags));
+    rpmtsSetNotifyCallback(ts, rpmShowProgress, (void *) ((long)notifyFlags));
 }
 
 struct rpmEIU {
     Header h;
-    FD_t fd;
     int numFailed;
     int numPkgs;
     char ** pkgURL;
@@ -262,7 +259,7 @@ static int rpmcliTransaction(rpmts ts, struct rpmInstallArguments_s * ia,
 	ps = rpmpsFree(ps);
     }
 
-    if ((eflags? 1 : (!stop)) && !(ia->installInterfaceFlags & INSTALL_NOORDER)) {
+    if (!stop && !(ia->installInterfaceFlags & INSTALL_NOORDER)) {
 	if (rpmtsOrder(ts)) {
 	    rc = numPackages;
 	    stop = 1;
@@ -270,16 +267,10 @@ static int rpmcliTransaction(rpmts ts, struct rpmInstallArguments_s * ia,
     }
 
     if (numPackages && !stop) {
-
-	if (eflags) {
-	    rpmlog(RPMLOG_DEBUG, "erasing packages\n");
-	    rpmtsClean(ts);
-	    rc = rpmtsRun(ts, NULL, ia->probFilter & (RPMPROB_FILTER_DISKSPACE|RPMPROB_FILTER_DISKNODES));
-	} else {
-	    rpmlog(RPMLOG_DEBUG, "installing binary packages\n");
-	    rpmtsClean(ts);
-	    rc = rpmtsRun(ts, NULL, ia->probFilter);
-	}
+	rpmlog(RPMLOG_DEBUG, eflags ? "erasing packages\n" :
+				      "installing binary packages\n");
+	rpmtsClean(ts);
+	rc = rpmtsRun(ts, NULL, ia->probFilter);
 
 	ps = rpmtsProblems(ts);
 
@@ -293,28 +284,28 @@ static int rpmcliTransaction(rpmts ts, struct rpmInstallArguments_s * ia,
 
 static int tryReadManifest(struct rpmEIU * eiu)
 {
-    int rc, xx;
+    int rc;
 
     /* Try to read a package manifest. */
-    eiu->fd = Fopen(*eiu->fnp, "r.fpio");
-    if (eiu->fd == NULL || Ferror(eiu->fd)) {
+    FD_t fd = Fopen(*eiu->fnp, "r.fpio");
+    if (fd == NULL || Ferror(fd)) {
         rpmlog(RPMLOG_ERR, _("open of %s failed: %s\n"), *eiu->fnp,
-	       Fstrerror(eiu->fd));
-	if (eiu->fd != NULL) {
-	    xx = Fclose(eiu->fd);
-	    eiu->fd = NULL;
+	       Fstrerror(fd));
+	if (fd != NULL) {
+	    Fclose(fd);
+	    fd = NULL;
 	}
 	eiu->numFailed++; *eiu->fnp = NULL;
 	return RPMRC_FAIL;
     }
 
     /* Read list of packages from manifest. */
-    rc = rpmReadPackageManifest(eiu->fd, &eiu->argc, &eiu->argv);
+    rc = rpmReadPackageManifest(fd, &eiu->argc, &eiu->argv);
     if (rc != RPMRC_OK)
         rpmlog(RPMLOG_ERR, _("%s: not an rpm package (or package manifest): %s\n"),
-	       *eiu->fnp, Fstrerror(eiu->fd));
-    xx = Fclose(eiu->fd);
-    eiu->fd = NULL;
+	       *eiu->fnp, Fstrerror(fd));
+    Fclose(fd);
+    fd = NULL;
 
     if (rc != RPMRC_OK)
         eiu->numFailed++; *eiu->fnp = NULL;
@@ -322,6 +313,68 @@ static int tryReadManifest(struct rpmEIU * eiu)
     return rc;
 }
 
+static int tryReadHeader(rpmts ts, struct rpmEIU * eiu, rpmVSFlags vsflags)
+{
+   rpmVSFlags tvsflags;
+
+   /* Try to read the header from a package file. */
+   FD_t fd = Fopen(*eiu->fnp, "r.ufdio");
+   if (fd == NULL || Ferror(fd)) {
+       rpmlog(RPMLOG_ERR, _("open of %s failed: %s\n"), *eiu->fnp,
+	      Fstrerror(fd));
+       if (fd != NULL) {
+           Fclose(fd);
+	   fd = NULL;
+       }
+       eiu->numFailed++; *eiu->fnp = NULL;
+       return RPMRC_FAIL;
+   }
+
+   /* Read the header, verifying signatures (if present). */
+   tvsflags = rpmtsSetVSFlags(ts, vsflags);
+   eiu->rpmrc = rpmReadPackageFile(ts, fd, *eiu->fnp, &eiu->h);
+   tvsflags = rpmtsSetVSFlags(ts, tvsflags);
+   Fclose(fd);
+   fd = NULL;
+   
+   /* Honor --nomanifest */
+   if (eiu->rpmrc == RPMRC_NOTFOUND && (giFlags & RPMGI_NOMANIFEST))
+       eiu->rpmrc = RPMRC_FAIL;
+
+   if(eiu->rpmrc == RPMRC_FAIL) {
+       rpmlog(RPMLOG_ERR, _("%s cannot be installed\n"), *eiu->fnp);
+       eiu->numFailed++; *eiu->fnp = NULL;
+   }
+
+   return RPMRC_OK;
+}
+
+
+/* On --freshen, verify package is installed and newer */
+static int checkFreshenStatus(rpmts ts, struct rpmEIU * eiu)
+{
+    rpmdbMatchIterator mi = NULL;
+    const char * name = headerGetString(eiu->h, RPMTAG_NAME);
+    const char *arch = headerGetString(eiu->h, RPMTAG_ARCH);
+    Header oldH = NULL;
+
+    if (name != NULL)
+        mi = rpmtsInitIterator(ts, RPMDBI_NAME, name, 0);
+    if (rpmtsColor(ts) && arch)
+	rpmdbSetIteratorRE(mi, RPMTAG_ARCH, RPMMIRE_DEFAULT, arch);
+
+    while ((oldH = rpmdbNextIterator(mi)) != NULL) {
+	/* Package is newer than those currently installed. */
+        if (rpmVersionCompare(oldH, eiu->h) < 0)
+	    break;
+    }
+
+    mi = rpmdbFreeIterator(mi);
+    if (oldH == NULL) {
+        eiu->h = headerFree(eiu->h);
+    }
+   return (oldH != NULL);
+}
 
 /** @todo Generalize --freshen policies. */
 int rpmInstall(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_t fileArgv)
@@ -329,9 +382,8 @@ int rpmInstall(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_t fileArgv)
     struct rpmEIU * eiu = xcalloc(1, sizeof(*eiu));
     rpmRelocation * relocations;
     char * fileURL = NULL;
-    rpmVSFlags vsflags, ovsflags, tvsflags;
+    rpmVSFlags vsflags, ovsflags;
     int rc;
-    int xx;
     int i;
 
     if (fileArgv == NULL) goto exit;
@@ -398,7 +450,7 @@ restart:
 	case URL_IS_HTTPS:
 	case URL_IS_HTTP:
 	case URL_IS_FTP:
-	{   char *tfn;
+	{   char *tfn = NULL;
 	    FD_t tfd;
 
 	    if (rpmIsVerbose())
@@ -448,48 +500,15 @@ restart:
 	rpmlog(RPMLOG_DEBUG, "============== %s\n", *eiu->fnp);
 	(void) urlPath(*eiu->fnp, &fileName);
 
-	/* Try to read the header from a package file. */
-	eiu->fd = Fopen(*eiu->fnp, "r.ufdio");
-	if (eiu->fd == NULL || Ferror(eiu->fd)) {
-	    rpmlog(RPMLOG_ERR, _("open of %s failed: %s\n"), *eiu->fnp,
-			Fstrerror(eiu->fd));
-	    if (eiu->fd != NULL) {
-		xx = Fclose(eiu->fd);
-		eiu->fd = NULL;
-	    }
-	    eiu->numFailed++; *eiu->fnp = NULL;
+	if (tryReadHeader(ts, eiu, vsflags) == RPMRC_FAIL)
 	    continue;
-	}
 
-	/* Read the header, verifying signatures (if present). */
-	tvsflags = rpmtsSetVSFlags(ts, vsflags);
-	eiu->rpmrc = rpmReadPackageFile(ts, eiu->fd, *eiu->fnp, &eiu->h);
-	tvsflags = rpmtsSetVSFlags(ts, tvsflags);
-	xx = Fclose(eiu->fd);
-	eiu->fd = NULL;
-
-	/* Honor --nomanifest */
-	if (eiu->rpmrc == RPMRC_NOTFOUND && (giFlags & RPMGI_NOMANIFEST))
-	    eiu->rpmrc = RPMRC_FAIL;
-
-	switch (eiu->rpmrc) {
-	case RPMRC_FAIL:
-	    rpmlog(RPMLOG_ERR, _("%s cannot be installed\n"), *eiu->fnp);
-	    eiu->numFailed++; *eiu->fnp = NULL;
-	    continue;
-	    break;
-	case RPMRC_NOTFOUND:
+	if (eiu->rpmrc == RPMRC_NOTFOUND) {
 	    rc = tryReadManifest(eiu);
 	    if (rc == RPMRC_OK) {
 	        eiu->prevx++;
-		goto restart;
+	        goto restart;
 	    }
-	    break;
-	case RPMRC_NOTTRUSTED:
-	case RPMRC_NOKEY:
-	case RPMRC_OK:
-	default:
-	    break;
 	}
 
 	eiu->isSource = headerIsSource(eiu->h);
@@ -514,39 +533,16 @@ restart:
 		eiu->relocations->oldPath = xstrdup(rpmtdGetString(&prefixes));
 		rpmtdFreeData(&prefixes);
 	    } else {
-		const char * name;
-		xx = headerNVR(eiu->h, &name, NULL, NULL);
-		rpmlog(RPMLOG_ERR,
-			       _("package %s is not relocatable\n"), name);
+		rpmlog(RPMLOG_ERR, _("package %s is not relocatable\n"),
+		       headerGetString(eiu->h, RPMTAG_NAME));
 		eiu->numFailed++;
 		goto exit;
 	    }
 	}
 
-	/* On --freshen, verify package is installed and newer */
-	if (ia->installInterfaceFlags & INSTALL_FRESHEN) {
-	    rpmdbMatchIterator mi;
-	    const char * name;
-	    Header oldH;
-	    int count;
-
-	    xx = headerNVR(eiu->h, &name, NULL, NULL);
-	    mi = rpmtsInitIterator(ts, RPMTAG_NAME, name, 0);
-	    count = rpmdbGetIteratorCount(mi);
-	    while ((oldH = rpmdbNextIterator(mi)) != NULL) {
-		if (rpmVersionCompare(oldH, eiu->h) < 0)
-		    continue;
-		/* same or newer package already installed */
-		count = 0;
-		break;
-	    }
-	    mi = rpmdbFreeIterator(mi);
-	    if (count == 0) {
-		eiu->h = headerFree(eiu->h);
-		continue;
-	    }
-	    /* Package is newer than those currently installed. */
-	}
+	if (ia->installInterfaceFlags & INSTALL_FRESHEN)
+	    if (checkFreshenStatus(ts, eiu) != 1)
+	        continue;
 
 	rc = rpmtsAddInstallElement(ts, eiu->h, (fnpyKey)fileName,
 			(ia->installInterfaceFlags & INSTALL_UPGRADE) != 0,
@@ -643,9 +639,6 @@ int rpmErase(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_const_t argv)
     vsflags = setvsFlags(ia);
     ovsflags = rpmtsSetVSFlags(ts, vsflags);
 
-    /* XXX suggest mechanism only meaningful when installing */
-    ia->transFlags |= RPMTRANS_FLAG_NOSUGGEST;
-
     (void) rpmtsSetFlags(ts, ia->transFlags);
 
 #ifdef	NOTYET	/* XXX no callbacks on erase yet */
