@@ -5,6 +5,10 @@
 #include <setjmp.h>
 #endif
 
+#if HAVE_SYS_UTSNAME_H
+#include <sys/utsname.h>
+#endif
+#include <netdb.h>
 #include <ctype.h>	/* XXX for /etc/rpm/platform contents */
 
 #if HAVE_SYS_SYSTEMCFG_H
@@ -119,7 +123,7 @@ static struct tableType_s tables[RPM_MACHTABLE_COUNT] = {
 /* this *must* be kept in alphabetical order */
 /* The order of the flags is archSpecific, required, macroize, localize */
 
-static const struct rpmOption const optionTable[] = {
+static const struct rpmOption optionTable[] = {
     { "include",		RPMVAR_INCLUDE,			0, 1,	0, 2 },
     { "macrofiles",		RPMVAR_MACROFILES,		0, 0,	0, 1 },
     { "optflags",		RPMVAR_OPTFLAGS,		1, 0,	1, 0 },
@@ -426,6 +430,7 @@ static void setDefaults(void)
     if (!macrofiles) {
 	macrofiles = rstrscat(NULL, confdir, "/macros", ":",
 				confdir, "/platform/%{_target}/macros", ":",
+				confdir, "/fileattrs/*.attr", ":",
   				confdir, "/" RPMCANONVENDOR "/macros", ":",
 				SYSCONFDIR "/rpm/macros.*", ":",
 				SYSCONFDIR "/rpm/macros", ":",
@@ -619,9 +624,9 @@ static rpmRC rpmPlatform(const char * platform)
     ssize_t blen = 0;
     int init_platform = 0;
     char * p, * pe;
-    int rc;
+    rpmRC rc;
 
-    rc = rpmioSlurp(platform, &b, &blen);
+    rc = (rpmioSlurp(platform, &b, &blen) == 0) ? RPMRC_OK : RPMRC_FAIL;
 
     if (rc || b == NULL || blen <= 0) {
 	rc = RPMRC_FAIL;
@@ -1148,9 +1153,9 @@ static void defaultMachine(const char ** arch,
 
 #	if defined(__linux__) && defined(__i386__)
 	{
-	    char class = (char) (RPMClass() | '0');
+	    char mclass = (char) (RPMClass() | '0');
 
-	    if ((class == '6' && is_athlon()) || class == '7')
+	    if ((mclass == '6' && is_athlon()) || mclass == '7')
 	    	strcpy(un.machine, "athlon");
 	    else if (is_pentium4())
 		strcpy(un.machine, "pentium4");
@@ -1158,8 +1163,8 @@ static void defaultMachine(const char ** arch,
 		strcpy(un.machine, "pentium3");
 	    else if (is_geode())
 		strcpy(un.machine, "geode");
-	    else if (strchr("3456", un.machine[1]) && un.machine[1] != class)
-		un.machine[1] = class;
+	    else if (strchr("3456", un.machine[1]) && un.machine[1] != mclass)
+		un.machine[1] = mclass;
 	}
 #	endif
 
@@ -1262,8 +1267,10 @@ void rpmSetTables(int archTable, int osTable)
 
 int rpmMachineScore(int type, const char * name)
 {
-    machEquivInfo info = machEquivSearch(&tables[type].equiv, name);
-    return (info != NULL ? info->score : 0);
+    machEquivInfo info = NULL;
+    if (name)
+	info = machEquivSearch(&tables[type].equiv, name);
+    return info ? info->score : 0;
 }
 
 int rpmIsKnownArch(const char *name)
@@ -1438,8 +1445,7 @@ static void rpmRebuildTargetVars(const char ** target, const char ** canontarget
 
     /* XXX For now, set canonical target to arch-os */
     if (ct == NULL) {
-	ct = xmalloc(strlen(ca) + sizeof("-") + strlen(co));
-	sprintf(ct, "%s-%s", ca, co);
+	rasprintf(&ct, "%s-%s", ca, co);
     }
 
 /*
@@ -1600,10 +1606,6 @@ exit:
 
 int rpmReadConfigFiles(const char * file, const char * target)
 {
-    mode_t mode = 0022;
-    /* Reset umask to its default umask(2) value. */
-    mode = umask(mode);
-
     /* Force preloading of dlopen()'ed libraries in case we go chrooting */
     (void) gethostbyname("localhost");
     (void) rpmInitCrypto();
