@@ -5,23 +5,20 @@
 #include "system.h"
 
 #include <rpm/rpmtypes.h>
-#include <rpm/rpmlib.h>		/* rpmGetFilesystem*() */
+#include <rpm/rpmlib.h>
 #include <rpm/rpmmacro.h>	/* XXX for %_i18ndomains */
 #include <rpm/rpmfi.h>
 #include <rpm/rpmstring.h>
+#include <rpm/rpmlog.h>
+#include "lib/misc.h"		/* tag function proto */
 
 #include "debug.h"
 
 struct headerTagFunc_s {
     rpmTag tag;		/*!< Tag of extension. */
-    void *func;		/*!< Pointer to formatter function. */	
+    headerTagTagFunction func;	/*!< Pointer to formatter function. */	
 };
 
-/* forward declarations */
-static const struct headerTagFunc_s rpmHeaderTagExtensions[];
-
-void *rpmHeaderTagFunc(rpmTag tag);
-
 /** \ingroup rpmfi
  * Retrieve file names from header.
  *
@@ -46,16 +43,13 @@ static void rpmfiBuildFNames(Header h, rpmTag tagN,
     uint32_t *dirIndexes;
     rpm_count_t count;
     size_t size;
-    rpmTag dirNameTag = 0;
-    rpmTag dirIndexesTag = 0;
+    rpmTag dirNameTag = RPMTAG_DIRNAMES;
+    rpmTag dirIndexesTag = RPMTAG_DIRINDEXES;
     char * t;
     int i;
     struct rpmtd_s bnames, dnames, dixs;
 
-    if (tagN == RPMTAG_BASENAMES) {
-	dirNameTag = RPMTAG_DIRNAMES;
-	dirIndexesTag = RPMTAG_DIRINDEXES;
-    } else if (tagN == RPMTAG_ORIGBASENAMES) {
+    if (tagN == RPMTAG_ORIGBASENAMES) {
 	dirNameTag = RPMTAG_ORIGDIRNAMES;
 	dirIndexesTag = RPMTAG_ORIGDIRINDEXES;
     }
@@ -97,7 +91,7 @@ static void rpmfiBuildFNames(Header h, rpmTag tagN,
     *fcp = count;
 }
 
-static int filedepTag(Header h, rpmTag tagN, rpmtd td)
+static int filedepTag(Header h, rpmTag tagN, rpmtd td, headerGetFlags hgflags)
 {
     rpmfi fi = rpmfiNew(NULL, h, RPMTAG_BASENAMES, RPMFI_NOHEADER);
     rpmds ds = NULL;
@@ -156,102 +150,13 @@ exit:
     return rc;
 }
 
-/**
- * Retrieve mounted file system paths.
- * @param h		header
- * @retval td		tag data container
- * @return		1 on success
- */
-static int fsnamesTag(Header h, rpmtd td)
-{
-    const char ** list;
-
-    if (rpmGetFilesystemList(&list, &(td->count)))
-	return 0;
-
-    td->type = RPM_STRING_ARRAY_TYPE;
-    td->data = list;
-
-    return 1; 
-}
-
-/**
- * Retrieve install prefixes.
- * @param h		header
- * @retval td		tag data container
- * @return		1 on success
- */
-static int instprefixTag(Header h, rpmtd td)
-{
-    struct rpmtd_s prefixes;
-    int flags = HEADERGET_MINMEM;
-
-    if (headerGet(h, RPMTAG_INSTALLPREFIX, td, flags)) {
-	return 1;
-    } else if (headerGet(h, RPMTAG_INSTPREFIXES, &prefixes, flags)) {
-	/* only return the first prefix of the array */
-	td->type = RPM_STRING_TYPE;
-	td->data = xstrdup(rpmtdGetString(&prefixes));
-	td->flags = RPMTD_ALLOCED;
-	rpmtdFreeData(&prefixes);
-	return 1;
-    }
-
-    return 0;
-}
-
-/**
- * Retrieve mounted file system space.
- * @param h		header
- * @retval td		tag data container
- * @return		1 on success
- */
-static int fssizesTag(Header h, rpmtd td)
-{
-    struct rpmtd_s fsizes, fnames;
-    const char ** filenames = NULL;
-    rpm_loff_t * filesizes = NULL;
-    rpm_loff_t * usages = NULL;
-    rpm_count_t numFiles = 0;
-
-    if (headerGet(h, RPMTAG_LONGFILESIZES, &fsizes, HEADERGET_EXT)) {
-	filesizes = fsizes.data;
-	headerGet(h, RPMTAG_FILENAMES, &fnames, HEADERGET_EXT);
-	filenames = fnames.data;
-	numFiles = rpmtdCount(&fnames);
-    }
-	
-    if (rpmGetFilesystemList(NULL, &(td->count)))
-	return 0;
-
-    td->type = RPM_INT64_TYPE;
-    td->flags = RPMTD_ALLOCED;
-
-    if (filenames == NULL) {
-	usages = xcalloc((td->count), sizeof(usages));
-	td->data = usages;
-
-	return 1;
-    }
-
-    if (rpmGetFilesystemUsage(filenames, filesizes, numFiles, &usages, 0))	
-	return 0;
-
-    td->data = usages;
-
-    rpmtdFreeData(&fnames);
-    rpmtdFreeData(&fsizes);
-
-    return 1;
-}
-
 /**
  * Retrieve trigger info.
  * @param h		header
  * @retval td		tag data container
  * @return		1 on success
  */
-static int triggercondsTag(Header h, rpmtd td)
+static int triggercondsTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
     uint32_t * indices;
     int i, j;
@@ -321,7 +226,7 @@ static int triggercondsTag(Header h, rpmtd td)
  * @retval td		tag data container
  * @return		1 on success
  */
-static int triggertypeTag(Header h, rpmtd td)
+static int triggertypeTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
     int i;
     char ** conds;
@@ -374,7 +279,7 @@ static int triggertypeTag(Header h, rpmtd td)
  * @retval td		tag data container
  * @return		1 on success
  */
-static int filenamesTag(Header h, rpmtd td)
+static int filenamesTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
     rpmfiBuildFNames(h, RPMTAG_BASENAMES, 
 		     (const char ***) &(td->data), &(td->count));
@@ -391,7 +296,7 @@ static int filenamesTag(Header h, rpmtd td)
  * @retval td		tag data container
  * @return		1 on success
  */
-static int origfilenamesTag(Header h, rpmtd td)
+static int origfilenamesTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
     rpmfiBuildFNames(h, RPMTAG_ORIGBASENAMES, 
 		     (const char ***) &(td->data), &(td->count));
@@ -407,7 +312,7 @@ static int origfilenamesTag(Header h, rpmtd td)
  * @retval td		tag data container
  * @return		1 on success
  */
-static int fileclassTag(Header h, rpmtd td)
+static int fileclassTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
     rpmfi fi = rpmfiNew(NULL, h, RPMTAG_BASENAMES, RPMFI_NOHEADER);
     char **fclasses;
@@ -443,9 +348,9 @@ exit:
  * @retval td		tag data container
  * @return		1 on success
  */
-static int fileprovideTag(Header h, rpmtd td)
+static int fileprovideTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
-    return filedepTag(h, RPMTAG_PROVIDENAME, td);
+    return filedepTag(h, RPMTAG_PROVIDENAME, td, hgflags);
 }
 
 /**
@@ -454,9 +359,9 @@ static int fileprovideTag(Header h, rpmtd td)
  * @retval td		tag data container
  * @return		1 on success
  */
-static int filerequireTag(Header h, rpmtd td)
+static int filerequireTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
-    return filedepTag(h, RPMTAG_REQUIRENAME, td);
+    return filedepTag(h, RPMTAG_REQUIRENAME, td, hgflags);
 }
 
 /* I18N look aside diversions */
@@ -475,10 +380,11 @@ static const char * const _macro_i18ndomains = "%{?_i18ndomains}";
  * @retval td		tag data container
  * @return		1 on success
  */
-static int i18nTag(Header h, rpmTag tag, rpmtd td)
+static int i18nTag(Header h, rpmTag tag, rpmtd td, headerGetFlags hgflags)
 {
-    char * dstring = rpmExpand(_macro_i18ndomains, NULL);
     int rc;
+#if defined(ENABLE_NLS)
+    char * dstring = rpmExpand(_macro_i18ndomains, NULL);
 
     td->type = RPM_STRING_TYPE;
     td->data = NULL;
@@ -489,18 +395,14 @@ static int i18nTag(Header h, rpmTag tag, rpmtd td)
 	const char * langval;
 	char * msgkey;
 	const char * msgid;
-	const char * n;
-	int xx;
 
-	xx = headerNVR(h, &n, NULL, NULL);
-	rasprintf(&msgkey, "%s(%s)", n, rpmTagGetName(tag));
+	rasprintf(&msgkey, "%s(%s)", headerGetString(h, RPMTAG_NAME), 
+		  rpmTagGetName(tag));
 
 	/* change to en_US for msgkey -> msgid resolution */
 	langval = getenv(language);
 	(void) setenv(language, "en_US", 1);
-#if defined(ENABLE_NLS)
         ++_nl_msg_cat_cntr;
-#endif
 
 	msgid = NULL;
 	for (domain = dstring; domain != NULL; domain = de) {
@@ -515,9 +417,7 @@ static int i18nTag(Header h, rpmTag tag, rpmtd td)
 	    (void) setenv(language, langval, 1);
 	else
 	    unsetenv(language);
-#if defined(ENABLE_NLS)
         ++_nl_msg_cat_cntr;
-#endif
 
 	if (domain && msgid) {
 	    td->data = dgettext(domain, msgid);
@@ -532,8 +432,9 @@ static int i18nTag(Header h, rpmTag tag, rpmtd td)
     }
 
     dstring = _free(dstring);
+#endif
 
-    rc = headerGet(h, tag, td, HEADERGET_DEFAULT);
+    rc = headerGet(h, tag, td, HEADERGET_ALLOC);
     return rc;
 }
 
@@ -543,9 +444,9 @@ static int i18nTag(Header h, rpmTag tag, rpmtd td)
  * @retval td		tag data container
  * @return		1 on success
  */
-static int summaryTag(Header h, rpmtd td)
+static int summaryTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
-    return i18nTag(h, RPMTAG_SUMMARY, td);
+    return i18nTag(h, RPMTAG_SUMMARY, td, hgflags);
 }
 
 /**
@@ -554,9 +455,9 @@ static int summaryTag(Header h, rpmtd td)
  * @retval td		tag data container
  * @return		1 on success
  */
-static int descriptionTag(Header h, rpmtd td)
+static int descriptionTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
-    return i18nTag(h, RPMTAG_DESCRIPTION, td);
+    return i18nTag(h, RPMTAG_DESCRIPTION, td, hgflags);
 }
 
 /**
@@ -565,9 +466,9 @@ static int descriptionTag(Header h, rpmtd td)
  * @retval td		tag data container
  * @return		1 on success
  */
-static int groupTag(Header h, rpmtd td)
+static int groupTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
-    return i18nTag(h, RPMTAG_GROUP, td);
+    return i18nTag(h, RPMTAG_GROUP, td, hgflags);
 }
 
 /*
@@ -611,38 +512,154 @@ static int get64(Header h, rpmtd td, rpmTag newtag, rpmTag oldtag)
  * @retval td		tag data container
  * @return		1 on success
  */
-static int longfilesizesTag(Header h, rpmtd td)
+static int longfilesizesTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
     return get64(h, td, RPMTAG_LONGFILESIZES, RPMTAG_FILESIZES);
 }
 
-static int longarchivesizeTag(Header h, rpmtd td)
+static int longarchivesizeTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
     return get64(h, td, RPMTAG_LONGARCHIVESIZE, RPMTAG_ARCHIVESIZE);
 }
 
-static int longsizeTag(Header h, rpmtd td)
+static int longsizeTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
     return get64(h, td, RPMTAG_LONGSIZE, RPMTAG_SIZE);
 }
 
-static int longsigsizeTag(Header h, rpmtd td)
+static int longsigsizeTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
     return get64(h, td, RPMTAG_LONGSIGSIZE, RPMTAG_SIGSIZE);
 }
 
-void *rpmHeaderTagFunc(rpmTag tag)
+static int numberTag(rpmtd td, uint32_t val)
 {
-    const struct headerTagFunc_s * ext;
-    void *func = NULL;
+    uint32_t *tval = xmalloc(sizeof(*tval));
 
-    for (ext = rpmHeaderTagExtensions; ext->func != NULL; ext++) {
-	if (ext->tag == tag) {
-	    func = ext->func;
-	    break;
-	}
+    tval[0] = val;
+    td->type = RPM_INT32_TYPE;
+    td->count = 1;
+    td->data = tval;
+    td->flags = RPMTD_ALLOCED;
+    return 1; /* this cannot fail */
+}
+
+static int dbinstanceTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return numberTag(td, headerGetInstance(h));
+}
+
+static int headercolorTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    rpm_color_t *fcolor, hcolor = 0;
+    struct rpmtd_s fcolors;
+
+    headerGet(h, RPMTAG_FILECOLORS, &fcolors, HEADERGET_MINMEM);
+    while ((fcolor = rpmtdNextUint32(&fcolors)) != NULL) {
+	hcolor |= *fcolor;
     }
-    return func;
+    rpmtdFreeData(&fcolors);
+    hcolor &= 0x0f;
+
+    return numberTag(td, hcolor);
+}
+
+enum nevraFlags_e {
+    NEVRA_NAME		= (1 << 0),
+    NEVRA_EPOCH		= (1 << 1),
+    NEVRA_VERSION	= (1 << 2),
+    NEVRA_RELEASE	= (1 << 3),
+    NEVRA_ARCH		= (1 << 4)
+};
+typedef rpmFlags nevraFlags;
+
+static int getNEVRA(Header h, rpmtd td, nevraFlags flags)
+{
+    const char *val = NULL;
+    char *res = NULL;
+
+    if ((flags & NEVRA_NAME)) {
+	val = headerGetString(h, RPMTAG_NAME);
+	if (val) rstrscat(&res, val, "-", NULL);
+    }
+    if ((flags & NEVRA_EPOCH)) {
+	char *e = headerGetAsString(h, RPMTAG_EPOCH);
+	if (e) rstrscat(&res, e, ":", NULL);
+	free(e);
+    }
+    if ((flags & NEVRA_VERSION)) {
+	val = headerGetString(h, RPMTAG_VERSION);
+	if (val) rstrscat(&res, val, "-", NULL);
+    }
+    if ((flags & NEVRA_RELEASE)) {
+	val = headerGetString(h, RPMTAG_RELEASE);
+	if (val) rstrscat(&res, val, NULL);
+    }
+    if ((flags & NEVRA_ARCH)) {
+	val = headerGetString(h, RPMTAG_ARCH);
+	if (headerIsSource(h) && val == NULL) val = "src";
+	if (val) rstrscat(&res, ".", val, NULL);
+    }
+
+    td->type = RPM_STRING_TYPE;
+    td->data = res;
+    td->count = 1;
+    td->flags = RPMTD_ALLOCED;
+
+    return 1;
+}
+
+static int evrTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return getNEVRA(h, td, NEVRA_EPOCH|NEVRA_VERSION|NEVRA_RELEASE);
+}
+
+static int nvrTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return getNEVRA(h, td, NEVRA_NAME|NEVRA_VERSION|NEVRA_RELEASE);
+}
+
+static int nvraTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return getNEVRA(h, td, NEVRA_NAME|NEVRA_VERSION|NEVRA_RELEASE|NEVRA_ARCH);
+}
+
+static int nevrTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return getNEVRA(h, td, NEVRA_NAME|NEVRA_EPOCH|NEVRA_VERSION|NEVRA_RELEASE);
+}
+
+static int nevraTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return getNEVRA(h, td, NEVRA_NAME|NEVRA_EPOCH|NEVRA_VERSION|NEVRA_RELEASE|NEVRA_ARCH);
+}
+
+static int verboseTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    if (rpmIsVerbose()) {
+	td->type = RPM_INT32_TYPE;
+	td->count = 1;
+	td->data = &(td->count);
+	td->flags = RPMTD_NONE;
+	return 1;
+    } else {
+	return 0;
+    }
+}
+
+static int epochnumTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    /* For consistency, always return malloced data */
+    if (!headerGet(h, RPMTAG_EPOCH, td, HEADERGET_ALLOC)) {
+	uint32_t *e = malloc(sizeof(*e));
+	*e = 0;
+	td->data = e;
+	td->type = RPM_INT32_TYPE;
+	td->count = 1;
+	td->flags = RPMTD_ALLOCED;
+    }
+    td->tag = RPMTAG_EPOCHNUM;
+    return 1;
 }
 
 static const struct headerTagFunc_s rpmHeaderTagExtensions[] = {
@@ -654,15 +671,35 @@ static const struct headerTagFunc_s rpmHeaderTagExtensions[] = {
     { RPMTAG_ORIGFILENAMES,	origfilenamesTag },
     { RPMTAG_FILEPROVIDE,	fileprovideTag },
     { RPMTAG_FILEREQUIRE,	filerequireTag },
-    { RPMTAG_FSNAMES,		fsnamesTag },
-    { RPMTAG_FSSIZES,		fssizesTag },
-    { RPMTAG_INSTALLPREFIX,	instprefixTag },
     { RPMTAG_TRIGGERCONDS,	triggercondsTag },
     { RPMTAG_TRIGGERTYPE,	triggertypeTag },
     { RPMTAG_LONGFILESIZES,	longfilesizesTag },
     { RPMTAG_LONGARCHIVESIZE,	longarchivesizeTag },
     { RPMTAG_LONGSIZE,		longsizeTag },
     { RPMTAG_LONGSIGSIZE,	longsigsizeTag },
+    { RPMTAG_DBINSTANCE,	dbinstanceTag },
+    { RPMTAG_EVR,		evrTag },
+    { RPMTAG_NVR,		nvrTag },
+    { RPMTAG_NEVR,		nevrTag },
+    { RPMTAG_NVRA,		nvraTag },
+    { RPMTAG_NEVRA,		nevraTag },
+    { RPMTAG_HEADERCOLOR,	headercolorTag },
+    { RPMTAG_VERBOSE,		verboseTag },
+    { RPMTAG_EPOCHNUM,		epochnumTag },
     { 0, 			NULL }
 };
 
+headerTagTagFunction rpmHeaderTagFunc(rpmTagVal tag)
+{
+    const struct headerTagFunc_s * ext;
+    headerTagTagFunction func = NULL;
+
+    for (ext = rpmHeaderTagExtensions; ext->func != NULL; ext++) {
+	if (ext->tag == tag) {
+	    func = ext->func;
+	    break;
+	}
+    }
+    return func;
+}
+
