@@ -11,49 +11,305 @@
 #include <rpm/rpmds.h>
 #include <rpm/rpmfileutil.h>
 #include <rpm/rpmstring.h>
-#include <rpm/argv.h>
 
-#include "lib/rpmdb_internal.h"	/* XXX for dbiIndexSetCount */
 #include "lib/fprint.h"
-#include "lib/psm.h"
+#include "lib/misc.h"
+#include "lib/rpmchroot.h"
 #include "lib/rpmlock.h"
 #include "lib/rpmfi_internal.h"	/* only internal apis */
 #include "lib/rpmte_internal.h"	/* only internal apis */
 #include "lib/rpmts_internal.h"
-#include "lib/cpio.h"
 #include "rpmio/rpmhook.h"
 
+/* XXX FIXME: merge with existing (broken?) tests in system.h */
+/* portability fiddles */
+#if STATFS_IN_SYS_STATVFS
+#include <sys/statvfs.h>
+
+#else
+# if STATFS_IN_SYS_VFS
+#  include <sys/vfs.h>
+# else
+#  if STATFS_IN_SYS_MOUNT
+#   include <sys/mount.h>
+#  else
+#   if STATFS_IN_SYS_STATFS
+#    include <sys/statfs.h>
+#   endif
+#  endif
+# endif
+#endif
+
 #include "debug.h"
 
+struct diskspaceInfo_s {
+    char * mntPoint;	/*!< File system mount point */
+    dev_t dev;		/*!< File system device number. */
+    int64_t bneeded;	/*!< No. of blocks needed. */
+    int64_t ineeded;	/*!< No. of inodes needed. */
+    int64_t bsize;	/*!< File system block size. */
+    int64_t bavail;	/*!< No. of blocks available. */
+    int64_t iavail;	/*!< No. of inodes available. */
+    int64_t obneeded;	/*!< Bookkeeping to avoid duplicate reports */
+    int64_t oineeded;	/*!< Bookkeeping to avoid duplicate reports */
+};
 
-/**
- */
-static int archOkay(const char * pkgArch)
+/* Adjust for root only reserved space. On linux e2fs, this is 5%. */
+#define	adj_fs_blocks(_nb)	(((_nb) * 21) / 20)
+#define BLOCK_ROUND(size, block) (((size) + (block) - 1) / (block))
+
+static int rpmtsInitDSI(const rpmts ts)
 {
-    if (pkgArch == NULL) return 0;
-    return (rpmMachineScore(RPM_MACHTABLE_INSTARCH, pkgArch) ? 1 : 0);
+    if (rpmtsFilterFlags(ts) & RPMPROB_FILTER_DISKSPACE)
+	return 0;
+    ts->dsi = _free(ts->dsi);
+    ts->dsi = xcalloc(1, sizeof(*ts->dsi));
+    return 0;
 }
 
-/**
- */
-static int osOkay(const char * pkgOs)
+static rpmDiskSpaceInfo rpmtsCreateDSI(const rpmts ts, dev_t dev,
+				       const char * dirName, int count)
 {
-    if (pkgOs == NULL) return 0;
-    return (rpmMachineScore(RPM_MACHTABLE_INSTOS, pkgOs) ? 1 : 0);
+    rpmDiskSpaceInfo dsi;
+    struct stat sb;
+    char * resolved_path;
+    char mntPoint[PATH_MAX];
+    int rc;
+
+#if STATFS_IN_SYS_STATVFS
+    struct statvfs sfb;
+    memset(&sfb, 0, sizeof(sfb));
+    rc = statvfs(dirName, &sfb);
+#else
+    struct statfs sfb;
+    memset(&sfb, 0, sizeof(sfb));
+#  if STAT_STATFS4
+/* This platform has the 4-argument version of the statfs call.  The last two
+ * should be the size of struct statfs and 0, respectively.  The 0 is the
+ * filesystem type, and is always 0 when statfs is called on a mounted
+ * filesystem, as we're doing.
+ */
+    rc = statfs(dirName, &sfb, sizeof(sfb), 0);
+#  else
+    rc = statfs(dirName, &sfb);
+#  endif
+#endif
+    if (rc)
+	return NULL;
+
+    rc = stat(dirName, &sb);
+    if (rc)
+	return NULL;
+    if (sb.st_dev != dev)
+	return NULL;
+
+    ts->dsi = xrealloc(ts->dsi, (count + 2) * sizeof(*ts->dsi));
+    dsi = ts->dsi + count;
+    memset(dsi, 0, 2 * sizeof(*dsi));
+
+    dsi->dev = sb.st_dev;
+    dsi->bsize = sfb.f_bsize;
+    if (!dsi->bsize)
+	dsi->bsize = 512;       /* we need a bsize */
+    dsi->bneeded = 0;
+    dsi->ineeded = 0;
+#ifdef STATFS_HAS_F_BAVAIL
+    dsi->bavail = (sfb.f_flag & ST_RDONLY) ? 0 : sfb.f_bavail;
+#else
+/* FIXME: the statfs struct doesn't have a member to tell how many blocks are
+ * available for non-superusers.  f_blocks - f_bfree is probably too big, but
+ * it's about all we can do.
+ */
+    dsi->bavail = sfb.f_blocks - sfb.f_bfree;
+#endif
+    /* XXX Avoid FAT and other file systems that have not inodes. */
+    /* XXX assigning negative value to unsigned type */
+    dsi->iavail = !(sfb.f_ffree == 0 && sfb.f_files == 0)
+	? sfb.f_ffree : -1;
+
+    /* Find mount point belonging to this device number */
+    resolved_path = realpath(dirName, mntPoint);
+    if (!resolved_path) {
+	strncpy(mntPoint, dirName, PATH_MAX);
+	mntPoint[PATH_MAX-1] = '\0';
+    }
+    char * end = NULL;
+    while (end != mntPoint) {
+	end = strrchr(mntPoint, '/');
+	if (end == mntPoint) { /* reached "/" */
+	    stat("/", &sb);
+	    if (dsi->dev != sb.st_dev) {
+		dsi->mntPoint = xstrdup(mntPoint);
+	    } else {
+		dsi->mntPoint = xstrdup("/");
+	    }
+	    break;
+	} else if (end) {
+	    *end = '\0';
+	} else { /* dirName doesn't start with / - should not happen */
+	    dsi->mntPoint = xstrdup(dirName);
+	    break;
+	}
+	stat(mntPoint, &sb);
+	if (dsi->dev != sb.st_dev) {
+	    *end = '/';
+	    dsi->mntPoint = xstrdup(mntPoint);
+	    break;
+	}
+    }
+
+    rpmlog(RPMLOG_DEBUG,
+	   "0x%08x %8" PRId64 " %12" PRId64 " %12" PRId64" %s\n",
+	   (unsigned) dsi->dev, dsi->bsize,
+	   dsi->bavail, dsi->iavail,
+	   dsi->mntPoint);
+    return dsi;
 }
 
+static rpmDiskSpaceInfo rpmtsGetDSI(const rpmts ts, dev_t dev,
+				    const char *dirName) {
+    rpmDiskSpaceInfo dsi;
+    dsi = ts->dsi;
+    if (dsi) {
+	while (dsi->bsize && dsi->dev != dev)
+	    dsi++;
+	if (dsi->bsize == 0) {
+	    /* create new entry */
+	    dsi = rpmtsCreateDSI(ts, dev, dirName, dsi - ts->dsi);
+	}
+    }
+    return dsi;
+}
+
+static void rpmtsUpdateDSI(const rpmts ts, dev_t dev, const char *dirName,
+		rpm_loff_t fileSize, rpm_loff_t prevSize, rpm_loff_t fixupSize,
+		rpmFileAction action)
+{
+    int64_t bneeded;
+    rpmDiskSpaceInfo dsi = rpmtsGetDSI(ts, dev, dirName);
+    if (dsi == NULL)
+	return;
+
+    bneeded = BLOCK_ROUND(fileSize, dsi->bsize);
+
+    switch (action) {
+    case FA_BACKUP:
+    case FA_SAVE:
+    case FA_ALTNAME:
+	dsi->ineeded++;
+	dsi->bneeded += bneeded;
+	break;
+
+    /*
+     * FIXME: If two packages share a file (same md5sum), and
+     * that file is being replaced on disk, will dsi->bneeded get
+     * adjusted twice? Quite probably!
+     */
+    case FA_CREATE:
+	dsi->bneeded += bneeded;
+	dsi->bneeded -= BLOCK_ROUND(prevSize, dsi->bsize);
+	break;
+
+    case FA_ERASE:
+	dsi->ineeded--;
+	dsi->bneeded -= bneeded;
+	break;
+
+    default:
+	break;
+    }
+
+    if (fixupSize)
+	dsi->bneeded -= BLOCK_ROUND(fixupSize, dsi->bsize);
+
+    /* adjust bookkeeping when requirements shrink */
+    if (dsi->bneeded < dsi->obneeded) dsi->obneeded = dsi->bneeded;
+    if (dsi->ineeded < dsi->oineeded) dsi->oineeded = dsi->ineeded;
+}
+
+/* return DSI of the device the rpmdb lives on */
+static rpmDiskSpaceInfo rpmtsDbDSI(const rpmts ts) {
+    const char *dbhome = rpmdbHome(rpmtsGetRdb(ts));
+    struct stat sb;
+    int rc;
+
+    rc = stat(dbhome, &sb);
+    if (rc) {
+	return NULL;
+    }
+    return rpmtsGetDSI(ts, sb.st_dev, dbhome);
+}
+
+/* Update DSI for changing size of the rpmdb */
+static void rpmtsUpdateDSIrpmDBSize(const rpmte p,
+				    rpmDiskSpaceInfo dsi) {
+    rpm_loff_t headerSize;
+    int64_t bneeded;
+
+    /* XXX somehow we can end up here with bsize 0 (RhBug:671056) */
+    if (dsi == NULL || dsi->bsize == 0) return;
+
+    headerSize = rpmteHeaderSize(p);
+    bneeded = BLOCK_ROUND(headerSize, dsi->bsize);
+    /* REMOVE doesn't neccessarily shrink the database */
+    if (rpmteType(p) == TR_ADDED) {
+	/* guessing that db grows 4 times more than the header size */
+	dsi->bneeded += (bneeded * 4);
+    }
+}
+
+
+static void rpmtsCheckDSIProblems(const rpmts ts, const rpmte te)
+{
+    rpmDiskSpaceInfo dsi = ts->dsi;
+
+    if (dsi == NULL || !dsi->bsize)
+	return;
+    if (rpmfiFC(rpmteFI(te)) <= 0)
+	return;
+
+    for (; dsi->bsize; dsi++) {
+
+	if (dsi->bavail >= 0 && adj_fs_blocks(dsi->bneeded) > dsi->bavail) {
+	    if (dsi->bneeded > dsi->obneeded) {
+		rpmteAddProblem(te, RPMPROB_DISKSPACE, NULL, dsi->mntPoint,
+		   (adj_fs_blocks(dsi->bneeded) - dsi->bavail) * dsi->bsize);
+		dsi->obneeded = dsi->bneeded;
+	    }
+	}
+
+	if (dsi->iavail >= 0 && adj_fs_blocks(dsi->ineeded) > dsi->iavail) {
+	    if (dsi->ineeded > dsi->oineeded) {
+		rpmteAddProblem(te, RPMPROB_DISKNODES, NULL, dsi->mntPoint,
+			(adj_fs_blocks(dsi->ineeded) - dsi->iavail));
+		dsi->oineeded = dsi->ineeded;
+	    }
+	}
+    }
+}
+
+static void rpmtsFreeDSI(rpmts ts){
+    rpmDiskSpaceInfo dsi;
+    if (ts == NULL)
+	return;
+    dsi = ts->dsi;
+    while (dsi && dsi->bsize != 0) {
+	dsi->mntPoint = _free(dsi->mntPoint);
+	dsi++;
+    }
+
+    ts->dsi = _free(ts->dsi);
+}
+
+
 /* Calculate total number of files involved in transaction */
 static uint64_t countFiles(rpmts ts)
 {
     uint64_t fc = 0;
     rpmtsi pi = rpmtsiInit(ts);
     rpmte p;
-    rpmfi fi;
-    while ((p = rpmtsiNext(pi, 0)) != NULL) {
-	if ((fi = rpmteFI(p)) == NULL)
-	    continue;   /* XXX can't happen */
-	fc += rpmfiFC(fi);
-    }
+    while ((p = rpmtsiNext(pi, 0)) != NULL)
+	fc += rpmfiFC(rpmteFI(p));
     pi = rpmtsiFree(pi);
     return fc;
 }
@@ -88,7 +344,7 @@ static int handleInstInstalledFile(const rpmts ts, rpmte p, rpmfi fi,
 
 	rConflicts = !(beingRemoved || (rpmtsFilterFlags(ts) & RPMPROB_FILTER_REPLACEOLDFILES));
 	/* Resolve file conflicts to prefer Elf64 (if not forced). */
-	if (tscolor != 0 && FColor != 0 && FColor != oFColor) {
+	if (tscolor != 0 && FColor != 0 && oFColor != 0 && FColor != oFColor) {
 	    if (oFColor & prefcolor) {
 		rpmfsSetAction(fs, fx, FA_SKIPCOLOR);
 		rConflicts = 0;
@@ -99,15 +355,10 @@ static int handleInstInstalledFile(const rpmts ts, rpmte p, rpmfi fi,
 	}
 
 	if (rConflicts) {
-	    char *altNEVR = headerGetNEVRA(otherHeader, NULL);
-	    rpmps ps = rpmtsProblems(ts);
-	    rpmpsAppend(ps, RPMPROB_FILE_CONFLICT,
-			rpmteNEVRA(p), rpmteKey(p),
-			rpmfiDN(fi), rpmfiBN(fi),
-			altNEVR,
-			0);
+	    char *altNEVR = headerGetAsString(otherHeader, RPMTAG_NEVRA);
+	    rpmteAddProblem(p, RPMPROB_FILE_CONFLICT, altNEVR, rpmfiFN(fi),
+			    headerGetInstance(otherHeader));
 	    free(altNEVR);
-	    rpmpsFree(ps);
 	}
 
 	/* Save file identifier to mark as state REPLACED. */
@@ -137,7 +388,6 @@ static int handleInstInstalledFile(const rpmts ts, rpmte p, rpmfi fi,
 static void handleOverlappedFiles(rpmts ts, rpmFpHash ht, rpmte p, rpmfi fi)
 {
     rpm_loff_t fixupSize = 0;
-    rpmps ps;
     const char * fn;
     int i, j;
     rpm_color_t tscolor = rpmtsColor(ts);
@@ -145,9 +395,7 @@ static void handleOverlappedFiles(rpmts ts, rpmFpHash ht, rpmte p, rpmfi fi)
     rpmfs fs = rpmteGetFileStates(p);
     rpmfs otherFs;
 
-    ps = rpmtsProblems(ts);
     fi = rpmfiInit(fi, 0);
-    if (fi != NULL)
     while ((i = rpmfiNext(fi)) >= 0) {
 	rpm_color_t oFColor, FColor;
 	struct fingerPrint_s * fiFps;
@@ -260,15 +508,11 @@ assert(otherFi != NULL);
 
 		rConflicts = reportConflicts;
 		/* Resolve file conflicts to prefer Elf64 (if not forced) ... */
-		if (tscolor != 0) {
+		if (tscolor != 0 && FColor != 0 && oFColor != 0 && FColor != oFColor) {
 		    if (FColor & prefcolor) {
 			/* ... last file of preferred colour is installed ... */
-			if (!XFA_SKIPPING(rpmfsGetAction(fs, i))) {
-			    /* XXX static helpers are order dependent. Ick. */
-			    if (strcmp(fn, "/usr/sbin/libgcc_post_upgrade")
-			     && strcmp(fn, "/usr/sbin/glibc_post_upgrade"))
-				rpmfsSetAction(otherFs, otherFileNum, FA_SKIPCOLOR);
-			}
+			if (!XFA_SKIPPING(rpmfsGetAction(fs, i)))
+			    rpmfsSetAction(otherFs, otherFileNum, FA_SKIPCOLOR);
 			rpmfsSetAction(fs, i, FA_CREATE);
 			rConflicts = 0;
 		    } else
@@ -282,11 +526,8 @@ assert(otherFi != NULL);
 		    done = 1;
 		}
 		if (rConflicts) {
-		    rpmpsAppend(ps, RPMPROB_NEW_FILE_CONFLICT,
-			rpmteNEVRA(p), rpmteKey(p),
-			fn, NULL,
-			rpmteNEVRA(otherTe),
-			0);
+		    rpmteAddProblem(p, RPMPROB_NEW_FILE_CONFLICT,
+				    rpmteNEVRA(otherTe), fn, 0);
 		}
 	    }
 
@@ -326,7 +567,7 @@ assert(otherFi != NULL);
 	    }
 		
 	    /* Here is a pre-existing modified config file that needs saving. */
-	    {	pgpHashAlgo algo = 0;
+	    {	int algo = 0;
 		size_t diglen = 0;
 		const unsigned char *digest;
 		if ((digest = rpmfiFDigest(fi, &algo, &diglen))) {
@@ -343,11 +584,11 @@ assert(otherFi != NULL);
 	}
 
 	/* Update disk space info for a file. */
-	rpmtsUpdateDSI(ts, fiFps->entry->dev, rpmfiFSize(fi),
-		       rpmfiFReplacedSize(fi), fixupSize, rpmfsGetAction(fs, i));
+	rpmtsUpdateDSI(ts, fiFps->entry->dev, fiFps->entry->dirName,
+		       rpmfiFSize(fi), rpmfiFReplacedSize(fi),
+		       fixupSize, rpmfsGetAction(fs, i));
 
     }
-    ps = rpmpsFree(ps);
 }
 
 /**
@@ -355,50 +596,103 @@ assert(otherFi != NULL);
  * @param p		current transaction element
  * @param h		installed header
  * @param ps		problem set
- * @return		0 if not newer, 1 if okay
  */
-static int ensureOlder(const rpmte p, const Header h, rpmps ps)
+static void ensureOlder(const rpmte p, const Header h)
 {
     rpmsenseFlags reqFlags = (RPMSENSE_LESS | RPMSENSE_EQUAL);
     rpmds req;
-    int rc;
-
-    if (p == NULL || h == NULL)
-	return 1;
 
     req = rpmdsSingle(RPMTAG_REQUIRENAME, rpmteN(p), rpmteEVR(p), reqFlags);
-    rc = rpmdsNVRMatchesDep(h, req, _rpmds_nopromote);
-    req = rpmdsFree(req);
-
-    if (rc == 0) {
-	char * altNEVR = headerGetNEVRA(h, NULL);
-	rpmpsAppend(ps, RPMPROB_OLDPACKAGE,
-		rpmteNEVRA(p), rpmteKey(p),
-		NULL, NULL,
-		altNEVR,
-		0);
-	altNEVR = _free(altNEVR);
-	rc = 1;
-    } else
-	rc = 0;
-
-    return rc;
+    if (rpmdsNVRMatchesDep(h, req, _rpmds_nopromote) == 0) {
+	char * altNEVR = headerGetAsString(h, RPMTAG_NEVRA);
+	rpmteAddProblem(p, RPMPROB_OLDPACKAGE, altNEVR, NULL,
+			headerGetInstance(h));
+	free(altNEVR);
+    }
+    rpmdsFree(req);
 }
 
+/**
+ * Check if the curent file in the file iterator is in the
+ * netshardpath and though should be excluded.
+ * @param ts            transaction set
+ * @param fi            file info set
+ * @returns pointer to matching path or NULL
+ */
+static char ** matchNetsharedpath(const rpmts ts, rpmfi fi)
+{
+    char ** nsp;
+    const char * dn, * bn;
+    size_t dnlen, bnlen;
+    char * s;
+    bn = rpmfiBN(fi);
+    bnlen = strlen(bn);
+    dn = rpmfiDN(fi);
+    dnlen = strlen(dn);
+    for (nsp = ts->netsharedPaths; nsp && *nsp; nsp++) {
+	size_t len;
+
+	len = strlen(*nsp);
+	if (dnlen >= len) {
+	    if (!rstreqn(dn, *nsp, len))
+		continue;
+	    /* Only directories or complete file paths can be net shared */
+	    if (!(dn[len] == '/' || dn[len] == '\0'))
+		continue;
+	} else {
+	    if (len < (dnlen + bnlen))
+		continue;
+	    if (!rstreqn(dn, *nsp, dnlen))
+		continue;
+	    /* Insure that only the netsharedpath basename is compared. */
+	    if ((s = strchr((*nsp) + dnlen, '/')) != NULL && s[1] != '\0')
+		continue;
+	    if (!rstreqn(bn, (*nsp) + dnlen, bnlen))
+		continue;
+	    len = dnlen + bnlen;
+	    /* Only directories or complete file paths can be net shared */
+	    if (!((*nsp)[len] == '/' || (*nsp)[len] == '\0'))
+		continue;
+	}
+
+	break;
+    }
+    return nsp;
+}
+
+static void skipEraseFiles(const rpmts ts, rpmte p)
+{
+    rpmfi fi = rpmteFI(p);
+    rpmfs fs = rpmteGetFileStates(p);
+    int i;
+    char ** nsp;
+    /*
+     * Skip net shared paths.
+     * Net shared paths are not relative to the current root (though
+     * they do need to take package relocations into account).
+     */
+    fi = rpmfiInit(fi, 0);
+    while ((i = rpmfiNext(fi)) >= 0)
+    {
+	nsp = matchNetsharedpath(ts, fi);
+	if (nsp && *nsp) {
+	    rpmfsSetAction(fs, i, FA_SKIPNETSHARED);
+	}
+    }
+}
+
+
 /**
  * Skip any files that do not match install policies.
  * @param ts		transaction set
  * @param fi		file info set
  */
-static void skipFiles(const rpmts ts, rpmte p)
+static void skipInstallFiles(const rpmts ts, rpmte p)
 {
     rpm_color_t tscolor = rpmtsColor(ts);
     rpm_color_t FColor;
     int noConfigs = (rpmtsFlags(ts) & RPMTRANS_FLAG_NOCONFIGS);
     int noDocs = (rpmtsFlags(ts) & RPMTRANS_FLAG_NODOCS);
-    const char * dn, * bn;
-    size_t dnlen, bnlen;
-    char * s;
     int * drc;
     char * dff;
     int dc;
@@ -415,20 +709,11 @@ static void skipFiles(const rpmts ts, rpmte p)
     dff = xcalloc(dc, sizeof(*dff));
 
     fi = rpmfiInit(fi, 0);
-    if (fi != NULL)	/* XXX lclint */
-    while ((i = rpmfiNext(fi)) >= 0)
-    {
+    while ((i = rpmfiNext(fi)) >= 0) {
 	char ** nsp;
 	const char *flangs;
 
-	bn = rpmfiBN(fi);
-	bnlen = strlen(bn);
 	ix = rpmfiDX(fi);
-	dn = rpmfiDN(fi);
-	dnlen = strlen(dn);
-	if (dn == NULL)
-	    continue;	/* XXX can't happen */
-
 	drc[ix]++;
 
 	/* Don't bother with skipped files */
@@ -450,35 +735,7 @@ static void skipFiles(const rpmts ts, rpmte p)
 	 * Net shared paths are not relative to the current root (though
 	 * they do need to take package relocations into account).
 	 */
-	for (nsp = ts->netsharedPaths; nsp && *nsp; nsp++) {
-	    size_t len;
-
-	    len = strlen(*nsp);
-	    if (dnlen >= len) {
-		if (strncmp(dn, *nsp, len))
-		    continue;
-		/* Only directories or complete file paths can be net shared */
-		if (!(dn[len] == '/' || dn[len] == '\0'))
-		    continue;
-	    } else {
-		if (len < (dnlen + bnlen))
-		    continue;
-		if (strncmp(dn, *nsp, dnlen))
-		    continue;
-		/* Insure that only the netsharedpath basename is compared. */
-		if ((s = strchr((*nsp) + dnlen, '/')) != NULL && s[1] != '\0')
-		    continue;
-		if (strncmp(bn, (*nsp) + dnlen, bnlen))
-		    continue;
-		len = dnlen + bnlen;
-		/* Only directories or complete file paths can be net shared */
-		if (!((*nsp)[len] == '/' || (*nsp)[len] == '\0'))
-		    continue;
-	    }
-
-	    break;
-	}
-
+	nsp = matchNetsharedpath(ts, fi);
 	if (nsp && *nsp) {
 	    drc[ix]--;	dff[ix] = 1;
 	    rpmfsSetAction(fs, i, FA_SKIPNETSHARED);
@@ -496,7 +753,7 @@ static void skipFiles(const rpmts ts, rpmte p)
 		for (l = flangs; *l != '\0'; l = le) {
 		    for (le = l; *le != '\0' && *le != '|'; le++)
 			{};
-		    if ((le-l) > 0 && !strncmp(*lang, l, (le-l)))
+		    if ((le-l) > 0 && rstreqn(*lang, l, (le-l)))
 			break;
 		    if (*le == '|') le++;	/* skip over | */
 		}
@@ -530,14 +787,9 @@ static void skipFiles(const rpmts ts, rpmte p)
     }
 
     /* Skip (now empty) directories that had skipped files. */
-#ifndef	NOTYET
-    if (fi != NULL)	/* XXX can't happen */
-    for (j = 0; j < dc; j++)
-#else
-    if ((fi = rpmfiInitD(fi)) != NULL)
-    while (j = rpmfiNextD(fi) >= 0)
-#endif
-    {
+    for (j = 0; j < dc; j++) {
+	const char * dn, * bn;
+	size_t dnlen, bnlen;
 
 	if (drc[j]) continue;	/* dir still has files. */
 	if (!dff[j]) continue;	/* dir was not emptied here. */
@@ -553,7 +805,6 @@ static void skipFiles(const rpmts ts, rpmte p)
 
 	/* If explicitly included in the package, skip the directory. */
 	fi = rpmfiInit(fi, 0);
-	if (fi != NULL)		/* XXX lclint */
 	while ((i = rpmfiNext(fi)) >= 0) {
 	    const char * fdn, * fbn;
 	    rpm_mode_t fFMode;
@@ -568,12 +819,12 @@ static void skipFiles(const rpmts ts, rpmte p)
 	    fdn = rpmfiDN(fi);
 	    if (strlen(fdn) != dnlen)
 		continue;
-	    if (strncmp(fdn, dn, dnlen))
+	    if (!rstreqn(fdn, dn, dnlen))
 		continue;
 	    fbn = rpmfiBN(fi);
 	    if (strlen(fbn) != bnlen)
 		continue;
-	    if (strncmp(fbn, bn, bnlen))
+	    if (!rstreqn(fbn, bn, bnlen))
 		continue;
 	    rpmlog(RPMLOG_DEBUG, "excluding directory %s\n", dn);
 	    rpmfsSetAction(fs, i, FA_SKIPNSTATE);
@@ -604,29 +855,28 @@ static void skipFiles(const rpmts ts, rpmte p)
 static
 rpmdbMatchIterator rpmFindBaseNamesInDB(rpmts ts, uint64_t fileCount)
 {
+    tsMembers tsmem = rpmtsMembers(ts);
     rpmtsi pi;  rpmte p;
     rpmfi fi;
     rpmdbMatchIterator mi;
-    int i, xx;
+    int xx;
+    int oc = 0;
     const char * baseName;
 
     rpmStringSet baseNames = rpmStringSetCreate(fileCount, 
 					hashFunctionString, strcmp, NULL);
 
-    mi = rpmdbInitIterator(rpmtsGetRdb(ts), RPMTAG_BASENAMES, NULL, 0);
+    mi = rpmdbNewIterator(rpmtsGetRdb(ts), RPMDBI_BASENAMES);
 
     pi = rpmtsiInit(ts);
     while ((p = rpmtsiNext(pi, 0)) != NULL) {
 	(void) rpmdbCheckSignals();
 
-	if ((fi = rpmteFI(p)) == NULL)
-	    continue;   /* XXX can't happen */
-	rpmtsNotify(ts, NULL, RPMCALLBACK_TRANS_PROGRESS, rpmtsiOc(pi),
-		    ts->orderCount);
+	rpmtsNotify(ts, NULL, RPMCALLBACK_TRANS_PROGRESS, oc++, tsmem->orderCount);
 
 	/* Gather all installed headers with matching basename's. */
-	fi = rpmfiInit(fi, 0);
-	while ((i = rpmfiNext(fi)) >= 0) {
+	fi = rpmfiInit(rpmteFI(p), 0);
+	while (rpmfiNext(fi) >= 0) {
 	    size_t keylen;
 	    baseName = rpmfiBN(fi);
 	    if (rpmStringSetHasEntry(baseNames, baseName))
@@ -656,6 +906,7 @@ rpmdbMatchIterator rpmFindBaseNamesInDB(rpmts ts, uint64_t fileCount)
 static
 void checkInstalledFiles(rpmts ts, uint64_t fileCount, rpmFpHash ht, fingerPrintCache fpc)
 {
+    tsMembers tsmem = rpmtsMembers(ts);
     rpmte p;
     rpmfi fi;
     rpmfs fs;
@@ -693,14 +944,7 @@ void checkInstalledFiles(rpmts ts, uint64_t fileCount, rpmFpHash ht, fingerPrint
 
 	/* Is this package being removed? */
 	installedPkg = rpmdbGetIteratorOffset(mi);
-	beingRemoved = 0;
-	if (ts->removedPackages != NULL)
-	for (j = 0; j < ts->numRemovedPackages; j++) {
-	    if (ts->removedPackages[j] != installedPkg)
-	        continue;
-	    beingRemoved = 1;
-	    break;
-	}
+	beingRemoved = intHashHasEntry(tsmem->removedPackages, installedPkg);
 
 	h = headerLink(h);
 	headerGet(h, RPMTAG_BASENAMES, &bnames, hgflags);
@@ -778,6 +1022,9 @@ void checkInstalledFiles(rpmts ts, uint64_t fileCount, rpmFpHash ht, fingerPrint
     mi = rpmdbFreeIterator(mi);
 }
 
+#define badArch(_a) (rpmMachineScore(RPM_MACHTABLE_INSTARCH, (_a)) == 0)
+#define badOs(_a) (rpmMachineScore(RPM_MACHTABLE_INSTOS, (_a)) == 0)
+
 /*
  * For packages being installed:
  * - verify package arch/os.
@@ -786,7 +1033,7 @@ void checkInstalledFiles(rpmts ts, uint64_t fileCount, rpmFpHash ht, fingerPrint
 static rpmps checkProblems(rpmts ts)
 {
     rpm_color_t tscolor = rpmtsColor(ts);
-    rpmps ps = rpmpsCreate();
+    rpmprobFilterFlags probFilter = rpmtsFilterFlags(ts);
     rpmtsi pi = rpmtsiInit(ts);
     rpmte p;
 
@@ -796,30 +1043,23 @@ static rpmps checkProblems(rpmts ts)
     while ((p = rpmtsiNext(pi, TR_ADDED)) != NULL) {
 	rpmdbMatchIterator mi;
 
-	if (!(rpmtsFilterFlags(ts) & RPMPROB_FILTER_IGNOREARCH))
-	    if (!archOkay(rpmteA(p)))
-		rpmpsAppend(ps, RPMPROB_BADARCH,
-			rpmteNEVRA(p), rpmteKey(p),
-			rpmteA(p), NULL,
-			NULL, 0);
+	if (!(probFilter & RPMPROB_FILTER_IGNOREARCH) && badArch(rpmteA(p)))
+	    rpmteAddProblem(p, RPMPROB_BADARCH, rpmteA(p), NULL, 0);
 
-	if (!(rpmtsFilterFlags(ts) & RPMPROB_FILTER_IGNOREOS))
-	    if (!osOkay(rpmteO(p)))
-		rpmpsAppend(ps, RPMPROB_BADOS,
-			rpmteNEVRA(p), rpmteKey(p),
-			rpmteO(p), NULL,
-			NULL, 0);
+	if (!(probFilter & RPMPROB_FILTER_IGNOREOS) && badOs(rpmteO(p)))
+	    rpmteAddProblem(p, RPMPROB_BADOS, rpmteO(p), NULL, 0);
 
-	if (!(rpmtsFilterFlags(ts) & RPMPROB_FILTER_OLDPACKAGE)) {
+	if (!(probFilter & RPMPROB_FILTER_OLDPACKAGE)) {
 	    Header h;
-	    mi = rpmtsInitIterator(ts, RPMTAG_NAME, rpmteN(p), 0);
+	    mi = rpmtsInitIterator(ts, RPMDBI_NAME, rpmteN(p), 0);
 	    while ((h = rpmdbNextIterator(mi)) != NULL)
-		ensureOlder(p, h, ps);
+		ensureOlder(p, h);
 	    mi = rpmdbFreeIterator(mi);
 	}
 
-	if (!(rpmtsFilterFlags(ts) & RPMPROB_FILTER_REPLACEPKG)) {
-	    mi = rpmtsInitIterator(ts, RPMTAG_NAME, rpmteN(p), 0);
+	if (!(probFilter & RPMPROB_FILTER_REPLACEPKG)) {
+	    Header h;
+	    mi = rpmtsInitIterator(ts, RPMDBI_NAME, rpmteN(p), 0);
 	    rpmdbSetIteratorRE(mi, RPMTAG_EPOCH, RPMMIRE_STRCMP, rpmteE(p));
 	    rpmdbSetIteratorRE(mi, RPMTAG_VERSION, RPMMIRE_STRCMP, rpmteV(p));
 	    rpmdbSetIteratorRE(mi, RPMTAG_RELEASE, RPMMIRE_STRCMP, rpmteR(p));
@@ -828,56 +1068,112 @@ static rpmps checkProblems(rpmts ts)
 		rpmdbSetIteratorRE(mi, RPMTAG_OS, RPMMIRE_STRCMP, rpmteO(p));
 	    }
 
-	    while (rpmdbNextIterator(mi) != NULL) {
-		rpmpsAppend(ps, RPMPROB_PKG_INSTALLED,
-			rpmteNEVRA(p), rpmteKey(p),
-			NULL, NULL,
-			NULL, 0);
-		break;
+	    if ((h = rpmdbNextIterator(mi)) != NULL) {
+		rpmteAddProblem(p, RPMPROB_PKG_INSTALLED, NULL, NULL,
+				headerGetInstance(h));
 	    }
 	    mi = rpmdbFreeIterator(mi);
 	}
     }
     pi = rpmtsiFree(pi);
-    return ps;
+    return rpmtsProblems(ts);
 }
 
 /*
  * Run pre/post transaction scripts for transaction set
  * param ts	Transaction set
- * param stag	RPMTAG_PRETRANS or RPMTAG_POSTTRANS
- * return	0 on success, -1 on error (invalid script tag)
+ * param goal	PKG_PRETRANS/PKG_POSTTRANS
+ * return	0 on success
  */
-static int runTransScripts(rpmts ts, rpmTag stag) 
+static int runTransScripts(rpmts ts, pkgGoal goal) 
 {
-    rpmtsi pi; 
     rpmte p;
-    rpmpsm psm;
-    rpmTag progtag = RPMTAG_NOT_FOUND;
-    int xx;
-
-    if (stag == RPMTAG_PRETRANS) {
-	progtag = RPMTAG_PRETRANSPROG;
-    } else if (stag == RPMTAG_POSTTRANS) {
-	progtag = RPMTAG_POSTTRANSPROG;
-    } else {
-	return -1;
-    }
-
-    pi = rpmtsiInit(ts);
+    rpmtsi pi = rpmtsiInit(ts);
     while ((p = rpmtsiNext(pi, TR_ADDED)) != NULL) {
-	/* Skip failed elements & those without pre/posttrans */
-	if (!rpmteHaveTransScript(p, stag) || rpmteFailed(p))
-	    continue;
+	rpmteProcess(p, goal);
+    }
+    pi = rpmtsiFree(pi);
+    return 0; /* what to do about failures? */
+}
 
-    	if (rpmteOpen(p, ts, 0)) {
-	    psm = rpmpsmNew(ts, p);
-	    xx = rpmpsmScriptStage(psm, stag, progtag);
-	    psm = rpmpsmFree(psm);
-	    rpmteClose(p, ts, 0);
+static int rpmtsSetupCollections(rpmts ts)
+{
+    /* seenCollectionsPost and TEs are basically a key-value pair. each item in
+     * seenCollectionsPost is a collection that has been seen from any package,
+     * and the associated index in the TEs is the last transaction element
+     * where that collection was seen. */
+    ARGV_t seenCollectionsPost = NULL;
+    rpmte *TEs = NULL;
+    int numSeenPost = 0;
+
+    /* seenCollectionsPre is a list of collections that have been seen from
+     * only removed packages */
+    ARGV_t seenCollectionsPre = NULL;
+    int numSeenPre = 0;
+
+    ARGV_const_t collname;
+    int installing = 1;
+    int i;
+
+    rpmte p;
+    rpmtsi pi = rpmtsiInit(ts);
+    while ((p = rpmtsiNext(pi, 0)) != NULL) {
+	/* detect when we switch from installing to removing packages, and
+	 * update the lastInCollectionAdd lists */
+	if (installing && rpmteType(p) == TR_REMOVED) {
+	    installing = 0;
+	    for (i = 0; i < numSeenPost; i++) {
+		rpmteAddToLastInCollectionAdd(TEs[i], seenCollectionsPost[i]);
+	    }
+	}
+
+	rpmteSetupCollectionPlugins(p);
+
+	for (collname = rpmteCollections(p); collname && *collname; collname++) {
+	    /* figure out if we've seen this collection in post before */
+	    for (i = 0; i < numSeenPost && strcmp(*collname, seenCollectionsPost[i]); i++) {
+	    }
+	    if (i < numSeenPost) {
+		/* we've seen the collection, update the index */
+		TEs[i] = p;
+	    } else {
+		/* haven't seen the collection yet, add it */
+		argvAdd(&seenCollectionsPost, *collname);
+		TEs = xrealloc(TEs, sizeof(*TEs) * (numSeenPost + 1));
+		TEs[numSeenPost] = p;
+		numSeenPost++;
+	    }
+
+	    /* figure out if we've seen this collection in pre remove before */
+	    if (installing == 0) {
+		for (i = 0; i < numSeenPre && strcmp(*collname, seenCollectionsPre[i]); i++) {
+		}
+		if (i >= numSeenPre) {
+		    /* haven't seen this collection, add it */
+		    rpmteAddToFirstInCollectionRemove(p, *collname);
+		    argvAdd(&seenCollectionsPre, *collname);
+		    numSeenPre++;
+		}
+	    }
 	}
     }
     pi = rpmtsiFree(pi);
+
+    /* we've looked at all the rpmte's, update the lastInCollectionAny lists */
+    for (i = 0; i < numSeenPost; i++) {
+	rpmteAddToLastInCollectionAny(TEs[i], seenCollectionsPost[i]);
+	if (installing == 1) {
+	    /* lastInCollectionAdd is only updated above if packages were
+	     * removed. if nothing is removed in the transaction, we need to
+	     * update that list here */
+	    rpmteAddToLastInCollectionAdd(TEs[i], seenCollectionsPost[i]);
+	}
+    }
+
+    argvFree(seenCollectionsPost);
+    argvFree(seenCollectionsPre);
+    _free(TEs);
+
     return 0;
 }
 
@@ -902,7 +1198,6 @@ static void addFingerprints(rpmts ts, uint64_t fileCount, rpmFpHash ht, fingerPr
 	rpmfiFpLookup(fi, fpc);
 	/* collect symbolic links */
  	fi = rpmfiInit(fi, 0);
- 	if (fi != NULL)		/* XXX lclint */
 	while ((i = rpmfiNext(fi)) >= 0) {
 	    struct rpmffi_s ffi;
 	    char const *linktarget;
@@ -929,11 +1224,8 @@ static void addFingerprints(rpmts ts, uint64_t fileCount, rpmFpHash ht, fingerPr
     while ((p = rpmtsiNext(pi, 0)) != NULL) {
 	(void) rpmdbCheckSignals();
 
-	if ((fi = rpmteFI(p)) == NULL)
-	    continue;	/* XXX can't happen */
-	fi = rpmfiInit(fi, 0);
+	fi = rpmfiInit(rpmteFI(p), 0);
 	(void) rpmswEnter(rpmtsOp(ts, RPMTS_OP_FINGERPRINT), 0);
-	if (fi != NULL)		/* XXX lclint */
 	while ((i = rpmfiNext(fi)) >= 0) {
 	    if (XFA_SKIPPING(rpmfsGetAction(rpmteGetFileStates(p), i)))
 		continue;
@@ -956,36 +1248,28 @@ static int rpmtsSetup(rpmts ts, rpmprobFilterFlags ignoreSet)
     if (rpmtsFlags(ts) & RPMTRANS_FLAG_NOTRIGGERS)
 	(void) rpmtsSetFlags(ts, (rpmtsFlags(ts) | _noTransTriggers));
 
-    if (rpmtsFlags(ts) & RPMTRANS_FLAG_JUSTDB)
-	(void) rpmtsSetFlags(ts, (rpmtsFlags(ts) | _noTransScripts | _noTransTriggers));
+    if (rpmtsFlags(ts) & (RPMTRANS_FLAG_JUSTDB | RPMTRANS_FLAG_TEST))
+	(void) rpmtsSetFlags(ts, (rpmtsFlags(ts) | _noTransScripts | _noTransTriggers | RPMTRANS_FLAG_NOCOLLECTIONS));
 
-    /* if SELinux isn't enabled or init fails, don't bother... */
-    if (!rpmtsSELinuxEnabled(ts)) {
+    /* if SELinux isn't enabled, init fails or test run, don't bother... */
+    if (!is_selinux_enabled() || (rpmtsFlags(ts) & RPMTRANS_FLAG_TEST)) {
         rpmtsSetFlags(ts, (rpmtsFlags(ts) | RPMTRANS_FLAG_NOCONTEXTS));
     }
 
     if (!(rpmtsFlags(ts) & RPMTRANS_FLAG_NOCONTEXTS)) {
-	char *fn = rpmGetPath("%{?_install_file_context_path}", NULL);
-	if (matchpathcon_init(fn) == -1) {
-	    rpmtsSetFlags(ts, (rpmtsFlags(ts) | RPMTRANS_FLAG_NOCONTEXTS));
-	}
-	free(fn);
+	rpmtsSELabelInit(ts, selinux_file_context_path());
     }
 
-    /* XXX Make sure the database is open RDWR for package install/erase. */
-    if (rpmtsOpenDB(ts, dbmode)) {
-	return -1;	/* XXX W2DO? */
-    }
+    /* 
+     * Make sure the database is open RDWR for package install/erase.
+     * Note that we initialize chroot state here even if it's just "/" as
+     * this ensures we can successfully perform open(".") which is
+     * required to reliably restore cwd after Lua scripts.
+     */ 
+    if (rpmtsOpenDB(ts, dbmode) || rpmChrootSet(rpmtsRootDir(ts)))
+	return -1;
 
     ts->ignoreSet = ignoreSet;
-    ts->probs = rpmpsFree(ts->probs);
-
-    {	char * currDir = rpmGetCwd();
-	rpmtsSetCurrDir(ts, currDir);
-	currDir = _free(currDir);
-    }
-
-    (void) rpmtsSetChrootDone(ts, 0);
     (void) rpmtsSetTid(ts, tid);
 
     /* Get available space on mounted file systems. */
@@ -997,58 +1281,53 @@ static int rpmtsSetup(rpmts ts, rpmprobFilterFlags ignoreSet)
 static int rpmtsFinish(rpmts ts)
 {
     if (!(rpmtsFlags(ts) & RPMTRANS_FLAG_NOCONTEXTS)) {
-	matchpathcon_fini();
+	rpmtsSELabelFini(ts);
     }
-    return 0;
+    return rpmChrootSet(NULL);
 }
 
 static int rpmtsPrepare(rpmts ts)
 {
+    tsMembers tsmem = rpmtsMembers(ts);
     rpmtsi pi;
     rpmte p;
     rpmfi fi;
-    int xx, rc = 0;
+    int rc = 0;
     uint64_t fileCount = countFiles(ts);
-    const char * rootDir = rpmtsRootDir(ts);
-    int dochroot = (rootDir != NULL && strcmp(rootDir, "/") && *rootDir == '/');
 
     fingerPrintCache fpc = fpCacheCreate(fileCount/2 + 10001);
     rpmFpHash ht = rpmFpHashCreate(fileCount/2+1, fpHashFunction, fpEqual,
 			     NULL, NULL);
+    rpmDiskSpaceInfo dsi;
 
     rpmlog(RPMLOG_DEBUG, "computing %" PRIu64 " file fingerprints\n", fileCount);
 
     /* Skip netshared paths, not our i18n files, and excluded docs */
     pi = rpmtsiInit(ts);
-    while ((p = rpmtsiNext(pi, TR_ADDED)) != NULL) {
-	if ((fi = rpmteFI(p)) == NULL)
-	    continue;	/* XXX can't happen */
-
-	if (rpmfiFC(fi) > 0) 
-	    skipFiles(ts, p);
+    while ((p = rpmtsiNext(pi, 0)) != NULL) {
+	if (rpmfiFC(rpmteFI(p)) == 0)
+	    continue;
+	if (rpmteType(p) == TR_ADDED) {
+	    skipInstallFiles(ts, p);
+	} else {
+	    skipEraseFiles(ts, p);
+	}
     }
     pi = rpmtsiFree(pi);
 
-    /* Enter chroot for fingerprinting if necessary */
-    if (!rpmtsChrootDone(ts)) {
-	xx = chdir("/");
-	if (dochroot) {
-	    /* opening db before chroot not optimal, see rhbz#103852 c#3 */
-	    xx = rpmdbOpenAll(ts->rdb);
-	    if (chroot(rootDir) == -1) {
-		rpmlog(RPMLOG_ERR, _("Unable to change root directory: %m\n"));
-		rc = -1;
-		goto exit;
-	    }
-	}
-	(void) rpmtsSetChrootDone(ts, 1);
+    /* Open rpmdb & enter chroot for fingerprinting if necessary */
+    if (rpmdbOpenAll(ts->rdb) || rpmChrootIn()) {
+	rc = -1;
+	goto exit;
     }
     
-    rpmtsNotify(ts, NULL, RPMCALLBACK_TRANS_START, 6, ts->orderCount);
+    rpmtsNotify(ts, NULL, RPMCALLBACK_TRANS_START, 6, tsmem->orderCount);
     addFingerprints(ts, fileCount, ht, fpc);
     /* check against files in the rpmdb */
     checkInstalledFiles(ts, fileCount, ht, fpc);
 
+    dsi = rpmtsDbDSI(ts);
+
     pi = rpmtsiInit(ts);
     while ((p = rpmtsiNext(pi, 0)) != NULL) {
 	if ((fi = rpmteFI(p)) == NULL)
@@ -1059,6 +1338,8 @@ static int rpmtsPrepare(rpmts ts)
 	   needs on each partition for this package. */
 	handleOverlappedFiles(ts, ht, p, fi);
 
+	rpmtsUpdateDSIrpmDBSize(p, dsi);
+
 	/* Check added package has sufficient space on each partition used. */
 	if (rpmteType(p) == TR_ADDED) {
 	    rpmtsCheckDSIProblems(ts, p);
@@ -1066,17 +1347,11 @@ static int rpmtsPrepare(rpmts ts)
 	(void) rpmswExit(rpmtsOp(ts, RPMTS_OP_FINGERPRINT), 0);
     }
     pi = rpmtsiFree(pi);
-    rpmtsNotify(ts, NULL, RPMCALLBACK_TRANS_STOP, 6, ts->orderCount);
+    rpmtsNotify(ts, NULL, RPMCALLBACK_TRANS_STOP, 6, tsmem->orderCount);
 
     /* return from chroot if done earlier */
-    if (rpmtsChrootDone(ts)) {
-	const char * currDir = rpmtsCurrDir(ts);
-	if (dochroot)
-	    xx = chroot(".");
-	(void) rpmtsSetChrootDone(ts, 0);
-	if (currDir != NULL)
-	    xx = chdir(currDir);
-    }
+    if (rpmChrootOut())
+	rc = -1;
 
     /* File info sets, fp caches etc not needed beyond here, free 'em up. */
     pi = rpmtsiInit(ts);
@@ -1088,6 +1363,7 @@ static int rpmtsPrepare(rpmts ts)
 exit:
     ht = rpmFpHashFree(ht);
     fpc = fpCacheFree(fpc);
+    rpmtsFreeDSI(ts);
     return rc;
 }
 
@@ -1101,46 +1377,15 @@ static int rpmtsProcess(rpmts ts)
 
     pi = rpmtsiInit(ts);
     while ((p = rpmtsiNext(pi, 0)) != NULL) {
-	int failed = 1;
-	rpmElementType tetype = rpmteType(p);
-	rpmtsOpX op = (tetype == TR_ADDED) ? RPMTS_OP_INSTALL : RPMTS_OP_ERASE;
+	int failed;
 
 	rpmlog(RPMLOG_DEBUG, "========== +++ %s %s-%s 0x%x\n",
 		rpmteNEVR(p), rpmteA(p), rpmteO(p), rpmteColor(p));
 
-	if (rpmteFailed(p)) {
-	    /* XXX this should be a warning, need a better message though */
-	    rpmlog(RPMLOG_DEBUG, "element %s marked as failed, skipping\n",
-					rpmteNEVRA(p));
-	    rc++;
-	    continue;
-	}
-	
-	if (rpmteOpen(p, ts, 1)) {
-	    rpmpsm psm = NULL;
-	    pkgStage stage = PSM_UNKNOWN;
-	    int async = (rpmtsiOc(pi) >= rpmtsUnorderedSuccessors(ts, -1)) ? 
-			1 : 0;
-
-	    switch (tetype) {
-	    case TR_ADDED:
-		stage = PSM_PKGINSTALL;
-		break;
-	    case TR_REMOVED:
-		stage = PSM_PKGERASE;
-		break;
-	    }
-	    psm = rpmpsmNew(ts, p);
-	    rpmpsmSetAsync(psm, async);
-
-	    (void) rpmswEnter(rpmtsOp(ts, op), 0);
-	    failed = rpmpsmStage(psm, stage);
-	    (void) rpmswExit(rpmtsOp(ts, op), 0);
-	    psm = rpmpsmFree(psm);
-	    rpmteClose(p, ts, 1);
-	}
+	failed = rpmteProcess(p, rpmteType(p));
 	if (failed) {
-	    rpmteMarkFailed(p, ts);
+	    rpmlog(RPMLOG_ERR, "%s: %s %s\n", rpmteNEVRA(p),
+		   rpmteTypeString(p), failed > 1 ? _("skipped") : _("failed"));
 	    rc++;
 	}
 	(void) rpmdbSync(rpmtsGetRdb(ts));
@@ -1152,10 +1397,14 @@ static int rpmtsProcess(rpmts ts)
 int rpmtsRun(rpmts ts, rpmps okProbs, rpmprobFilterFlags ignoreSet)
 {
     int rc = -1; /* assume failure */
-    void * lock = NULL;
+    rpmlock lock = NULL;
+    rpmps tsprobs = NULL;
+    /* Force default 022 umask during transaction for consistent results */
+    mode_t oldmask = umask(022);
 
-    /* XXX programmer error segfault avoidance. */
+    /* Empty transaction, nothing to do */
     if (rpmtsNElements(ts) <= 0) {
+	rc = 0;
 	goto exit;
     }
 
@@ -1171,44 +1420,52 @@ int rpmtsRun(rpmts ts, rpmps okProbs, rpmprobFilterFlags ignoreSet)
 	goto exit;
     }
 
+    rpmtsSetupCollections(ts);
+
     /* Check package set for problems */
-    ts->probs = checkProblems(ts);
+    tsprobs = checkProblems(ts);
 
     /* Run pre-transaction scripts, but only if there are no known
      * problems up to this point and not disabled otherwise. */
-    if (!((rpmtsFlags(ts) & (RPMTRANS_FLAG_BUILD_PROBS|RPMTRANS_FLAG_TEST|RPMTRANS_FLAG_NOPRE))
-     	  || (rpmpsNumProblems(ts->probs) &&
-		(okProbs == NULL || rpmpsTrim(ts->probs, okProbs))))) {
+    if (!((rpmtsFlags(ts) & (RPMTRANS_FLAG_BUILD_PROBS|RPMTRANS_FLAG_NOPRE))
+     	  || (rpmpsNumProblems(tsprobs)))) {
 	rpmlog(RPMLOG_DEBUG, "running pre-transaction scripts\n");
-	runTransScripts(ts, RPMTAG_PRETRANS);
+	runTransScripts(ts, PKG_PRETRANS);
     }
+    tsprobs = rpmpsFree(tsprobs);
 
     /* Compute file disposition for each package in transaction set. */
     if (rpmtsPrepare(ts)) {
 	goto exit;
     }
+    /* Check again for problems (now including file conflicts,  duh */
+    tsprobs = rpmtsProblems(ts);
 
      /* If unfiltered problems exist, free memory and return. */
-    if ((rpmtsFlags(ts) & RPMTRANS_FLAG_BUILD_PROBS) ||
-		(rpmpsNumProblems(ts->probs) &&
-		(okProbs == NULL || rpmpsTrim(ts->probs, okProbs)))) {
-	rc = ts->orderCount;
+    if ((rpmtsFlags(ts) & RPMTRANS_FLAG_BUILD_PROBS) || (rpmpsNumProblems(tsprobs))) {
+	tsMembers tsmem = rpmtsMembers(ts);
+	rc = tsmem->orderCount;
 	goto exit;
     }
 
+    /* Free up memory taken by problem sets */
+    tsprobs = rpmpsFree(tsprobs);
+    rpmtsCleanProblems(ts);
+
     /* Actually install and remove packages, get final exit code */
     rc = rpmtsProcess(ts) ? -1 : 0;
 
     /* Run post-transaction scripts unless disabled */
-    if (!(rpmtsFlags(ts) & (RPMTRANS_FLAG_TEST|RPMTRANS_FLAG_NOPOST))) {
+    if (!(rpmtsFlags(ts) & (RPMTRANS_FLAG_NOPOST))) {
 	rpmlog(RPMLOG_DEBUG, "running post-transaction scripts\n");
-	runTransScripts(ts, RPMTAG_POSTTRANS);
+	runTransScripts(ts, PKG_POSTTRANS);
     }
 
-    /* Finish up... */
-    (void) rpmtsFinish(ts);
-
 exit:
-    rpmtsFreeLock(lock);
+    /* Finish up... */
+    (void) umask(oldmask);
+    (void) rpmtsFinish(ts);
+    tsprobs = rpmpsFree(tsprobs);
+    rpmlockFree(lock);
     return rc;
 }
