@@ -5,6 +5,14 @@
 
 #include "system.h"
 
+#include <errno.h>
+#if WITH_CAP
+#include <sys/capability.h>
+#endif
+#if WITH_ACL
+#include <acl/libacl.h>
+#endif
+
 #include <rpm/rpmcli.h>
 #include <rpm/header.h>
 #include <rpm/rpmlog.h>
@@ -13,17 +21,15 @@
 #include <rpm/rpmdb.h>
 #include <rpm/rpmfileutil.h>
 
-#include "lib/psm.h"
-#include "lib/misc.h" 	/* uidToUname(), gnameToGid */
-#include "lib/rpmte_internal.h"	/* rpmteOpen(), rpmteClose() */
+#include "lib/misc.h"
+#include "lib/rpmchroot.h"
+#include "lib/rpmte_internal.h"	/* rpmteProcess() */
+#include "lib/rpmug.h"
 
 #include "debug.h"
 
 #define S_ISDEV(m) (S_ISBLK((m)) || S_ISCHR((m)))
 
-extern int _rpmds_unspecified_epoch_noise;
-extern int _cacheDependsRC;
-
 /* If cap_compare() (Linux extension) not available, do it the hard way */
 #if WITH_CAP && !defined(HAVE_CAP_COMPARE)
 static int cap_compare(cap_t acap, cap_t bcap)
@@ -64,11 +70,22 @@ int rpmVerifyFile(const rpmts ts, const rpmfi fi,
      */
     switch (rpmfiFState(fi)) {
     case RPMFILE_STATE_NETSHARED:
-    case RPMFILE_STATE_REPLACED:
     case RPMFILE_STATE_NOTINSTALLED:
-    case RPMFILE_STATE_WRONGCOLOR:
+    case RPMFILE_STATE_MISSING:
 	return 0;
 	break;
+    case RPMFILE_STATE_REPLACED:
+	/* For replaced files we can only verify if it exists at all */
+	flags = RPMVERIFY_LSTATFAIL;
+	break;
+    case RPMFILE_STATE_WRONGCOLOR:
+	/*
+	 * Files with wrong color are supposed to share some attributes
+	 * with the actually installed file - verify what we can.
+	 */
+	flags &= ~(RPMVERIFY_FILEDIGEST | RPMVERIFY_FILESIZE |
+		   RPMVERIFY_MTIME | RPMVERIFY_RDEV);
+	break;
     case RPMFILE_STATE_NORMAL:
 	break;
     }
@@ -87,9 +104,6 @@ int rpmVerifyFile(const rpmts ts, const rpmfi fi,
     else if (S_ISLNK(sb.st_mode)) {
 	flags &= ~(RPMVERIFY_FILEDIGEST | RPMVERIFY_FILESIZE | RPMVERIFY_MTIME |
 			RPMVERIFY_MODE | RPMVERIFY_CAPS);
-#if CHOWN_FOLLOWS_SYMLINK
-	    flags &= ~(RPMVERIFY_USER | RPMVERIFY_GROUP);
-#endif
     }
     else if (S_ISFIFO(sb.st_mode))
 	flags &= ~(RPMVERIFY_FILEDIGEST | RPMVERIFY_FILESIZE | RPMVERIFY_MTIME |
@@ -118,7 +132,7 @@ int rpmVerifyFile(const rpmts ts, const rpmfi fi,
 
     if (flags & RPMVERIFY_FILEDIGEST) {
 	const unsigned char *digest; 
-	pgpHashAlgo algo;
+	int algo;
 	size_t diglen;
 
 	/* XXX If --nomd5, then prelinked library sizes are not corrected. */
@@ -231,22 +245,18 @@ int rpmVerifyFile(const rpmts ts, const rpmfi fi,
 #endif
 
     if ((flags & RPMVERIFY_MTIME) && (sb.st_mtime != rpmfiFMtime(fi))) {
-	/* Filter out timestamp differences of shared files */
-	rpmdbMatchIterator mi = rpmtsInitIterator(ts, RPMTAG_BASENAMES, fn, 0);
-	if (rpmdbGetIteratorCount(mi) < 2) 
-	    *res |= RPMVERIFY_MTIME;
-	rpmdbFreeIterator(mi);
+	*res |= RPMVERIFY_MTIME;
     }
 
     if (flags & RPMVERIFY_USER) {
-	const char * name = uidToUname(sb.st_uid);
+	const char * name = rpmugUname(sb.st_uid);
 	const char * fuser = rpmfiFUser(fi);
 	if (name == NULL || fuser == NULL || !rstreq(name, fuser))
 	    *res |= RPMVERIFY_USER;
     }
 
     if (flags & RPMVERIFY_GROUP) {
-	const char * name = gidToGname(sb.st_gid);
+	const char * name = rpmugGname(sb.st_gid);
 	const char * fgroup = rpmfiFGroup(fi);
 	if (name == NULL || fgroup == NULL || !rstreq(name, fgroup))
 	    *res |= RPMVERIFY_GROUP;
@@ -257,82 +267,115 @@ int rpmVerifyFile(const rpmts ts, const rpmfi fi,
 
 /**
  * Return exit code from running verify script from header.
- * @todo malloc/free/refcount handling is fishy here.
- * @param qva		parsed query/verify options
  * @param ts		transaction set
  * @param h		header
- * @param scriptFd      file handle to use for stderr (or NULL)
  * @return              0 on success
  */
-static int rpmVerifyScript(QVA_t qva, rpmts ts, Header h, FD_t scriptFd)
+static int rpmVerifyScript(rpmts ts, Header h)
 {
-    rpmpsm psm = NULL;
-    rpmte te = NULL;
     int rc = 0;
 
-    /* fake up a erasure transaction element */
-    rc = rpmtsAddEraseElement(ts, h, -1);
-    te = rpmtsElement(ts, 0);
-    rpmteOpen(te, ts, 0);
-    
-    if (scriptFd != NULL)
-	rpmtsSetScriptFd(ts, scriptFd);
+    if (headerIsEntry(h, RPMTAG_VERIFYSCRIPT)) {
+	/* fake up a erasure transaction element */
+	rpmte p = rpmteNew(ts, h, TR_REMOVED, NULL, NULL);
+	rpmteSetHeader(p, h);
 
-    /* create psm to run the script */
-    psm = rpmpsmNew(ts, te);
-    rpmpsmScriptStage(psm, RPMTAG_VERIFYSCRIPT, RPMTAG_VERIFYSCRIPTPROG);
-    rc = rpmpsmStage(psm, PSM_SCRIPT);
-    psm = rpmpsmFree(psm);
+	rc = (rpmpsmRun(ts, p, PKG_VERIFY) != RPMRC_OK);
 
-    if (scriptFd != NULL)
-	rpmtsSetScriptFd(ts, NULL);
-
-    /* clean up our fake transaction bits */
-    rpmteClose(te, ts, 0);
-    rpmtsEmpty(ts);
+	/* clean up our fake transaction bits */
+	rpmteFree(p);
+    }
 
     return rc;
 }
 
+#define unknown "?"
+#define	_verify(_RPMVERIFY_F, _C, _pad)	\
+	((verifyResult & _RPMVERIFY_F) ? _C : _pad)
+#define	_verifylink(_RPMVERIFY_F, _C, _pad)	\
+	((verifyResult & RPMVERIFY_READLINKFAIL) ? unknown : \
+	 (verifyResult & _RPMVERIFY_F) ? _C : _pad)
+#define	_verifyfile(_RPMVERIFY_F, _C, _pad)	\
+	((verifyResult & RPMVERIFY_READFAIL) ? unknown : \
+	 (verifyResult & _RPMVERIFY_F) ? _C : _pad)
+char * rpmVerifyString(uint32_t verifyResult, const char *pad)
+{
+    char *fmt = NULL;
+    rasprintf(&fmt, "%s%s%s%s%s%s%s%s%s",
+		_verifyfile(RPMVERIFY_FILEDIGEST, "5", pad),
+		_verify(RPMVERIFY_FILESIZE, "S", pad),
+		_verifylink(RPMVERIFY_LINKTO, "L", pad),
+		_verify(RPMVERIFY_MTIME, "T", pad),
+		_verify(RPMVERIFY_RDEV, "D", pad),
+		_verify(RPMVERIFY_USER, "U", pad),
+		_verify(RPMVERIFY_GROUP, "G", pad),
+		_verify(RPMVERIFY_MODE, "M", pad),
+		_verify(RPMVERIFY_CAPS, "P", pad));
+		
+    return fmt;
+}
+#undef _verifyfile
+#undef _verifylink
+#undef _verify
+#undef aok
+#undef unknown
+
+char * rpmFFlagsString(uint32_t fflags, const char *pad)
+{
+    char *fmt = NULL;
+    rasprintf(&fmt, "%s%s%s%s%s%s%s%s",
+		(fflags & RPMFILE_DOC) ? "d" : pad,
+		(fflags & RPMFILE_CONFIG) ? "c" : pad,
+		(fflags & RPMFILE_SPECFILE) ? "s" : pad,
+		(fflags & RPMFILE_MISSINGOK) ? "m" : pad,
+		(fflags & RPMFILE_NOREPLACE) ? "n" : pad,
+		(fflags & RPMFILE_GHOST) ? "g" : pad,
+		(fflags & RPMFILE_LICENSE) ? "l" : pad,
+		(fflags & RPMFILE_README) ? "r" : pad);
+    return fmt;
+}
+
 /**
  * Check file info from header against what's actually installed.
- * @param qva		parsed query/verify options
  * @param ts		transaction set
  * @param h		header to verify
+ * @param omitMask	bits to disable verify checks
+ * @param ghosts	should ghosts be verified?
  * @return		0 no problems, 1 problems found
  */
-static int verifyHeader(QVA_t qva, const rpmts ts, Header h)
+static int verifyHeader(rpmts ts, Header h, rpmVerifyAttrs omitMask, int ghosts)
 {
     rpmVerifyAttrs verifyResult = 0;
-    /* FIX: union? */
-    rpmVerifyAttrs omitMask = ((qva->qva_flags & VERIFY_ATTRS) ^ VERIFY_ATTRS);
     int ec = 0;		/* assume no problems */
-    char *buf = NULL;
-
     rpmfi fi = rpmfiNew(ts, h, RPMTAG_BASENAMES, RPMFI_FLAGS_VERIFY);
+
     rpmfiInit(fi, 0);
     while (rpmfiNext(fi) >= 0) {
-	rpmfileAttrs fileAttrs;
+	rpmfileAttrs fileAttrs = rpmfiFFlags(fi);
+	char *buf = NULL, *attrFormat;
+	char ac;
 	int rc;
 
-	fileAttrs = rpmfiFFlags(fi);
-
 	/* If not verifying %ghost, skip ghost files. */
-	if (!(qva->qva_fflags & RPMFILE_GHOST)
-	&& (fileAttrs & RPMFILE_GHOST))
+	if ((fileAttrs & RPMFILE_GHOST) && !ghosts)
 	    continue;
 
 	rc = rpmVerifyFile(ts, fi, &verifyResult, omitMask);
+
+	/* Filter out timestamp differences of shared files */
+	if (rc == 0 && (verifyResult & RPMVERIFY_MTIME)) {
+	    rpmdbMatchIterator mi;
+	    mi = rpmtsInitIterator(ts, RPMDBI_BASENAMES, rpmfiFN(fi), 0);
+	    if (rpmdbGetIteratorCount(mi) > 1) 
+		verifyResult &= ~RPMVERIFY_MTIME;
+	    rpmdbFreeIterator(mi);
+	}
+
+	attrFormat = rpmFFlagsString(fileAttrs, "");
+	ac = rstreq(attrFormat, "") ? ' ' : attrFormat[0];
 	if (rc) {
 	    if (!(fileAttrs & (RPMFILE_MISSINGOK|RPMFILE_GHOST)) || rpmIsVerbose()) {
-		rasprintf(&buf, _("missing   %c %s"),
-			((fileAttrs & RPMFILE_CONFIG)	? 'c' :
-			 (fileAttrs & RPMFILE_DOC)	? 'd' :
-			 (fileAttrs & RPMFILE_GHOST)	? 'g' :
-			 (fileAttrs & RPMFILE_LICENSE)	? 'l' :
-			 (fileAttrs & RPMFILE_PUBKEY)	? 'P' :
-			 (fileAttrs & RPMFILE_README)	? 'r' : ' '), 
-			rpmfiFN(fi));
+		rasprintf(&buf, _("missing   %c %s"), ac, rpmfiFN(fi));
 		if ((verifyResult & RPMVERIFY_LSTATFAIL) != 0 &&
 		    errno != ENOENT) {
 		    char *app;
@@ -343,46 +386,13 @@ static int verifyHeader(QVA_t qva, const rpmts ts, Header h)
 		ec = rc;
 	    }
 	} else if (verifyResult || rpmIsVerbose()) {
-	    const char * size, * filedigest, * link, * mtime, * mode;
-	    const char * group, * user, * rdev, *caps;
-	    static const char *const aok = ".";
-	    static const char *const unknown = "?";
+	    char *verifyFormat = rpmVerifyString(verifyResult, ".");
+	    rasprintf(&buf, "%s  %c %s", verifyFormat, ac, rpmfiFN(fi));
+	    free(verifyFormat);
 
 	    if (verifyResult) ec = 1;
-
-#define	_verify(_RPMVERIFY_F, _C)	\
-	((verifyResult & _RPMVERIFY_F) ? _C : aok)
-#define	_verifylink(_RPMVERIFY_F, _C)	\
-	((verifyResult & RPMVERIFY_READLINKFAIL) ? unknown : \
-	 (verifyResult & _RPMVERIFY_F) ? _C : aok)
-#define	_verifyfile(_RPMVERIFY_F, _C)	\
-	((verifyResult & RPMVERIFY_READFAIL) ? unknown : \
-	 (verifyResult & _RPMVERIFY_F) ? _C : aok)
-	
-	    filedigest = _verifyfile(RPMVERIFY_FILEDIGEST, "5");
-	    size = _verify(RPMVERIFY_FILESIZE, "S");
-	    link = _verifylink(RPMVERIFY_LINKTO, "L");
-	    mtime = _verify(RPMVERIFY_MTIME, "T");
-	    rdev = _verify(RPMVERIFY_RDEV, "D");
-	    user = _verify(RPMVERIFY_USER, "U");
-	    group = _verify(RPMVERIFY_GROUP, "G");
-	    mode = _verify(RPMVERIFY_MODE, "M");
-	    caps = _verify(RPMVERIFY_CAPS, "P");
-
-#undef _verifyfile
-#undef _verifylink
-#undef _verify
-
-	    rasprintf(&buf, "%s%s%s%s%s%s%s%s%s  %c %s",
-			size, mode, filedigest, rdev, link, user, group, mtime, caps,
-			((fileAttrs & RPMFILE_CONFIG)	? 'c' :
-			 (fileAttrs & RPMFILE_DOC)	? 'd' :
-			 (fileAttrs & RPMFILE_GHOST)	? 'g' :
-			 (fileAttrs & RPMFILE_LICENSE)	? 'l' :
-			 (fileAttrs & RPMFILE_PUBKEY)	? 'P' :
-			 (fileAttrs & RPMFILE_README)	? 'r' : ' '), 
-			rpmfiFN(fi));
 	}
+	free(attrFormat);
 
 	if (buf) {
 	    rpmlog(RPMLOG_NOTICE, "%s\n", buf);
@@ -396,41 +406,38 @@ static int verifyHeader(QVA_t qva, const rpmts ts, Header h)
 
 /**
  * Check installed package dependencies for problems.
- * @param qva		parsed query/verify options
  * @param ts		transaction set
  * @param h		header
  * @return		number of problems found (0 for no problems)
  */
-static int verifyDependencies(QVA_t qva, rpmts ts,
-		Header h)
+static int verifyDependencies(rpmts ts, Header h)
 {
     rpmps ps;
-    rpmpsi psi;
-    int rc = 0;		/* assume no problems */
-    int xx;
+    rpmte te;
+    int rc;
 
     rpmtsEmpty(ts);
     (void) rpmtsAddInstallElement(ts, h, NULL, 0, NULL);
 
-    xx = rpmtsCheck(ts);
-    ps = rpmtsProblems(ts);
+    (void) rpmtsCheck(ts);
+    te = rpmtsElement(ts, 0);
+    ps = rpmteProblems(te);
+    rc = rpmpsNumProblems(ps);
 
-    psi = rpmpsInitIterator(ps);
-    if (rpmpsNumProblems(ps) > 0) {
-	char *nevra = headerGetAsString(h, RPMTAG_NEVRA);
-	rpmlog(RPMLOG_NOTICE, _("Unsatisfied dependencies for %s:\n"), nevra);
-	free(nevra);
-	while (rpmpsNextIterator(psi) >= 0) {
-	    rpmProblem p = rpmpsGetProblem(psi);
+    if (rc > 0) {
+	rpmlog(RPMLOG_NOTICE, _("Unsatisfied dependencies for %s:\n"),
+	       rpmteNEVRA(te));
+	rpmpsi psi = rpmpsInitIterator(ps);
+	rpmProblem p;
+
+	while ((p = rpmpsiNext(psi)) != NULL) {
 	    char * ps = rpmProblemString(p);
 	    rpmlog(RPMLOG_NOTICE, "\t%s\n", ps);
 	    free(ps);
-	    rc++;	
 	}
+	psi = rpmpsFreeIterator(psi);
     }
-    psi = rpmpsFreeIterator(psi);
     ps = rpmpsFree(ps);
-
     rpmtsEmpty(ts);
 
     return rc;
@@ -438,25 +445,22 @@ static int verifyDependencies(QVA_t qva, rpmts ts,
 
 int showVerifyPackage(QVA_t qva, rpmts ts, Header h)
 {
+    rpmVerifyAttrs omitMask = ((qva->qva_flags & VERIFY_ATTRS) ^ VERIFY_ATTRS);
+    int ghosts = (qva->qva_fflags & RPMFILE_GHOST);
     int ec = 0;
     int rc;
 
     if (qva->qva_flags & VERIFY_DEPS) {
-	if ((rc = verifyDependencies(qva, ts, h)) != 0)
+	if ((rc = verifyDependencies(ts, h)) != 0)
 	    ec = rc;
     }
     if (qva->qva_flags & VERIFY_FILES) {
-	if ((rc = verifyHeader(qva, ts, h)) != 0)
+	if ((rc = verifyHeader(ts, h, omitMask, ghosts)) != 0)
 	    ec = rc;
     }
-    if ((qva->qva_flags & VERIFY_SCRIPT)
-     && headerIsEntry(h, RPMTAG_VERIFYSCRIPT))
-    {
-	FD_t fdo = fdDup(STDOUT_FILENO);
-	if ((rc = rpmVerifyScript(qva, ts, h, fdo)) != 0)
+    if (qva->qva_flags & VERIFY_SCRIPT) {
+	if ((rc = rpmVerifyScript(ts, h)) != 0)
 	    ec = rc;
-	if (fdo != NULL)
-	    rc = Fclose(fdo);
     }
 
     return ec;
@@ -465,9 +469,7 @@ int showVerifyPackage(QVA_t qva, rpmts ts, Header h)
 int rpmcliVerify(rpmts ts, QVA_t qva, char * const * argv)
 {
     rpmVSFlags vsflags, ovsflags;
-    int ec = 0, xx;
-    int dirfd = -1;
-    const char * rootDir = rpmtsRootDir(ts);
+    int ec = 0;
     FD_t scriptFd = fdDup(STDOUT_FILENO);
 
     /* 
@@ -476,49 +478,39 @@ int rpmcliVerify(rpmts ts, QVA_t qva, char * const * argv)
      */
     rpmtsOpenDB(ts, O_RDONLY);
     rpmdbOpenAll(rpmtsGetRdb(ts));
-    if (rootDir && !rstreq(rootDir, "/")) {
-	dirfd = open(".", O_RDONLY);
-	if (dirfd == -1 || chdir("/") == -1 || chroot(rootDir) == -1) {
-	    rpmlog(RPMLOG_ERR, _("Unable to change root directory: %m\n"));
-	    ec = 1;
-	    goto exit;
-	} else {
-	    rpmtsSetChrootDone(ts, 1);
-	}
+    if (rpmChrootSet(rpmtsRootDir(ts)) || rpmChrootIn()) {
+	ec = 1;
+	goto exit;
     }
 
     if (qva->qva_showPackage == NULL)
         qva->qva_showPackage = showVerifyPackage;
 
-    /* XXX verify flags are inverted from query. */
     vsflags = rpmExpandNumeric("%{?_vsflags_verify}");
-    if (!(qva->qva_flags & VERIFY_DIGEST))
+    if (rpmcliQueryFlags & VERIFY_DIGEST)
 	vsflags |= _RPMVSF_NODIGESTS;
-    if (!(qva->qva_flags & VERIFY_SIGNATURE))
+    if (rpmcliQueryFlags & VERIFY_SIGNATURE)
 	vsflags |= _RPMVSF_NOSIGNATURES;
-    if (!(qva->qva_flags & VERIFY_HDRCHK))
+    if (rpmcliQueryFlags & VERIFY_HDRCHK)
 	vsflags |= RPMVSF_NOHDRCHK;
     vsflags &= ~RPMVSF_NEEDPAYLOAD;
 
+    rpmtsSetScriptFd(ts, scriptFd);
     ovsflags = rpmtsSetVSFlags(ts, vsflags);
     ec = rpmcliArgIter(ts, qva, argv);
     vsflags = rpmtsSetVSFlags(ts, ovsflags);
+    rpmtsSetScriptFd(ts, NULL);
 
     if (qva->qva_showPackage == showVerifyPackage)
         qva->qva_showPackage = NULL;
 
     rpmtsEmpty(ts);
 
-    if (rpmtsChrootDone(ts)) {
-	/* only done if previous chroot succeeded, assume success */
-	xx = chroot(".");
-	xx = fchdir(dirfd);
-	rpmtsSetChrootDone(ts, 0);
-    }
+    if (rpmChrootOut() || rpmChrootSet(NULL))
+	ec = 1;
 
 exit:
     Fclose(scriptFd);
-    if (dirfd >= 0) close(dirfd);
 
     return ec;
 }
