@@ -125,3 +125,17 @@ PyObject *rpmprob_Wrap(PyTypeObject *subtype, rpmProblem prob)
     return (PyObject *) s;
 }
 
+PyObject *rpmps_AsList(rpmps ps)
+{
+    PyObject *problems = PyList_New(0);
+    rpmpsi psi = rpmpsInitIterator(ps);
+    rpmProblem prob;
+
+    while ((prob = rpmpsiNext(psi))) {
+        PyObject *pyprob = rpmprob_Wrap(&rpmProblem_Type, prob);
+        PyList_Append(problems, pyprob);
+        Py_DECREF(pyprob);
+    }
+    rpmpsFreeIterator(psi);
+    return problems;
+}
