@@ -26,11 +26,11 @@ typedef Py_ssize_t (*lenfunc)(PyObject *);
 #endif
 
 #if ((PY_MAJOR_VERSION << 8) | (PY_MINOR_VERSION << 0)) < 0x0206
-#define PyXBytes_Check PyString_Check
-#define PyXBytes_FromString PyString_FromString
-#define PyXBytes_FromStringAndSize PyString_FromStringAndSize
-#define PyXBytes_Size PyString_Size
-#define PyXBytes_AsString PyString_AsString
+#define PyBytes_Check PyString_Check
+#define PyBytes_FromString PyString_FromString
+#define PyBytes_FromStringAndSize PyString_FromStringAndSize
+#define PyBytes_Size PyString_Size
+#define PyBytes_AsString PyString_AsString
 #endif
 
 /* For Python 3, use the PyLong type throughout in place of PyInt */
@@ -38,6 +38,8 @@ typedef Py_ssize_t (*lenfunc)(PyObject *);
 #define PyInt_Check PyLong_Check
 #define PyInt_AsLong PyLong_AsLong
 #define PyInt_FromLong PyLong_FromLong
+#define PyInt_AsUnsignedLongMask PyLong_AsUnsignedLongMask
+#define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
 #endif
 
 #endif	/* H_SYSTEM_PYTHON */
