@@ -23,6 +23,11 @@ static int _print = 0;
 static int _crypto_initialized = 0;
 static int _new_process = 1;
 
+typedef const struct pgpValTbl_s {
+    int val;
+    char const * const str;
+} * pgpValTbl;
+ 
 static struct pgpValTbl_s const pgpSigTypeTbl[] = {
     { PGPSIGTYPE_BINARY,	"Binary document signature" },
     { PGPSIGTYPE_TEXT,		"Text document signature" },
@@ -87,6 +92,7 @@ static struct pgpValTbl_s const pgpHashTbl[] = {
     { PGPHASHALGO_SHA256,	"SHA256" },
     { PGPHASHALGO_SHA384,	"SHA384" },
     { PGPHASHALGO_SHA512,	"SHA512" },
+    { PGPHASHALGO_SHA224,	"SHA224" },
     { -1,			"Unknown hash algorithm" },
 };
 
@@ -202,14 +208,7 @@ static void pgpPrtStr(const char *pre, const char *s)
     fprintf(stderr, " %s", s);
 }
 
-/** \ingroup rpmpgp
- * Return string representation of am OpenPGP value.
- * @param vs		table of (string,value) pairs
- * @param val		byte value to lookup
- * @return		string value of byte
- */
-static inline
-const char * pgpValStr(pgpValTbl vs, uint8_t val)
+static const char * pgpValStr(pgpValTbl vs, uint8_t val)
 {
     do {
 	if (vs->val == val)
@@ -218,6 +217,31 @@ const char * pgpValStr(pgpValTbl vs, uint8_t val)
     return vs->str;
 }
 
+static pgpValTbl pgpValTable(pgpValType type)
+{
+    switch (type) {
+    case PGPVAL_TAG:		return pgpTagTbl;
+    case PGPVAL_ARMORBLOCK:	return pgpArmorTbl;
+    case PGPVAL_ARMORKEY:	return pgpArmorKeyTbl;
+    case PGPVAL_SIGTYPE:	return pgpSigTypeTbl;
+    case PGPVAL_SUBTYPE:	return pgpSubTypeTbl;
+    case PGPVAL_PUBKEYALGO:	return pgpPubkeyTbl;
+    case PGPVAL_SYMKEYALGO:	return pgpSymkeyTbl;
+    case PGPVAL_COMPRESSALGO:	return pgpCompressionTbl;
+    case PGPVAL_HASHALGO: 	return pgpHashTbl;
+    case PGPVAL_SERVERPREFS:	return pgpKeyServerPrefsTbl;
+    default:
+	break;
+    }
+    return NULL;
+}
+
+const char * pgpValString(pgpValType type, uint8_t val)
+{
+    pgpValTbl tbl = pgpValTable(type);
+    return (tbl != NULL) ? pgpValStr(tbl, val) : NULL;
+}
+
 static void pgpPrtHex(const char *pre, const uint8_t *p, size_t plen)
 {
     char *hex = NULL;
@@ -286,7 +310,7 @@ int pgpValTok(pgpValTbl vs, const char * s, const char * se)
 {
     do {
 	size_t vlen = strlen(vs->str);
-	if (vlen <= (se-s) && !strncmp(s, vs->str, vlen))
+	if (vlen <= (se-s) && rstreqn(s, vs->str, vlen))
 	    break;
     } while ((++vs)->val != -1);
     return vs->val;
@@ -295,12 +319,12 @@ int pgpValTok(pgpValTbl vs, const char * s, const char * se)
  * @return		0 on success
  */
 static int pgpMpiSet(const char * pre, unsigned int lbits,
-		void *dest, const uint8_t * p, const uint8_t * pend)
+		uint8_t *dest, const uint8_t * p, const uint8_t * pend)
 {
     unsigned int mbits = pgpMpiBits(p);
     unsigned int nbits;
     size_t nbytes;
-    char *t = dest;
+    uint8_t *t = dest;
     unsigned int ix;
 
     if ((p + ((mbits+7) >> 3)) > pend)
@@ -375,16 +399,6 @@ static SECKEYPublicKey *pgpNewPublicKey(KeyType type)
     return key;
 }
 
-static SECKEYPublicKey *pgpNewRSAKey(void)
-{
-    return pgpNewPublicKey(rsaKey);
-}
-
-static SECKEYPublicKey *pgpNewDSAKey(void)
-{
-    return pgpNewPublicKey(dsaKey);
-}
-
 /** \ingroup rpmpgp
  * Is buffer at beginning of an OpenPGP packet?
  * @param p		buffer
@@ -595,8 +609,8 @@ static int pgpPrtSigParams(pgpTag tag, uint8_t pubkey_algo, uint8_t sigtype,
 	    {
 		switch (i) {
 		case 0:		/* m**d */
-		    _dig->rsasig = pgpMpiItem(NULL, _dig->rsasig, p);
-		    if (_dig->rsasig == NULL)
+		    _dig->sigdata = pgpMpiItem(NULL, _dig->sigdata, p);
+		    if (_dig->sigdata == NULL)
 			return 1;
 		    break;
 		default:
@@ -619,13 +633,13 @@ static int pgpPrtSigParams(pgpTag tag, uint8_t pubkey_algo, uint8_t sigtype,
 		    break;
 		case 1:		/* s */
 		    xx = pgpMpiSet(pgpSigDSA[i], DSA_SUBPRIME_LEN*8, dsaraw.data + DSA_SUBPRIME_LEN, p, pend);
-		    if (_dig->dsasig != NULL)
-		    	SECITEM_FreeItem(_dig->dsasig, PR_FALSE);
-		    else if ((_dig->dsasig=SECITEM_AllocItem(NULL, NULL, 0)) == NULL) {
+		    if (_dig->sigdata != NULL)
+		    	SECITEM_FreeItem(_dig->sigdata, PR_FALSE);
+		    else if ((_dig->sigdata=SECITEM_AllocItem(NULL, NULL, 0)) == NULL) {
 		        xx = 1;
 		        break;
 		    }
-		    if (DSAU_EncodeDerSig(_dig->dsasig, &dsaraw) != SECSuccess)
+		    if (DSAU_EncodeDerSig(_dig->sigdata, &dsaraw) != SECSuccess)
 		    	xx = 1;
 		    break;
 		default:
@@ -820,22 +834,26 @@ static const uint8_t * pgpPrtPubkeyParams(uint8_t pubkey_algo,
 {
     size_t i;
 
+    /* XXX we can't handle more than one key in a packet, error out */
+    if (_dig && _dig->keydata)
+	return NULL;
+
     for (i = 0; p < &h[hlen]; i++, p += pgpMpiLen(p)) {
 	char * mpi;
 	if (pubkey_algo == PGPPUBKEYALGO_RSA) {
 	    if (i >= 2) break;
 	    if (_dig) {
-		if (_dig->rsa == NULL) {
-		    _dig->rsa = pgpNewRSAKey();
-		    if (_dig->rsa == NULL)
-			break; /* error abort? */
+		if (_dig->keydata == NULL) {
+		    _dig->keydata = pgpNewPublicKey(rsaKey);
+		    if (_dig->keydata == NULL)
+			return NULL;
 		}
 		switch (i) {
 		case 0:		/* n */
-		    pgpMpiItem(_dig->rsa->arena, &_dig->rsa->u.rsa.modulus, p);
+		    pgpMpiItem(_dig->keydata->arena, &_dig->keydata->u.rsa.modulus, p);
 		    break;
 		case 1:		/* e */
-		    pgpMpiItem(_dig->rsa->arena, &_dig->rsa->u.rsa.publicExponent, p);
+		    pgpMpiItem(_dig->keydata->arena, &_dig->keydata->u.rsa.publicExponent, p);
 		    break;
 		default:
 		    break;
@@ -845,23 +863,23 @@ static const uint8_t * pgpPrtPubkeyParams(uint8_t pubkey_algo,
 	} else if (pubkey_algo == PGPPUBKEYALGO_DSA) {
 	    if (i >= 4) break;
 	    if (_dig) {
-		if (_dig->dsa == NULL) {
-		    _dig->dsa = pgpNewDSAKey();
-		    if (_dig->dsa == NULL)
-			break; /* error abort? */
+		if (_dig->keydata == NULL) {
+		    _dig->keydata = pgpNewPublicKey(dsaKey);
+		    if (_dig->keydata == NULL)
+			return NULL;
 		}
 		switch (i) {
 		case 0:		/* p */
-		    pgpMpiItem(_dig->dsa->arena, &_dig->dsa->u.dsa.params.prime, p);
+		    pgpMpiItem(_dig->keydata->arena, &_dig->keydata->u.dsa.params.prime, p);
 		    break;
 		case 1:		/* q */
-		    pgpMpiItem(_dig->dsa->arena, &_dig->dsa->u.dsa.params.subPrime, p);
+		    pgpMpiItem(_dig->keydata->arena, &_dig->keydata->u.dsa.params.subPrime, p);
 		    break;
 		case 2:		/* g */
-		    pgpMpiItem(_dig->dsa->arena, &_dig->dsa->u.dsa.params.base, p);
+		    pgpMpiItem(_dig->keydata->arena, &_dig->keydata->u.dsa.params.base, p);
 		    break;
 		case 3:		/* y */
-		    pgpMpiItem(_dig->dsa->arena, &_dig->dsa->u.dsa.publicValue, p);
+		    pgpMpiItem(_dig->keydata->arena, &_dig->keydata->u.dsa.publicValue, p);
 		    break;
 		default:
 		    break;
@@ -988,7 +1006,7 @@ static int pgpPrtKey(pgpTag tag, const uint8_t *h, size_t hlen,
 
 	p = ((uint8_t *)v) + sizeof(*v);
 	p = pgpPrtPubkeyParams(v->pubkey_algo, p, h, hlen, _dig);
-	rc = 0;
+	rc = (p == NULL);
     }	break;
     case 4:
     {   pgpPktKeyV4 v = (pgpPktKeyV4)h;
@@ -1009,7 +1027,7 @@ static int pgpPrtKey(pgpTag tag, const uint8_t *h, size_t hlen,
 	p = pgpPrtPubkeyParams(v->pubkey_algo, p, h, hlen, _dig);
 	if (!(tag == PGPTAG_PUBLIC_KEY || tag == PGPTAG_PUBLIC_SUBKEY))
 	    p = pgpPrtSeckeyParams(v->pubkey_algo, p, h, hlen);
-	rc = 0;
+	rc = (p == NULL);
     }	break;
     default:
 	rc = 1;
@@ -1026,9 +1044,9 @@ static int pgpPrtUserID(pgpTag tag, const uint8_t *h, size_t hlen,
 	fprintf(stderr, " \"%.*s\"", (int)hlen, (const char *)h);
     pgpPrtNL();
     if (_digp) {
-	char * t;
-	_digp->userid = t = memcpy(xmalloc(hlen+1), h, hlen);
-	t[hlen] = '\0';
+	free(_digp->userid);
+	_digp->userid = memcpy(xmalloc(hlen+1), h, hlen);
+	_digp->userid[hlen] = '\0';
     }
     return 0;
 }
@@ -1190,11 +1208,6 @@ static int pgpPrtPkt(const uint8_t *pkt, size_t pleft,
 	    else
 		memset(_digp->signid, 0, sizeof(_digp->signid));
 	}
-    case PGPTAG_PUBLIC_SUBKEY:
-	rc = pgpPrtKey(tag, h, hlen, _dig, _digp);
-	break;
-    case PGPTAG_SECRET_KEY:
-    case PGPTAG_SECRET_SUBKEY:
 	rc = pgpPrtKey(tag, h, hlen, _dig, _digp);
 	break;
     case PGPTAG_USER_ID:
@@ -1205,6 +1218,9 @@ static int pgpPrtPkt(const uint8_t *pkt, size_t pleft,
 	rc = pgpPrtComment(tag, h, hlen);
 	break;
 
+    case PGPTAG_PUBLIC_SUBKEY:
+    case PGPTAG_SECRET_KEY:
+    case PGPTAG_SECRET_SUBKEY:
     case PGPTAG_RESERVED:
     case PGPTAG_PUBLIC_SESSION_KEY:
     case PGPTAG_SYMMETRIC_SESSION_KEY:
@@ -1253,29 +1269,15 @@ void pgpCleanDig(pgpDig dig)
 	memset(&dig->signature, 0, sizeof(dig->signature));
 	memset(&dig->pubkey, 0, sizeof(dig->pubkey));
 
-	dig->md5 = _free(dig->md5);
-	dig->sha1 = _free(dig->sha1);
-
-	if (dig->dsa != NULL) {
-	    SECKEY_DestroyPublicKey(dig->dsa);
-	    dig->dsa = NULL;
+	if (dig->keydata != NULL) {
+	    SECKEY_DestroyPublicKey(dig->keydata);
+	    dig->keydata = NULL;
 	}
 
-	if (dig->dsasig != NULL) {
-	    SECITEM_ZfreeItem(dig->dsasig, PR_TRUE);
-	    dig->dsasig = NULL;
+	if (dig->sigdata != NULL) {
+	    SECITEM_ZfreeItem(dig->sigdata, PR_TRUE);
+	    dig->sigdata = NULL;
 	}
-
-	if (dig->rsa != NULL) {
-	    SECKEY_DestroyPublicKey(dig->rsa);
-	    dig->rsa = NULL;
-	}
-
-	if (dig->rsasig != NULL) {
-	    SECITEM_ZfreeItem(dig->rsasig, PR_TRUE);
-	    dig->rsasig = NULL;
-	}
-
     }
     return;
 }
@@ -1286,23 +1288,6 @@ pgpDig pgpFreeDig(pgpDig dig)
 
 	/* DUmp the signature/pubkey data. */
 	pgpCleanDig(dig);
-
-	if (dig->hdrsha1ctx != NULL)
-	    (void) rpmDigestFinal(dig->hdrsha1ctx, NULL, NULL, 0);
-	dig->hdrsha1ctx = NULL;
-
-	if (dig->sha1ctx != NULL)
-	    (void) rpmDigestFinal(dig->sha1ctx, NULL, NULL, 0);
-	dig->sha1ctx = NULL;
-
-	if (dig->hdrmd5ctx != NULL)
-	    (void) rpmDigestFinal(dig->hdrmd5ctx, NULL, NULL, 0);
-	dig->hdrmd5ctx = NULL;
-
-	if (dig->md5ctx != NULL)
-	    (void) rpmDigestFinal(dig->md5ctx, NULL, NULL, 0);
-	dig->md5ctx = NULL;
-
 	dig = _free(dig);
     }
     return dig;
@@ -1334,10 +1319,131 @@ int pgpPrtPkts(const uint8_t * pkts, size_t pktlen, pgpDig dig, int printing)
     return 0;
 }
 
-pgpArmor pgpReadPkts(const char * fn, uint8_t ** pkt, size_t * pktlen)
+static SECOidTag getSigAlg(pgpDigParams sigp)
+{
+    SECOidTag sigalg = SEC_OID_UNKNOWN;
+    if (sigp->pubkey_algo == PGPPUBKEYALGO_DSA) {
+	/* assume SHA1 for now, NSS doesn't have SECOID's for other types */
+	sigalg = SEC_OID_ANSIX9_DSA_SIGNATURE_WITH_SHA1_DIGEST;
+    } else if (sigp->pubkey_algo == PGPPUBKEYALGO_RSA) {
+	switch (sigp->hash_algo) {
+	case PGPHASHALGO_MD5:
+	    sigalg = SEC_OID_PKCS1_MD5_WITH_RSA_ENCRYPTION;
+	    break;
+	case PGPHASHALGO_MD2:
+	    sigalg = SEC_OID_PKCS1_MD2_WITH_RSA_ENCRYPTION;
+	    break;
+	case PGPHASHALGO_SHA1:
+	    sigalg = SEC_OID_PKCS1_SHA1_WITH_RSA_ENCRYPTION;
+	    break;
+	case PGPHASHALGO_SHA256:
+	    sigalg = SEC_OID_PKCS1_SHA256_WITH_RSA_ENCRYPTION;
+	    break;
+	case PGPHASHALGO_SHA384:
+	     sigalg = SEC_OID_PKCS1_SHA384_WITH_RSA_ENCRYPTION;
+	    break;
+	case PGPHASHALGO_SHA512:
+	    sigalg = SEC_OID_PKCS1_SHA512_WITH_RSA_ENCRYPTION;
+	    break;
+	default:
+	    break;
+	}
+    }
+    return sigalg;
+}
+
+char *pgpIdentItem(pgpDigParams digp)
+{
+    char *id = NULL;
+    if (digp) {
+	
+	char *signid = pgpHexStr(digp->signid+4, sizeof(digp->signid)-4);
+	rasprintf(&id, _("V%d %s/%s %s, key ID %s"),
+			digp->version,
+			pgpValStr(pgpPubkeyTbl, digp->pubkey_algo),
+			pgpValStr(pgpHashTbl, digp->hash_algo),
+			pgpValStr(pgpTagTbl, digp->tag),
+			signid);
+	free(signid);
+    } else {
+	id = xstrdup(_("(none)"));
+    }
+    return id;
+}
+
+rpmRC pgpVerifySig(pgpDig dig, DIGEST_CTX hashctx)
+{
+    DIGEST_CTX ctx = rpmDigestDup(hashctx);
+    uint8_t *hash = NULL;
+    size_t hashlen = 0;
+    rpmRC res = RPMRC_FAIL; /* assume failure */
+    pgpDigParams sigp = dig ? &dig->signature : NULL;
+
+    if (sigp == NULL || ctx == NULL)
+	goto exit;
+
+    if (sigp->hash != NULL)
+	rpmDigestUpdate(ctx, sigp->hash, sigp->hashlen);
+
+    if (sigp->version == 4) {
+	/* V4 trailer is six octets long (rfc4880) */
+	uint8_t trailer[6];
+	uint32_t nb = sigp->hashlen;
+	nb = htonl(nb);
+	trailer[0] = sigp->version;
+	trailer[1] = 0xff;
+	memcpy(trailer+2, &nb, 4);
+	rpmDigestUpdate(ctx, trailer, sizeof(trailer));
+    }
+
+    rpmDigestFinal(ctx, (void **)&hash, &hashlen, 0);
+
+    /* Compare leading 16 bits of digest for quick check. */
+    if (hash == NULL || memcmp(hash, sigp->signhash16, 2) != 0)
+	goto exit;
+
+    /*
+     * If we have a key, verify the signature for real. Otherwise we've
+     * done all we can, return NOKEY to indicate "looks okay but dunno."
+     */
+    if (dig->keydata == NULL) {
+	res = RPMRC_NOKEY;
+    } else {
+	SECItem digest = { .type = siBuffer, .data = hash, .len = hashlen };
+	SECItem *sig = dig->sigdata;
+
+	/* Zero-pad RSA signature to expected size if necessary */
+	if (sigp->pubkey_algo == PGPPUBKEYALGO_RSA) {
+	    size_t siglen = SECKEY_SignatureLen(dig->keydata);
+	    if (siglen > sig->len) {
+		size_t pad = siglen - sig->len;
+		if ((sig = SECITEM_AllocItem(NULL, NULL, siglen)) == NULL) {
+		    goto exit;
+		}
+		memset(sig->data, 0, pad);
+		memcpy(sig->data+pad, dig->sigdata->data, dig->sigdata->len);
+	    }
+	}
+
+	/* XXX VFY_VerifyDigest() is deprecated in NSS 3.12 */ 
+	if (VFY_VerifyDigest(&digest, dig->keydata, sig,
+			     getSigAlg(sigp), NULL) == SECSuccess) {
+	    res = RPMRC_OK;
+	}
+
+	if (sig != dig->sigdata) {
+	    SECITEM_ZfreeItem(sig, 1);
+	}
+    }
+
+exit:
+    free(hash);
+    return res;
+
+}
+
+static pgpArmor decodePkts(uint8_t *b, uint8_t **pkt, size_t *pktlen)
 {
-    uint8_t * b = NULL;
-    ssize_t blen;
     const char * enc = NULL;
     const char * crcenc = NULL;
     uint8_t * dec;
@@ -1349,13 +1455,6 @@ pgpArmor pgpReadPkts(const char * fn, uint8_t ** pkt, size_t * pktlen)
     char * t, * te;
     int pstate = 0;
     pgpArmor ec = PGPARMOR_ERR_NO_BEGIN_PGP;	/* XXX assume failure */
-    int rc;
-
-    rc = rpmioSlurp(fn, &b, &blen);
-    if (rc || b == NULL || blen <= 0) {
-	goto exit;
-    }
-
     if (pgpIsPkt(b)) {
 #ifdef NOTYET	/* XXX ASCII Pubkeys only, please. */
 	ec = 0;	/* XXX fish out pkt type. */
@@ -1363,9 +1462,10 @@ pgpArmor pgpReadPkts(const char * fn, uint8_t ** pkt, size_t * pktlen)
 	goto exit;
     }
 
-#define	TOKEQ(_s, _tok)	(!strncmp((_s), (_tok), sizeof(_tok)-1))
+#define	TOKEQ(_s, _tok)	(rstreqn((_s), (_tok), sizeof(_tok)-1))
 
     for (t = (char *)b; t && *t; t = te) {
+	int rc;
 	if ((te = strchr(t, '\n')) == NULL)
 	    te = t + strlen(t);
 	else
@@ -1385,10 +1485,13 @@ pgpArmor pgpReadPkts(const char * fn, uint8_t ** pkt, size_t * pktlen)
 	    }
 	    if (rc != PGPARMOR_PUBKEY)	/* XXX ASCII Pubkeys only, please. */
 		continue;
-	    armortype = t;
 
-	    t = te - (sizeof("-----\n")-1);
-	    if (!TOKEQ(t, "-----\n"))
+	    armortype = pgpValStr(pgpArmorTbl, rc);
+	    t += strlen(armortype);
+	    if (!TOKEQ(t, "-----"))
+		continue;
+	    t += sizeof("-----")-1;
+	    if (*t != '\n' && *t != '\r')
 		continue;
 	    *t = '\0';
 	    pstate++;
@@ -1398,7 +1501,7 @@ pgpArmor pgpReadPkts(const char * fn, uint8_t ** pkt, size_t * pktlen)
 	    rc = pgpValTok(pgpArmorKeyTbl, t, te);
 	    if (rc >= 0)
 		continue;
-	    if (*t != '\n') {
+	    if (*t != '\n' && *t != '\r') {
 		pstate = 0;
 		continue;
 	    }
@@ -1425,8 +1528,7 @@ pgpArmor pgpReadPkts(const char * fn, uint8_t ** pkt, size_t * pktlen)
 
 	    if (armortype == NULL) /* XXX can't happen */
 		continue;
-	    rc = strncmp(t, armortype, strlen(armortype));
-	    if (rc)
+	    if (!rstreqn(t, armortype, strlen(armortype)))
 		continue;
 
 	    t += strlen(armortype);
@@ -1460,9 +1562,8 @@ pgpArmor pgpReadPkts(const char * fn, uint8_t ** pkt, size_t * pktlen)
 		ec = PGPARMOR_ERR_CRC_CHECK;
 		goto exit;
 	    }
-	    b = _free(b);
-	    b = dec;
-	    blen = declen;
+	    if (pkt) *pkt = dec;
+	    if (pktlen) *pktlen = declen;
 	    ec = PGPARMOR_PUBKEY;	/* XXX ASCII Pubkeys only, please. */
 	    goto exit;
 	    break;
@@ -1471,12 +1572,30 @@ pgpArmor pgpReadPkts(const char * fn, uint8_t ** pkt, size_t * pktlen)
     ec = PGPARMOR_NONE;
 
 exit:
-    if (ec > PGPARMOR_NONE && pkt)
-	*pkt = b;
-    else if (b != NULL)
-	b = _free(b);
-    if (pktlen)
-	*pktlen = blen;
+    return ec;
+}
+
+pgpArmor pgpReadPkts(const char * fn, uint8_t ** pkt, size_t * pktlen)
+{
+    uint8_t * b = NULL;
+    ssize_t blen;
+    pgpArmor ec = PGPARMOR_ERR_NO_BEGIN_PGP;	/* XXX assume failure */
+    int rc = rpmioSlurp(fn, &b, &blen);
+    if (rc == 0 && b != NULL && blen > 0) {
+	ec = decodePkts(b, pkt, pktlen);
+    }
+    free(b);
+    return ec;
+}
+
+pgpArmor pgpParsePkts(const char *armor, uint8_t ** pkt, size_t * pktlen)
+{
+    pgpArmor ec = PGPARMOR_ERR_NO_BEGIN_PGP;	/* XXX assume failure */
+    if (armor && strlen(armor) > 0) {
+	uint8_t *b = (uint8_t*) xstrdup(armor);
+	ec = decodePkts(b, pkt, pktlen);
+	free(b);
+    }
     return ec;
 }
 
