@@ -50,21 +50,13 @@ marc@redhat.com and ewt@redhat.com.
    
 */
 
-#include "system.h"
-
-/*
-  these includes are for my use, rpm will use #include "system.h"*
-*/
-
-/*
 #include <stdio.h>
 #include <stdlib.h>
-#include <string.h>
-#include <ctype.h>
-#include <netinet/in.h>
-*/
-
 #include <stdarg.h>
+#include <string.h>
+#include <netdb.h>
+#include <rpm/rpmutil.h>
+#include <ctype.h>
 #include "debug.h"
 
 /*---------typedefs---------*/
@@ -281,7 +273,7 @@ usage (void)
   exit(-1);
 }
 
-
+RPM_GNUC_NORETURN
 void outofmemory(void) {
 
   /* Its doubtful we could do a printf if there is really a memory
@@ -291,7 +283,7 @@ void outofmemory(void) {
   exit(-1);
 }
 
-
+RPM_GNUC_NORETURN
 void die(const char *format, ...) {
   /* Most errors are fatal.
      This function throws a fatal error and 
@@ -993,7 +985,7 @@ char*
 findClassName (FILE *fileHandle, symbolTable_t *symbolTable) {
   char ignore[10];
   unsigned short type = 0;
-  unsigned short class = 0;
+  unsigned short jclass = 0;
   char *className;
 
   /* seek a little past the end of the table */
@@ -1004,13 +996,13 @@ findClassName (FILE *fileHandle, symbolTable_t *symbolTable) {
   
   my_fread(&type, 2, 1, fileHandle);
   type=ntohs(type);
-  class = symbolTable->classRef[type];
-  if( !class ||
-      !symbolTable->stringList[class] ) {
-      die("Couln't find class: %d, provided by file.\n", class);
+  jclass = symbolTable->classRef[type];
+  if( !jclass ||
+      !symbolTable->stringList[jclass] ) {
+      die("Couln't find class: %d, provided by file.\n", jclass);
   }
 
-  className=symbolTable->stringList[class];
+  className=symbolTable->stringList[jclass];
   
   return className;
 
